/**
 * @file aatree.c
 * Cop-out implementation of an AA-tree. Really a static vector; use until the
 * real tree is fixed. Maintains the same tree interface as the original tree.
 * @ingroup AATree
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "aatree.h"
#include <stdlib.h>
#include <assert.h>

#include "vector.h"

static Vector copout = { NULL, NULL, NULL, sizeof(AANode), 0, 0, 16 };

AANode aaTerminatorNode = { { NULL }, /* &aaTerminatorNode, &aaTerminatorNode, */
	0, -1 };

AANode *AATreeFind(AANode *tree, AATreeKeyLevelType key) {
	AANode *node;
	int loop;
	if (copout.array == NULL) return NULL;
	for (loop = copout.items, node = copout.array; loop > 0; loop--, node++) {
		if (node->key == key)
			return node;
	}
	// the item sought is not in the tree
	return NULL;
}

AANode *AATreeAdd(AANode **tree, AATreeKeyLevelType key) {
	AANode *node;
	// the key must not be the same as the terminator
	assert(key != 0xFFFF);
	if (copout.array == NULL) {
		VectorInit(&copout, 16);
	}
	// attempt to allocate a new node and check for success
	if ((node = VectorAdd(&copout)) != NULL) {
		// initialize the new node
		node->key = key;
		*tree = node;
		return node;
	}
	// failed
	return NULL;
}

AANode *_AATreeRemove(AANode *tree, AATreeRemovalData *rd) {
	AANode *node;
	unsigned int loop;
	if (copout.array == NULL) return NULL;
	for (loop = 0, node = copout.array; loop < copout.items; loop++, node++) {
		if (node->key == rd->key) {
			VectorRemove(&copout, loop);
			if (copout.items)
				return (AANode*)copout.array;
			return &aaTerminatorNode;
		}
	}
	// the item sought is not in the tree
	return NULL;
}

static Bool setExit = FALSE;

static void treeExit(void) {
	VectorDestroy(&copout);
}

void AATreeDestroy(AANode *tree) {
	VectorClear(&copout);
	if (!setExit) {
		atexit(treeExit);
		setExit = TRUE;
	}
}

