/**
 * @file aatree.h
 * Interface to an AA-tree.
 * @ingroup AATree
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

/**
 * @defgroup DataStruct Generic Data Structures
 * A set of data structure implementations with interfaces that allow for
 * generic use.
 *
 * To assist in implementing these algorithms, I used Mark Allen Weiss's text
 * "Data Structures and Problem Solving Using C++", second edition. While I
 * didn't have a copy of the C version of the text, I still found it extremely
 * useful. The source code here is not copied from the text; I developed it
 * based on what I learned from the text.
 */

/**
 * @defgroup AATree AA-Tree
 * An implementation of an AA-Tree -- a balanced binary tree. AA-Trees are
 * simplified Red-Black trees with an easier implementation.
 *
 * @bug The tree is broken. Nodes do not continue to track the same item. If
 *      the item contains the key value, like MsgDescr, eventually AATreeFind()
 *      will return a node who's key field is correct, but the data item will
 *      have a different key value indicating that the node's data has
 *      incorrectly changed.
 *
 * This AA-Tree implementation lacks a data type that represents the tree as
 * a whole. This was done as a way to improve efficiency a little, although
 * it may make using this implementation a little more confusing, so maybe
 * I'll change that in the future. For now, a tree is represented only by a
 * pointer to its root node. New, empty trees should have a pointer to
 * ::AA_ROOT_NODE as thier root node.
 *
 * The implementation does not manage memory for any data being tracked by a
 * tree. However, it does dynamically allocate and deallocate the AANode
 * structures that make up the trees.
 *
 * @ingroup DataStruct
 * 
 * @{
 */

#include <SDL_types.h>
#include <stddef.h>

/**
 * The data type used to identify the items in the AA-Tree. This must be an
 * integer type. A 16-bit value was selected because it works well for Tank!
 */
typedef Uint16  AATreeKeyLevelType;

typedef struct AANode_t  AANode;

/**
 * The node data used to track the structure of an AA-Tree.
 */
struct AANode_t {
	union {
		/**
		 * A pointer to the arbitrary data being tracked by the tree. This is
		 * needed because the remove algorithm moves nodes about in a way that
		 * requires copying node data and takes advantage of the fact that other
		 * nodes still have pointers to the copied node. So without this pointer,
		 * all the data for the node would also have to be copied.
		 */
		void *data;
		/**
		 * An alternative to a pointer to the associated data; a value intended
		 * to be an index into an array. It can safely be used in any other way.
		 */
		size_t index;
	};
	/**
	 * The lesser branch of the tree.
	 */
	//AANode *left;
	/**
	 * The greater branch of the tree.
	 */
	//AANode *right;
	/**
	 * The item's level in the tree.
	 * @internal
	 */
	AATreeKeyLevelType level;
	/**
	 * The unique identifier for the item.
	 * @warning This should be set prior to adding the item to the tree and
	 *          should never be changed until after the item is removed.
	 */
	AATreeKeyLevelType key;
};

/**
 * The node used to indicate that a particular branch is empty. This is used in
 * place of NULL to simplify the code, and defined only once to avoid having
 * multiple copies.
 * @note  This is not defined as a const type becuase the code will still run
 *        assignments that change values within the terminator, although only
 *        to the value that is already there. It simplifies the code to able
 *        to use assignments in this way.
 */
extern AANode aaTerminatorNode;

/**
 * A pointer to the initial root node; if it is the root node of a tree, then
 * the tree is empty.
 */
#define AA_ROOT_NODE  (&aaTerminatorNode)

/**
 * Evaluates to true if the given tree is empty.
 * @param  tree  A pointer to the root node of the tree to check.
 * @return True if the tree is empty, false if there is something within the
 *         tree.
 * @author Jeff Jackowski
 */
#define AATreeIsEmpty(tree)  ((tree) == &aaTerminatorNode)

/**
 * Used by AATreeRemove() to store data used to remove items from the tree.
 * The data's scope lasts the duration of the removal operation rather than
 * the duration of the tree's existance.
 * @internal
 * @author Jeff Jackowski
 */
struct AATreeRemovalData_t {
	/**
	 * The node visited previously by the removal algorithm.
	 */
	AANode *lastNode;
	/**
	 * Points to a node that may need to be removed.
	 */
	AANode *deletedNode;
	/**
	 * The key identifying the data that needs to be removed.
	 */
	AATreeKeyLevelType key;
};

typedef struct AATreeRemovalData_t  AATreeRemovalData;

/**
 * Finds a node in the tree.
 * @param tree  A pointer to the root node of the tree to search.
 * @param key   The key identifying the node to find.
 * @return  A pointer to the sought after node, or NULL if the node is not in
 *          the tree.
 * @author  Jeff Jackowski
 */
AANode *AATreeFind(AANode *tree, AATreeKeyLevelType key);

/**
 * Adds a node to the indicated tree.
 * @param tree  A pointer to the root node of the tree that will be given
 *              the new item. This pointer will be changed to the new root
 *              node. In case of error, it will remain unchanged.
 * @param key   The key of the node to add.
 * @return      A pointer to the added node, or NULL if the node could not
 *              be added. The node's key field will be filled out. The
 *              data / index field will be uninitialized.
 * @author  Jeff Jackowski
 */
AANode *AATreeAdd(AANode **tree, AATreeKeyLevelType key);

/**
 * Implements the recursive removal algorithim; removes and deallocates the
 * node with the given key.
 * @internal    Use the AATreeRemove() macro instead of calling this function
 *              directly.
 * @param tree  A pointer to the next branch of the tree to work with.
 * @param rd    Removal scope data.
 * @return      The new value for the passed in tree pointer, or NULL if the
 *              node sought is not in the tree.
 * @author      Jeff Jackowski
 */
AANode *_AATreeRemove(AANode *tree, AATreeRemovalData *rd);

/**
 * Removes an item from the given tree. This is a macro implementation that has
 * a little code to start the removal operationn to avoid putting that code in
 * a really small function. I just didn't feel like making that function.
 * @param tree  A pointer to the root node of a tree. This pointer will be
 *              altered to point to the new root node.
 * @param key   The key that identifies the item to remove from the tree. The
 *              corresponding node will be deallocated.
 * @author Jeff Jackowski
 */
#define AATreeRemove(tree, key) if ((tree) != &aaTerminatorNode) { \
	AANode *newRoot;  \
	AATreeRemovalData rd = { NULL, &aaTerminatorNode, key }; \
	if ((newRoot = _AATreeRemove(tree, &rd)) != NULL) tree = newRoot; \
}

/**
 * Deallocates all AANodes that comprise a tree.
 * @warning     The root node of the tree will not be valid after this
 *              function returns.
 * @param tree  The root node of the tree to deallocate. The tree must not
 *              already be empty. The root node will not be valid after this
 *              function returns.
 * @author Jeff Jackowski
 */
void AATreeDestroy(AANode *tree);

/** @} */

