/**
 * @file audio.h
 * Interface to audio output.
 * @ingroup Audio
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "bool.h"

#ifdef WIN32
#undef PlaySound
#define PlaySound _PlaySound_tank_
#endif

/**
 * @defgroup Audio Audio
 * Handles audio output for the game through a high level interface.
 *
 * The sound setup is intentionally simple because I haven't found much
 * motiviation for this part of the project, but the game does need sound.
 *
 * SDL_mixer is used for the sound output. Several of SDL_mixer's channels are
 * used to easily support several independent sounds that can play
 * simultaneously. Sound volume drops linearly with distance such that a tank
 * 212 units distant (just off-screen at initial window size) produces sound at
 * half the volumne of something 0 units away.
 *
 * The audio functions are not critical to gameplay, so if InitAudio() fails,
 * the game will continue. All other audio functions will correctly do nothing
 * if called when audio is not initialized.
 *
 * @{
 */

/**
 * The available sounds.
 */
enum SND_SOUND {
	/**
	 * The sound of a tank's engine idling.
	 */
	SND_IDLE,
	/**
	 * The sound of a tank moving forward.
	 */
	SND_MOVE,
	/**
	 * The sound of a tank shooting its cannon.
	 */
	SND_SHOOT,
	/**
	 * The sound of a tank being destroyed.
	 */
	SND_DEATH,
	SND_MAX
};

/**
 * The audio channels that are mixed together to produce the final output.
 */
enum SND_CHANNEL {
	/**
	 * This channel plays the local player's shooting sound.
	 */
	SND_CH_LOCALSHOOT,
	/**
	 * A channel for the movement sound of tanks in the game. The local player
	 * is moving, the sound will be SND_SOUND::SND_MOVE at full volumne. If
	 * not, and another tank close by is moving, the sound will be
	 * SND_SOUND::SND_MOVE attenuated according to distance. Otherwise, the
	 * sound will be SND_SOUND::SND_IDLE.
	 */
	SND_CH_TANK,
	/**
	 * The number of reserved channels.
	 */
	SND_CH_RESERVE,
	/**
	 * The total number of mixing channels. The additional channels used beyond
	 * the reservered channels play the sounds of shots and kills of tanks other
	 * than the local player.
	 */
	SND_CH_MAX = 8
};

/**
 * The maximum squared distance that can be heard by the player. The value was
 * selected from what seemed to work well during trials.
 */
#define MAX_SQ_DIST  (27500<<2)  // 178374

/**
 * Prepares for auido output by requesting output resources and loading in
 * the game sounds.
 * @return  True on success, false on failure. Failure can be caused by lack
 *          of available sound hardware or inability to load any sounds.
 *          Failure to load some but not all sounds is not considered an error.
 */          
Bool InitAudio();

/**
 * Plays a sound; intended to play sounds caused by actions taken by the local
 * player.
 * @param snd    The sound from the ::SND_SOUND enum to play.
 * @param chan   The channel on which to play the sound, from the ::SND_CHANNEL
 *               enum. A specific channel must be specified; -1 is not accepted.
 * @param count  The number of times to repeat the sound, or -1 to loop forever.
 * @return  True on success, false on failure. Failure can be caused by lack of
 *          availble channels or an inability to load the requested sound (error
 *          during InitAudio()).
 * @author  Jeff Jackowski
 */
Bool PlaySound(int snd, int chan, int count);

/**
 * Plays a sound with its volume attenuated according to its distance from the
 * local player. This should not be used for sounds corresponding to the local
 * player.
 * @param snd    The sound from the ::SND_SOUND enum to play.
 * @param chan   The channel on which to play the sound, from the ::SND_CHANNEL
 *               enum, or -1 to select a free channel.
 * @param distsq The square of the distance from the player's tank to the
 *               source of the sound.
 * @param count  The number of times to repeat the sound, or -1 to loop forever.
 *               If -1, then the specific channel must be specified; @a chan
 *               must not be -1.
 * @return  True on success, false on failure. Failure can be caused by lack of
 *          availble channels or an inability to load the requested sound (error
 *          during InitAudio()).
 * @author  Jeff Jackowski
 */
Bool PlaySoundDist(int snd, int chan, int distsq, int count);

/**
 * Plays the sound of a tank, other than the local player's tank, moving.
 * @param distsq  The square of the distance from the player's tank to the
 *                source of the sound.
 * @return  True on success, false on failure.
 * @author  Jeff Jackowski
 */
#define PlayTankMove(distsq)   PlaySoundDist(SND_MOVE, SND_CH_TANK, distsq, -1)

/**
 * Plays the sound of the local player's tank moving.
 * @return  True on success, false on failure.
 * @author  Jeff Jackowski
 */
#define PlayLocalTankMove()   PlaySound(SND_MOVE, SND_CH_TANK, -1)

/**
 * Plays the sound of the local player's tank idling.
 * @return  True on success, false on failure.
 * @author  Jeff Jackowski
 */
#define PlayTankIdle()   PlaySound(SND_IDLE, SND_CH_TANK, -1)

/**
 * Plays the sound of a tank, other than the local player's tank, shooting.
 * @param distsq  The square of the distance from the player's tank to the
 *                source of the sound.
 * @return  True on success, false on failure.
 * @author  Jeff Jackowski
 */
#define PlayTankShoot(distsq)  PlaySoundDist(SND_SHOOT, -1, distsq, 0)

/**
 * Plays the sound of the destruction of a tank other than the local player's
 * tank.
 * @param distsq  The square of the distance from the player's tank to the
 *                source of the sound.
 * @return  True on success, false on failure.
 * @author  Jeff Jackowski
 */
#define PlayTankKilled(distsq) PlaySoundDist(SND_DEATH, -1, distsq, 0)

/**
 * Computes the square of the distance between two players.
 * @param  p1  A pointer to a Player.
 * @param  p2  A pointer to another Player.
 * @return An int with the square of the distance.
 * @author  Jeff Jackowski
 */
#define DistSq(p1, p2) \
	((int)(((p1)->x - (p2)->x) >> 8) * (int)(((p1)->x - (p2)->x) >> 8) +  \
	 (int)(((p1)->y - (p2)->y) >> 8) * (int)(((p1)->y - (p2)->y) >> 8))

/**
 * Stops all sounds and plays the sound of the local player's tank being
 * destroyed.
 * @return  True on success, false on failure. Failure can be caused by lack of
 *          availble channels or an inability to load the requested sound (error
 *          during InitAudio()).
 * @author  Jeff Jackowski
 */
Bool PlayLocalDeath();

/**
 * Pauses the sound on a specific mixing channel. Intended to pause the sound
 * of tanks moving in the SND_CHANNEL::SND_CH_TANK channel.
 * @author  Jeff Jackowski
 */
void PauseSound(int chan);

/**
 * Halts all playing sounds on all channels.
 * @author  Jeff Jackowski
 */
void StopAllSounds();

/**
 * Stops all sounds and frees all resources needed for audio output.
 * @author  Jeff Jackowski
 */
void UninitAudio();

/** @} */
