/**
 * @file bflderrstate.c
 * Implements the operating state for the battlefield error screen
 * (STATE_CODES::STATE_BFLDERR).
 * @ingroup StateBfldErr
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include <SDL_image.h>
#include <string.h>
#include <stdlib.h>
#include "states.h"
#include "color.h"
#include "gameconfig.h"
#include "tanksprite.h"
#include "message.h"

extern RenderItem background;

/**
 * The time when the splash screen was initalized.
 */
extern Uint32 startTime;

/**
 * A couple of Players used to draw tanks on the screen.
 */
extern Player *tanks;

/**
 * Places the text used on the splash screen. The positioning adjusts with the
 * size of the window or full-screen display.
 * @ingroup StateSplash
 * @author Jeff Jackowski
 */
static void BfldErrResize() {
	int loop;
	// assure good color values
	SetupColorVals();	
	// put the backgound image in place
	background.loc.w = dInfo.pw;
	background.loc.h = dInfo.ph;
	PlaceRenderItem(&background);
	// title at the top
	textItems[TEXTITEM_TITLE].loc.y = TOP_MARGIN;
	// continue at the bottom
	textItems[TEXTITEM_ERRCONT].loc.y = dInfo.ph - (BOTTOM_MARGIN << 1) -
	textItems[TEXTITEM_ERRCONT].loc.h;
	// place the tank
	tanks->tank.loc.w = tanks->tank.loc.h = dInfo.luw << 3;
	tanks->tank.loc.x = (dInfo.pw >> 1) - (dInfo.luw << 2);
	tanks->tank.loc.y = textItems[TEXTITEM_TITLE].loc.y +
	textItems[TEXTITEM_TITLE].loc.h + (TOP_MARGIN << 1);
	PlaceRenderItem(&(tanks->tank));
	// re-render the error text
	TextRenderString(&textItems[TEXTITEM_ERROR], FONT_SCORE, dInfo.pw - 40,
	colorFmtInd[COLOR_TEXT], bfErr.msg);
	// place the error a little below the tank
	textItems[TEXTITEM_ERROR].loc.y = tanks->tank.loc.y + tanks->tank.loc.h +
	((dInfo.ph - tanks->tank.loc.y - tanks->tank.loc.h -
	textItems[TEXTITEM_ERROR].loc.h) >> 1);
	// center all text horizontally
	for (loop = TEXTITEM_TITLE; loop < TEXTITEM_ERRCONT + 1; loop++) {
		// compute the centered location
		textItems[loop].loc.x = (dInfo.pw >> 1) - (textItems[loop].loc.w >> 1);
		// add the text
		if (textItems[loop].parent == NULL) {
			AddRenderItem(&textItems[loop], LAYER_TEXT);
		}
		// place the item
		PlaceRenderItem(&(textItems[loop]));
	}
}

int BfldErrInit() {
	memset(&background, 0, sizeof(background));
	// get memory for a tank to use as an icon
	if ((tanks = calloc(1, sizeof(Player))) == NULL) {
		// out of memory
		return STATE_ERROR;
	}
	// setup position of screen parts
	dInfo.luw = dInfo.luh = 8;
	dInfo.lw = dInfo.pw / dInfo.luw;
	dInfo.lh = dInfo.ph / dInfo.luh;
	// put the backgound image in place
	background.renderer = SolidFillRenderer;
	background.data = &(colorVals[COLOR_REGBACK]);
	AddRenderItem(&background, LAYER_BACKGROUND);
	// initalize a tank
	tanks->tank.renderer = DrawTank;
	AddRenderItem(&(tanks->tank), LAYER_TANKS);
	// set the text color
	colorFmtInd[COLOR_TEXT] = colorFmtInd[COLOR_BLACK];
	// attempt to render the title and prompt text off-screen
	if (!TextRenderString(&textItems[TEXTITEM_TITLE], FONT_MESSAGE, 0,
	colorFmtInd[COLOR_TEXT], strings[STRING_TITLE]) ||
	!TextRenderString(&textItems[TEXTITEM_ERRCONT], FONT_SCORE, 0,
	colorFmtInd[COLOR_TEXT], strings[STRING_FIRECONT])) {
		// couldn't render the text
		free(tanks);
		return STATE_ERROR;
	}
	// place all the text
	BfldErrResize();
	// record the time
	startTime = SDL_GetTicks();
	return STATE_OK;
}

int BfldErrUninit() {
	// see if a background image, rather than a color, was used
	if (background.renderer == GenericSurfaceRenderer) {
		// remove the background image
		SDL_FreeSurface(background.data);
	}
	RemoveRenderItem(&background);
	RemoveRenderItem(&(tanks->tank));
	free(tanks);
	TextUnsetState();
	return STATE_OK;
}

int BfldErrRun() {
	SDL_Event event;
	
	// loop through all waiting events
	while (SDL_PollEvent(&event)) {
		// figure event type
		switch (event.type) {
			case SDL_QUIT:
				return STATE_QUIT;
				break;
			case SDL_KEYDOWN:
				// check for a request to quit the program
				if (event.key.keysym.sym == SDLK_ESCAPE)
					return STATE_QUIT;
				break;
			case SDL_KEYUP:
			case SDL_JOYBUTTONDOWN:
				// enforce a minimum delay
				if ((SDL_GetTicks() - startTime) > 511)
					return STATE_MENUROOT;
				break;
			case SDL_VIDEORESIZE:
				ResizeWindow(event.resize.w, event.resize.h);
				BfldErrResize();
			default:
				break;
		}
	}
	// rotate the tank
	tanks->rot += 128;
	// make the new rotaion visible
	PlaceRenderItem(&(tanks->tank));
	
	// draw it
	//Render();
	// don't monoplize the processor
	//SDL_Delay(16);
	// ----- run network code ----
	// Without this, if the client quickly connects to a server it will not
	// respond until running the menu state, which can be long enough for the
	// server to decide that the client is not responding.
	ProcessNetworkMessages();
	// continue with the error screen
	return STATE_BFLDERR;
}

