/**
 * @file client.h
 * Definitions for server side data detailing the clients.
 * @ingroup Networking
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#ifndef _client_h_
#define _client_h_

#include "hostlookup.h"
#include "player.h"
#include <SDL_types.h>

typedef struct GameConfigId_t GameConfigId;
typedef struct ClientData_t   ClientData;

/**
 * Stores information on a client for use on the server.
 * @ingroup Networking
 * @author Jeff Jackowski
 */
struct ClientData_t {
	/**
	 * A pointer to the next client that needs some part of the
	 * @ref NetGConf "game configuration".
	 */
	ClientData *configNext;
	/**
	 * A pointer to the previous client that needs some part of the
	 * @ref NetGConf "game configuration".
	 */
	ClientData *configPrev;
	/**
	 * The time of the transmitted update for each of the players.
	 * @todo  Reevaluate the need for this field.
	 */
	UpdateTimes sentUpdateTimes[MAX_PLAYERS];
	/**
	 * The time when the client attempted to disconnect. This is used to
	 * support reliable disconnects by maintaing client data for up to 4
	 * seconds after the disconnect message. Only valid when
	 * the ClientData::disconn flag is set.
	 */
	Uint32  disconnTime;
	/**
	 * The socket address used to communicate directly with the client.
	 */
	SocketAddr addr;
	/**
	 * The random ID sent from the client during the connection process.
	 */
	Uint32 randId;
	/**
	 * The index of the first configuration message sent to this client.
	 */
	Uint16 configFirst;
	/**
	 * The number of game configuration messages that this client has
	 * acknowledged.
	 */
	Uint16 configAckCnt;
	/**
	 * The index of the client. The client data is stored in an array and this
	 * value is the index into the array. It is needed when only a pointer to
	 * the client is available, such as when the game configuration messages
	 * are sent and a linked list of clients (see configNext and configPrev)
	 * is traversed.
	 */
	Uint8 cid;
	union {
		/**
		 * Flags that modify how this client is handled. The initial value
		 * is zero and is reset when clients disconnect.
		 */
		Uint8 flags;
		struct {
			/**
			 * Indicates that the client has contacted the server and that
			 * a connection is being established.
			 */
			Bool contact : 1;
			/**
			 * Indicates the client has joined the server's game.
			 */
			Bool active  : 1;
			/**
			 * Indicates the client is in the process of disconnecting.
			 */
			Bool disconn : 1;
			/**
			 * Indicates the client can be contacted with IPv4 UDP broadcasts.
			 */
			Bool bcast   : 1;
			/**
			 * Indicates the client can be contacted with IPv6 UDP multicasts.
			 * @warning  This is not yet implemented.
			 */
			Bool mcast   : 1;
			/**
			 * Indicates that the client has acknowledged the game options.
			 */
			Bool hasGameOpts    : 1;
		};
	};
};

/**
 * An array of data on each client indexed by player ID. This array is only
 * allocated when the process is acting as a server, and all bytes are
 * initialized to zero. When not a server, the pointer should be NULL.
 * Defined in net.c.
 * @ingroup Networking
 */
extern ClientData *clients;

/**
 * The random ID generated by this process to uniquely identify it as a client
 * before it has been assigned a client ID by the server. Defined in net.c.
 */
extern Uint32 clientRandId;

#endif

