/**
 * @file game.c
 * Game specific code.
 * @ingroup GameImpl
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#define QueuePlaceTank   __func_QueuePlaceTank
#define QueuePlaceShell  __func_QueuePlaceShell
#define QueuePlaceName   __func_QueuePlaceName

#include "game.h"
#include "gameconfig.h"
#include "color.h"
#include "net.h"
#include "obstacle.h"
#include "tanksprite.h"
#include "playernet.h"
#include "random.h"
#include "notice.h"
#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#ifdef USE_AUDIO
#include "audio.h"
#endif

#undef QueuePlaceTank
#undef QueuePlaceShell
#undef QueuePlaceName

#ifdef _MSC_VER
// For some reason it seems Microsoft's math.h in Visual Studio 2003 only has
// prototypes for the fabsf and sqrtf functions for C++. Or maybe I did
// something wrong. I can't imagine this wouldn't be a major problem. The
// following macros are a workaround. The problem doesn't exist in Visual
// Studio 2005.
#ifndef fabsf
#define fabsf(x)  (float)fabs(x)
#endif
#ifndef sqrtf
#define sqrtf(x)  (float)sqrt(x)
#endif
#endif

/**
 * A table containing the x and y movement deltas based on the rotation value.
 * The first dimension is the @ref Player::rot "rotation value", the second
 * contains the x and y offsets.
 * @ingroup GameImpl
 */
static const Sint32 motionTable[16][2] = {
	// up
	{ 0, -rounddiv(MOTION_0_1, MOTION_DIV) },
	// 22.5 right of up
	{ rounddiv(MOTION_22_0, MOTION_DIV), -rounddiv(MOTION_22_1, MOTION_DIV) },
	// 45 right and up
	{ rounddiv(MOTION_45_0, MOTION_DIV), -rounddiv(MOTION_45_1, MOTION_DIV) },
	// 22.5 up of right
	{ rounddiv(MOTION_22_1, MOTION_DIV), -rounddiv(MOTION_22_0, MOTION_DIV) },
	// right
	{ rounddiv(MOTION_0_1, MOTION_DIV), 0 },
	// 22.5 down of right
	{ rounddiv(MOTION_22_1, MOTION_DIV), rounddiv(MOTION_22_0, MOTION_DIV) },
	// 45 right and down
	{ rounddiv(MOTION_45_0, MOTION_DIV), rounddiv(MOTION_45_1, MOTION_DIV) },
	// 22.5 right of down
	{ rounddiv(MOTION_22_0, MOTION_DIV), rounddiv(MOTION_22_1, MOTION_DIV) },
	// down
	{ 0, rounddiv(MOTION_0_1, MOTION_DIV) },
	// 22.5 left of down
	{ -rounddiv(MOTION_22_0, MOTION_DIV), rounddiv(MOTION_22_1, MOTION_DIV) },
	// 45 left and down
	{ -rounddiv(MOTION_45_0, MOTION_DIV), rounddiv(MOTION_45_1, MOTION_DIV) },
	// 22.5 down of left
	{ -rounddiv(MOTION_22_1, MOTION_DIV), rounddiv(MOTION_22_0, MOTION_DIV) },
	// left
	{ -rounddiv(MOTION_0_1, MOTION_DIV), 0 },
	// 22.5 up of left
	{ -rounddiv(MOTION_22_1, MOTION_DIV), -rounddiv(MOTION_22_0, MOTION_DIV) },
	// 45 left and up
	{ -rounddiv(MOTION_45_0, MOTION_DIV), -rounddiv(MOTION_45_1, MOTION_DIV) },
	// 22.5 left of up
	{ -rounddiv(MOTION_22_0, MOTION_DIV), -rounddiv(MOTION_22_1, MOTION_DIV) }
};

#ifdef DOXYGEN
/**
 * A table containing normal vectors based on the rotation value.
 * The first dimension is the @ref Player::rot "rotation value", the second
 * contains the x and y offsets.
 * @ingroup GameImpl
 */
static const float normalTable[16][2] = {
	// up
	{ 0, -FMOTION_0_1 },
	// 22.5 right of up
	{ FMOTION_22_0, -FMOTION_22_1 },
	// 45 right and up
	{ FMOTION_45_0, -FMOTION_45_1 },
	// 22.5 up of right
	{ FMOTION_22_1, -FMOTION_22_0 },
	// right
	{ FMOTION_0_1, 0 },
	// 22.5 down of right
	{ FMOTION_22_1, FMOTION_22_0 },
	// 45 right and down
	{ FMOTION_45_0, FMOTION_45_1 },
	// 22.5 right of down
	{ FMOTION_22_0, FMOTION_22_1 },
	// down
	{ 0, FMOTION_0_1 },
	// 22.5 left of down
	{ -FMOTION_22_0, FMOTION_22_1 },
	// 45 left and down
	{ -FMOTION_45_0, FMOTION_45_1 },
	// 22.5 down of left
	{ -FMOTION_22_1, FMOTION_22_0 },
	// left
	{ -FMOTION_0_1, 0 },
	// 22.5 up of left
	{ -FMOTION_22_1, -FMOTION_22_0 },
	// 45 left and up
	{ -FMOTION_45_0, -FMOTION_45_1 },
	// 22.5 left of up
	{ -FMOTION_22_0, -FMOTION_22_1 }
};
#endif

#if defined(USE_AUDIO) || defined(DOXYGEN)
/**
 * The last computed closest distance from the player's tank to any other
 * moving tank. Used to attenuate the volumne of the sound of other moving
 * tanks.
 */
static unsigned int oldClosest = 0xFFFFFFFF;
#endif

void PlacePlayerLabel(Player *p) {
	if (p->labelName.renderer == NULL) return;
	// see if the whole frame will be rendered
	if (RenderWillRenderAll()) {
		// set the required physical location fields that are not set when
		// the whole frame is rendered
		p->tank.physDest.x = p->tank.loc.x + p->tank.parent->offX;
		p->tank.physDest.y = p->tank.loc.y + p->tank.parent->offY;
		p->tank.physDest.w = p->tank.loc.w;
		p->tank.physDest.h = p->tank.loc.h;
		// clip to the display region
		ClipRectToWH(&(p->tank.physDest), dInfo.pw, dInfo.ph);
	}
	// see if the tank is visible enough
	if ((p->tank.physDest.h >= (dInfo.luh << 2)) &&
	(p->tank.physDest.w >= (dInfo.luw << 2))) {
		// horizontally place the name centered with the tank
		p->labelName.loc.x = p->tank.loc.x + (dInfo.luw << 2) -
		(p->labelName.loc.w >> 1);
		// check for adequate room below the tank
		if (((dInfo.luw << 3) + p->tank.physLocY +
		p->labelName.loc.h) < dInfo.ph) {
			// vertically place beneth the tank
			p->labelName.loc.y = (dInfo.luw << 3) + p->tank.loc.y;
		}
		else {
			// vertically place above the tank
			p->labelName.loc.y = p->tank.loc.y - p->labelName.loc.h;
		}
		// place the label
		if (dInfo.showLabels)
			PlaceRenderItem(&(p->labelName));
	}
	// check for a visible name
	else if (p->labelName.enableDraw && dInfo.showLabels) {
		// make it invisible
		ClearRenderItem(&(p->labelName));
	}
}

void EnablePlayerLabels() {
	Player *p;
	int loop;

	// place all player's labels
	for (p = players, loop = MAX_PLAYERS; loop > 0; loop--, p++) {
		if (((p->flags & (PLAYER_ACTIVE | PLAYER_DEAD)) == PLAYER_ACTIVE) &&
		!p->labelName.enableDraw && p->labelName.renderer) {
			PlaceRenderItem(&(p->labelName));
		}
	}
	// set the flag
	dInfo.showLabels = TRUE;
}

void DisablePlayerLabels() {
	Player *p;
	int loop;

	// clear all player's labels
	for (p = players, loop = MAX_PLAYERS; loop > 0; loop--, p++) {
		if ((p->flags & PLAYER_ACTIVE) && p->labelName.enableDraw) {
			ClearRenderItem(&(p->labelName));
		}
	}
	// clear the flag
	dInfo.showLabels = FALSE;
}

/**
 * A set of bit flags indicating which player's tanks need to be placed by the
 * renderer.
 * @internal
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
static Uint32 batchTankDraw[rounddivup(MAX_PLAYERS, 32)] = { 0 };

/**
 * A set of bit flags indicating which player's shells need to be placed by the
 * renderer.
 * @internal
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
static Uint32 batchShellDraw[rounddivup(MAX_PLAYERS, 32)] = { 0 };

/**
 * A set of bit flags indicating which player's name labels need to be placed
 * by the renderer.
 * @internal
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
static Uint32 batchNameDraw[rounddivup(MAX_PLAYERS, 32)] = { 0 };

/**
 * Computes the index of a player's flag within the batch flags.
 * @internal
 * @code
 * batchTankDraw[BatchIndex(p->pid)]; // the correct element with the flag
 *                                    // for player p
 * @endcode
 * @param   pid  The player's ID.
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
#define BatchIndex(pid)  ((pid) >> 5)

/**
 * Computes the bit of a player's flag at the corresponding index of the
 * batch flags.
 * @internal
 * @code
 * if (batchTankDraw[BatchIndex(p->pid)] & BatchBit(p->pid)) {
 *    // the flag for player p is set
 * }
 * @endcode
 * @param   pid  The player's ID.
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
#define BatchBit(pid)    (1 << ((pid) & 0x1F))

/**
 * Sets the flag that indicates the given player's tank needs to be redrawn.
 * @internal
 * @param   pid  The player's ID.
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
#define QueuePlaceTank(pid) \
	batchTankDraw[BatchIndex(pid)] |= BatchBit(pid)

/**
 * Sets the flag that indicates the given player's shell needs to be redrawn.
 * @internal
 * @param   pid  The player's ID.
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
#define QueuePlaceShell(pid) \
	batchShellDraw[BatchIndex(pid)] |= BatchBit(pid)

/**
 * Sets the flag that indicates the given player's name label needs to be
 * redrawn.
 * @internal
 * @param   pid  The player's ID.
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
#define QueuePlaceName(pid) \
	batchNameDraw[BatchIndex(pid)] |= BatchBit(pid)

/**
 * The function prototype used to place tanks, shells, and names whenever they
 * change or move.
 * @internal
 * @param   data  A pointer to some data. The type varies with the function.
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
typedef void (*PlacementFunc)(void *data);

/**
 * Renders the player's items who's flags are set using the given function.
 * @internal
 * @param batch  A pointer to the flags to inspect. All flags will be cleared
 *               by this function.
 * @param placement  The function to call that will cause the item to be
 *                   rendered. It will be passed a pointer to an element within
 *                   the corresponding Player struct.
 * @param offset  The offset into the Player struct of the item to pass to the
 *                @a placement function.
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
static void BatchRender(Uint32 *batch, PlacementFunc placement, long offset) {
	Player *p;
	int index, bit;
	Uint32 byte;
	// loop through all the batch flags
	for (index = rounddivup(MAX_PLAYERS, 32), p = players; index > 0;
	batch++, index--) {
		// see if any of these 32 items need to be placed
		if (*batch) {
			// at least one shell needs placement, so consider them 8 at a time
			for (bit = 1, byte = 0xFF; byte; byte <<= 8) {
				// see if any of these 8 items need to be placed
				if (*batch & byte) {
					// consider these 8 items one at a time
					for (; bit & byte; p++, bit <<= 1) {
						// place this item?
						if (*batch & bit) {
							(*placement)(((char*)p) + offset);
						}
					}
				}
				else {
					// none of these 8 items need placement; advance past them
					p += 8;
					bit <<= 8;
				}
			}
		}
		else {
			// none of these 32 items need placement; advance past them
			p += 32;
		}
		// clear the batch flags
		*batch = 0;
	}
}

void RenderGame() {
	// batch render tanks
	BatchRender(batchTankDraw, (PlacementFunc)PlaceRenderItem,
	offsetof(Player, tank));
	// batch render shells
	BatchRender(batchShellDraw, (PlacementFunc)PlaceRenderItem,
	offsetof(Player, shell));
	// if player labes are currently visible . . .
	if (dInfo.showLabels) {
		// batch render tanks
		BatchRender(batchNameDraw, (PlacementFunc)PlacePlayerLabel, 0);
	}
}

void MoveTank(Player *p, int factor) {
	int tempLoc;
	// forward motion
	if (factor > 0) {
		// move the tank
		p->x += motionTable[(p->rot >> 8) & 0xF][0] * factor;
		p->y += motionTable[(p->rot >> 8) & 0xF][1] * factor;
	}
	// backward motion
	else if (factor < 0) {
		p->x += (motionTable[(p->rot >> 8) & 0xF][0] * factor) >> 1;
		p->y += (motionTable[(p->rot >> 8) & 0xF][1] * factor) >> 1;
	}
	// check for out-of-bounds tank and correct
	if (gameOpts.width > 0) {
		if (p->x < 0)
			p->x = 0;
		else if (p->x > ((gameOpts.width - 8) << 8))
			p->x = (gameOpts.width - 8) << 8;
		if (p->y < 0)
			p->y = 0;
		else if (p->y > ((gameOpts.height - 8) << 8))
			p->y = (gameOpts.height - 8) << 8;
	}
	// see if this tank needs to be drawn
	if ((tempLoc = (p->y >> 8) * dInfo.luh) != p->tank.loc.y) {
		// place the item accordingly
		p->tank.loc.y = tempLoc;
		p->tank.loc.x = (p->x >> 8) * dInfo.luw;
		// flag the tank and the player name label for a redraw
		QueuePlaceTank(p->pid);
		// put the label in place
		QueuePlaceName(p->pid);
	}
	else if ((tempLoc = (p->x >> 8) * dInfo.luw) != p->tank.loc.x) {
		// place the item accordingly
		p->tank.loc.x = tempLoc;
		QueuePlaceTank(p->pid);
		// put the label in place
		QueuePlaceName(p->pid);
	}
}

void MoveShell(Player *p) {
	// record distance traveled and check for an uninitalized shell
	if (p->shellDist++ == 0) {
		// place the shell on the tank
		p->shellX = p->x + (shellInitLoc[(p->rot >> 8) & 0xF][0] << 8);
		p->shellY = p->y + (shellInitLoc[(p->rot >> 8) & 0xF][1] << 8);
		// set the shell's rotation to match the tank's
		p->shellRot = p->rot;
		goto MoveShell_placeShell;
	}
	// check for maximum distance
	else if (p->flags & PLAYER_PONG) { // test the pong flag once
		if (p->shellDist > SHELL_PONG_DIST) {
			MoveShell_killShell:
			// quit rendering the shell
			batchShellDraw[BatchIndex(p->pid)] &= ~BatchBit(p->pid);
			ClearRenderItem(&(p->shell));
			// flag the shot as no longer in flight
			p->flags &= ~PLAYER_SHOT;
			// don't move the shell
			return;
		}
		// here is were the bouncing of the shell can be done
	}
	else if (p->shellDist > SHELL_GUIDED_DIST) {
		goto MoveShell_killShell;
	}
	// forward motion
	p->shellX += motionTable[(p->shellRot >> 8) & 0xF][0] * MOTION_SHELL_FACTOR;
	p->shellY += motionTable[(p->shellRot >> 8) & 0xF][1] * MOTION_SHELL_FACTOR;
	// kill out-of-bounds shells
	if ((p->shellX < 0) || (p->shellY < 0) ||
	((p->shellX >> 8) > gameOpts.width) || ((p->shellY >> 8) > gameOpts.height))
		goto MoveShell_killShell;
	// see if the shell has moved a visible amount
	if ((p->shell.loc.x != (p->shellX >> 8)) ||
	(p->shell.loc.y != (p->shellY >> 8))) {
		MoveShell_placeShell:
		// setup the shell's location
		p->shell.loc.x = (p->shellX >> 8) * dInfo.luh;
		p->shell.loc.y = (p->shellY >> 8) * dInfo.luh;
		// place it onscreen
		QueuePlaceShell(p->pid);
	}
}

/**
 * Roates a player's tank. Sets the player's dirty flag if the visible
 * rotation changes.
 * @warning  Must be called after MoveTank().
 * @param p   A pointer to the player to rotate.
 * @param rot  A delta that will be applied to the player's rotation.
 * @ingroup GameImpl
 * @author  Jeff Jackowski
 */
static void RotateTank(Player *p, Sint16 rot) {
	Sint16 oldRot;
	// keep the old rotation
	oldRot = p->rot;
	// compute the new rotation
	p->rot += rot;
	// check for a change and a need for a redraw
	if (!p->tank.placed && ((oldRot & 0xF00) != (p->rot & 0xF00))) {
		// record that the whole tank must be redrawn
		QueuePlaceTank(p->pid);
		//PlaceRenderItem(&(p->tank));
	}
	// see if a non-pong shell is in use
	if ((p->flags & PLAYER_SHOT) && !(p->flags & PLAYER_PONG)) {
		// adjust the shell's rotation to match
		p->shellRot = p->rot;
	}
}

void SpawnTank(Player *p) {
	Player *c;
	SDL_Rect *rect;
	int attempts = 8, maxArea = gameOpts.spawnAreas[p->team], loop;
	Uint32 rndnum;
	Bool bad;
	// this function is only for servers
	assert(networkState == NETSTATE_SERVER);
	// up to some maximum tries . . .
	do {
		// pick a spawn area at random
		rndnum = RandNum();
		p->rot = rndnum >> 16;
		rect = &(spawnAreas[p->team][rndnum % maxArea]);
		// pick a spot in the area at random
		rndnum = RandNum();
		p->x = (p->tank.loc.x = (rect->x + (rndnum & 0xFFFF) % rect->w)) << 8;
		p->y = (p->tank.loc.y = (rect->y + (rndnum >> 16) % rect->h)) << 8;
		p->tank.loc.x *= dInfo.luw;
		p->tank.loc.y *= dInfo.luh;
		// might be good
		bad = FALSE;
		// max tries?
		if (--attempts == 0) {
			// use this attempt no matter what
			break;
		}
		// check for a collision with other tanks
		for (c = players, loop = MAX_PLAYERS; loop > 0; loop--, c++) {
			// is this tank not collidable?
			if ((c == p) || !(c->flags & PLAYER_ACTIVE) ||
			(c->flags & PLAYER_DEAD)) {
				// c is p, or c isn't in the game, or c is dead
				continue;
			}
			// check for collision
			if (IsCollidingTank(p, c)) {
				// not good
				bad = TRUE;
				break;
			}
		}
	} while (bad);
	// finalize tank's position
	QueuePlaceTank(p->pid);
	QueuePlaceName(p->pid);
	// make the tank not dead
	p->flags = (p->flags & ~PLAYER_DEAD) | PLAYER_UPDATE_TANK |
	PLAYER_UPDATE_DFS;
	p->deadTime = 0;
}

void HandleInput(Uint8 *eventState) {
	// get a reference to the local player
	Player *p = &(players[localPlayer]);
	// do a limited set for dead players
	if (!(p->flags & PLAYER_DEAD)) {
		// new request to move forward
		if (eventState[EC_FORWARD] == 1) {
			// flag the request as handled
			eventState[EC_FORWARD]++;
			// set the forward motion flag and request the player's data be sent
			// over the network
			p->flags |= PLAYER_FORWARD | PLAYER_UPDATE_TANK;
			#ifdef USE_AUDIO
			// play the movement sound
			oldClosest = 0;
			//PlaySound(SND_MOVE, SND_CH_LOCALTANK, -1);
			#endif
		}
		// new request to stop forward motion
		else if ((eventState[EC_FORWARD] == 0) && (p->flags & PLAYER_FORWARD)) {
			// clear the forward motion flag and request the player's data be
			// sent over the network
			p->flags = (p->flags & ~PLAYER_FORWARD) | PLAYER_UPDATE_TANK;
			#ifdef USE_AUDIO
			if (!(p->flags & PLAYER_BACKWARD)) {
				// play the idle engine sound
				oldClosest = 0;
				//PlaySound(SND_IDLE, SND_CH_LOCALTANK, -1);
			}
			#endif
		}
		// new request to move backward
		if (eventState[EC_BACKWARD] == 1) {
			// flag the request as handled
			eventState[EC_BACKWARD]++;
			// set the backward motion flag and request the player's data be
			// sent over the network
			p->flags |= PLAYER_BACKWARD | PLAYER_UPDATE_TANK;
			#ifdef USE_AUDIO
			// play the movement sound
			oldClosest = 0;
			//PlaySound(SND_MOVE, SND_CH_LOCALTANK, -1);
			#endif
		}
		else if ((eventState[EC_BACKWARD] == 0) && (p->flags & PLAYER_BACKWARD)) {
			// clear the backward motion flag and request the player's data be
			// sent over the network
			p->flags = (p->flags & ~PLAYER_BACKWARD) | PLAYER_UPDATE_TANK;
			#ifdef USE_AUDIO
			if (!(p->flags & PLAYER_FORWARD)) {
				// play the idle engine sound
				oldClosest = 0;
				//PlaySound(SND_IDLE, SND_CH_LOCALTANK, -1);
			}
			#endif
		}
		if (eventState[EC_ROTRIGHT] == 1) {
			eventState[EC_ROTRIGHT]++;
			p->flags |= PLAYER_STARTROTRIGHT | PLAYER_UPDATE_TANK;
		}
		else if ((eventState[EC_ROTRIGHT] == 0) && p->flags & PLAYER_ROTRIGHT) {
			p->flags = (p->flags & ~PLAYER_ROTRIGHT) | PLAYER_UPDATE_TANK;
		}
		if (eventState[EC_ROTLEFT] == 1) {
			eventState[EC_ROTLEFT]++;
			p->flags |= PLAYER_STARTROTLEFT | PLAYER_UPDATE_TANK;
		}
		else if ((eventState[EC_ROTLEFT] == 0) && p->flags & PLAYER_ROTLEFT) {
			p->flags = (p->flags & ~PLAYER_ROTLEFT) | PLAYER_UPDATE_TANK;
		}
	}
	// Check for request to fire while there is no shell in flight for the
	// local player. Unlike the other inputs, this one will be serviced each
	// time it changes the game state rather than when the user changes the
	// input.
	if (eventState[EC_FIRE] == 1) {
		// if the player has been dead long enough, this is a request to spawn
		if ((p->flags & PLAYER_DEAD) && (p->deadTime > DEAD_TIME)) {
			// must not be taken as an immediate request to fire next frame
			eventState[EC_FIRE]++;
			// for servers . . .
			if (networkState == NETSTATE_SERVER) {
				// spwan the player
				SpawnTank(p);
			}
			// for clients . . .
			else {
				// send a spawn request to the server
				SendSpawnRequest();
			}
			#ifdef USE_AUDIO
			// start the idle engine sound
			PlaySound(SND_IDLE, SND_CH_TANK, -1);
			#endif
		}
		// an input to shoot
		else if ((p->flags & (PLAYER_SHOT | PLAYER_DEAD)) == 0) {
			// set the flag that brings the shell into existance and
			// request a network update
			p->flags |= PLAYER_SHOT | PLAYER_UPDATE_SHELL;
			// initalize the shell
			p->shellDist = 0;
			// place the shell later to avoid rendering issues that might be
			// caused by the tank moving based on the input
			#ifdef USE_AUDIO
			// play the shooting sound
			PlaySound(SND_SHOOT, SND_CH_LOCALSHOOT, 0);
			#endif
		}
	}
}

/**
 * Attempts to correct a collision between this tank and another tank. Call
 * the function for each colliding tank/player.
 * @param p  A player who's tank is colliding with another tank.
 * @author   Jeff Jackowski
 */
static void FixTankCollision(Player *p) {
	// check for motion on tank p
	if (p->flags & (PLAYER_FORWARD | PLAYER_BACKWARD | PLAYER_ROTRIGHT |
	PLAYER_ROTLEFT)) {
		// move in the opposite direction
		if (p->flags & PLAYER_FORWARD) {
			MoveTank(p, -10);
		}
		else if (p->flags & PLAYER_BACKWARD) {
			MoveTank(p, 8);
		}
		// rotate in the opposite direction
		else if (p->flags & PLAYER_ROTRIGHT) {
			RotateTank(p, -64);
		}
		else if (p->flags & PLAYER_ROTLEFT) {
			RotateTank(p, 64);
		}
		// for the server . . .
		if (networkState == NETSTATE_SERVER) {
			// send out a network update for this tank
			p->flags |= PLAYER_UPDATE_TANK;
		}
	}
}

void UpdatePlayers(Uint32 time) {
	// Stores the updates to scores from this frame. Used to produce the
    // ::MSG_TYPE_SCORE network messages.
	ScoreUpdate scores[MAX_PLAYERS / 2];
	int numScores = 0, loop, inner;
	// number of players still out-of-date
	int outdated;
	#ifdef USE_AUDIO
	unsigned int dist, closest = 0xFFFFFFFF;
	Player *local = &(players[localPlayer]);
	#endif
	Player *p, *c;
	Obstacle *obst;

	// loop until all players are at least up-to-date
	do {

		/**
		 * @todo  Update players furthest out of date first
		 */

		// clear the number of players still needing an update
		outdated = 0;
		// loop to update player locations
		for (p = players, loop = MAX_PLAYERS; loop > 0; loop--, p++) {
			// if this player is inactive or is already current . . .
			if (!(p->flags & PLAYER_ACTIVE) || (p->localUpdate >= time))
				// skip this player
				continue;
			// increment the time
			p->localUpdate += FRAME_DURATION;
			// check for a dead player
			if (p->flags & PLAYER_DEAD) {
				// count the time spent dead and see if the tank should not stay
				// dead
				if (p->deadTime == DEAD_TIME) {
					// stop drawing the player
					batchTankDraw[BatchIndex(p->pid)] &= ~BatchBit(p->pid);
					ClearRenderItem(&(p->tank));
					ClearRenderItem(&(p->labelName));
					p->deadTime++;
				}
				else if (p->deadTime < DEAD_TIME) {
					// Explode the tank!  By explode, I mean spin really fast.
					RotateTank(p, 256);
					p->deadTime++;
				}
			}
			// for live players . . .
			else {
				// check for forward motion
				if (p->flags & PLAYER_FORWARD) {
					MoveTank(p, 1);
				}
				// check for backward motion
				if (p->flags & PLAYER_BACKWARD) {
					MoveTank(p, -1);
				}
				// check for the start of a right turn
				if (p->flags & PLAYER_STARTROTRIGHT) {
					// turn started; flag a continuing turn
					p->flags = (p->flags & ~PLAYER_STARTROTRIGHT) |
					PLAYER_ROTRIGHT;
					RotateTank(p, 64);
				}
				// check for a continuing right turn
				else if (p->flags & PLAYER_ROTRIGHT) {
					RotateTank(p, 16);
				}
				// check for the start of a left turn
				if (p->flags & PLAYER_STARTROTLEFT) {
					// turn started; flag a continuing turn
					p->flags = (p->flags & ~PLAYER_STARTROTLEFT) |
					PLAYER_ROTLEFT;
					RotateTank(p, -64);
				}
				// check for a continuing left turn
				else if (p->flags & PLAYER_ROTLEFT) {
					RotateTank(p, -16);
				}
			}
			// see if the player has a shell in flight
			if (p->flags & PLAYER_SHOT) {
				// move the shell
				MoveShell(p);
			}
		}
		// loop to detect and correct collisions
		for (p = players, loop = MAX_PLAYERS; loop > 0; loop--, p++) {
			// only check collisions on active players
			if (!(p->flags & PLAYER_ACTIVE) || ((p->flags & PLAYER_DEAD) &&
			(p->deadTime > DEAD_TIME))) {
				continue;
			}
			// check collisions for this player against others
			if (networkState == NETSTATE_SERVER) {
				c = players;
				inner = MAX_PLAYERS;
			}
			else {
				c = p + 1;
				inner = loop - 1;
			}
			for ( ; inner > 0; inner--, c++) {
				// see if player c is not active or dead
				if ((c == p) || !(c->flags & PLAYER_ACTIVE) ||
				((c->flags & PLAYER_DEAD) && (c->deadTime > DEAD_TIME))) {
					// do not collide with this player
					continue;
				}
				// if this code is for the server . . .
				if (networkState == NETSTATE_SERVER) {
					// Check for a shell collision. Only the server may decide a
					// player has been hit and count the score.
					if ((c->flags & PLAYER_SHOT) && IsCollidingShell(p, c)) {
						// c hit p!
						// get rid of the shell
						c->flags &= ~PLAYER_SHOT;
						ClearRenderItem(&(c->shell));
						batchShellDraw[BatchIndex(c->pid)] &= ~BatchBit(c->pid);
						// force an update to be sent
						c->flags |= PLAYER_UPDATE_DFS;
						// see if p was not dead
						if (!(p->flags & PLAYER_DEAD)) {
							// reset the time the player has spent dead
							p->deadTime = 0;
							// set the dead flag and force an update to be sent
							p->flags |= PLAYER_DEAD | PLAYER_UPDATE_DFS;
							// clear the movement flags
							p->flags &= ~(PLAYER_FORWARD | PLAYER_BACKWARD |
							PLAYER_ROTRIGHT | PLAYER_ROTLEFT | PLAYER_SHOT);
							// remove the dead player's shell
							ClearRenderItem(&(p->shell));
							batchShellDraw[BatchIndex(p->pid)] &=
							~BatchBit(p->pid);
							// check for a team match
							if (c->team == p->team) {
								// is p in good standing?
								if (p->kills >= 0) {
									// count this against c
									c->kills -= 4;
								}
								// else, p probably deserved it
							}
							else {
								// count the kill; check for non-CTF match
								if ((++c->kills > 0) &&
								!(gameOpts.flags & OPT_ISCTF)) {
									// count another kill for the team
									gameOpts.teamScore[c->team]++;
									// flag the need to update the score
									gameOpts.flags |= OPT_UPDATESCORE;
								}
							}
							// record the score and show a notice on-screen
							AddScoreNotice(scores[numScores].killerId = c->pid,
							scores[numScores].killedId = p->pid, c->kills);
							numScores++;
							#ifdef USE_AUDIO
							// is p the local (server) player?
							if (p == local) {
								// play the death sound
								PlayLocalDeath();
							}
							else {
								// compute the square of the distance
								int distsq = DistSq(p, local);
								// close enough to hear?
								if (distsq < (MAX_SQ_DIST >> 1)) {
									// play the death sound
									PlayTankKilled(distsq);
								}
							}
							#endif
						}
					}
					// finished with O(n*2) stuff; make rest of loop O(n log n)
					if (inner > loop) continue;
				}
				// check for a tank to tank collision
				if (IsCollidingTank(p, c)) {
					// fix the problem
					FixTankCollision(p);
					FixTankCollision(c);
				}
			}
			// iterate through the obstacles
			for (inner = gameOpts.obstacles, obst = obstacles; inner > 0;
			obst++, inner--) {
				// check for a collision between a moving tank and an obstacle
				if ((p->flags & (PLAYER_FORWARD | PLAYER_BACKWARD |
				PLAYER_ROTRIGHT | PLAYER_ROTLEFT)) &&
				IsCollidingObstacle(p, obst)) {
					// for the server . . .
					if (networkState == NETSTATE_SERVER) {
						// send out a network update the colliding tank
						p->flags |= PLAYER_UPDATE_TANK;
					}
					// check for motion
					if (p->flags & PLAYER_FORWARD) {
						MoveTank(p, -10);
						goto UpdatePlayers_contObst;
					}
					else if (p->flags & PLAYER_BACKWARD) {
						MoveTank(p, 8);
						goto UpdatePlayers_contObst;
					}
					// check for rotation
					if (p->flags & PLAYER_ROTRIGHT) {
						RotateTank(p, -64);
					}
					else if (p->flags & PLAYER_ROTLEFT) {
						RotateTank(p, 64);
					}
				}
				UpdatePlayers_contObst:
				// check for a collision between a shell and an obstacle
				if ((p->flags & PLAYER_SHOT) &&
				((p->shellX >> 8) >= obst->loc.x) &&
				((p->shellX >> 8) < (obst->loc.x + obst->loc.w)) &&
				((p->shellY >> 8) >= obst->loc.y) &&
				((p->shellY >> 8) < (obst->loc.y + obst->loc.h))) {
					// The shell collided, so kill it.
					p->flags &= ~PLAYER_SHOT;
					ClearRenderItem(&(p->shell));
					batchShellDraw[BatchIndex(p->pid)] &= ~BatchBit(p->pid);
				}
			}
			#ifdef USE_AUDIO
			// if the local player is not dead . . .
			// for players other than the local player that are moving . . .
			if ((p != local) &&
			!(local->flags & (PLAYER_DEAD | PLAYER_MOVEMENT_SOUND)) &&
			(p->flags & PLAYER_MOVEMENT_SOUND)) {
				// compute the square of the distance to the local player
				dist = DistSq(p, local);
				// see if this is the closest tank to the player's tank
				// so far
				if (dist < closest) {
					// record it
					closest = dist;
				}
			}
			#endif
		}
	} while (outdated);
	// for servers, if there was a score update . . .
	if ((networkState == NETSTATE_SERVER) && (numScores > 0)) {
		// send the score update
		SendScoreUpdate(scores, numScores);
	}
	// if the game should be rendered . . .
	if (dInfo.showGame) {
		RenderGame();
	}
	#ifdef USE_AUDIO
	// if there has been a change in the tank distance . . .
	if (closest != oldClosest) {
		// if there is a tank movement sound other than for the local player
		// that is close enough to be audible . . .
		if (closest < (MAX_SQ_DIST >> 2)) {
			// request to play the sound
			PlayTankMove(closest);
		}
		else {
			// play sound for the local player's tank instead
			if (local->flags & PLAYER_MOVEMENT_SOUND) {
				PlayLocalTankMove();
			}
			else {
				PlayTankIdle();
			}
		}
		// record the distance for the next iteration of the game loop
		oldClosest = closest;
	}
	#endif
}

#undef QueuePlaceTank
#undef QueuePlaceShell
#undef QueuePlaceName

void QueuePlaceTank(Uint8 pid) {
	batchTankDraw[BatchIndex(pid)] |= BatchBit(pid);
}

void QueuePlaceShell(Uint8 pid) { 
	batchShellDraw[BatchIndex(pid)] |= BatchBit(pid);
}

void RemoveShell(Uint8 pid) { 
	batchShellDraw[BatchIndex(pid)] &= ~BatchBit(pid);
}

void QueuePlaceName(Uint8 pid) {
	batchNameDraw[BatchIndex(pid)] |= BatchBit(pid);
}
