/**
 * @file game.h
 * Interface to game specific code.
 * @ingroup GameImpl
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "player.h"

/**
 * @defgroup GameImpl Game Play
 * Implements the game play stuff.
 *
 * @{
 */

/**
 * Responds to the collected user input by changing data for the local player.
 * @param eventState  An array of input states. The array is indexed by
 *                    ::EVENT_TYPES.
 * @author Jeff Jackowski
 */
void HandleInput(Uint8 *eventState);

/**
 * Updates all player data to the present time.
 * @todo Revamp this function to support more robust network update scheme (as
 *       per design), all gameplay functions, and sound.
 *
 * @author Jeff Jackowski
 */
void UpdatePlayers(Uint32 time);

/**
 * Puts the label for the player in the correct position. The tank must already
 * be positioned onscreen with PlaceRenderItem() prior to calling this function.
 * Call this function every time a tank moves whether or not the player labels
 * are displayed.
 * @param p  The player who's label must be placed for display.
 * @author   Jeff Jackowski
 */
void PlacePlayerLabel(Player *p);

/**
 * Enables the display of labels showing player names.
 * @author Jeff Jackowski
 */
void EnablePlayerLabels();

/**
 * Disables the display of labels showing player names.
 * @author Jeff Jackowski
 */
void DisablePlayerLabels();

/**
 * Moves a player's tank. The direction moved is the facing of the player.
 * Sets the player's dirty flag if the visible position changes.
 * @param p  A pointer to the player to move.
 * @param factor  A factor multiplied to the distance to move. Negative values
 *                move the tank backwards and are divided by two. Zero maintains
 *                the currently set location and will redarw if needed -- use
 *                this when the tank is moved by command of the server.
 * @author  Jeff Jackowski
 */
void MoveTank(Player *p, int factor);

/**
 * Moves a player's shell. Respomsible for removing shells that have traveled
 * their maximum distance.
 *
 * @todo     Add pong shell behavior.
 *
 * @param p  A pointer to the player who's shell must be moved.
 *
 * @author  Jeff Jackowski
 */
void MoveShell(Player *p);

/**
 * Spawns the given tank.
 * @pre  @a p is dead or just joined the game, and this process is a server.
 * @param p   A pointer to the player to spawn.
 * @author  Jeff Jackowski
 */
void SpawnTank(Player *p);

/**
 * Sets the flag that indicates the given player's tank needs to be redrawn.
 * @param   pid  The player's ID.
 * @author  Jeff Jackowski
 */
void QueuePlaceTank(Uint8 pid);

/**
 * Sets the flag that indicates the given player's shell needs to be redrawn.
 * @param   pid  The player's ID.
 * @author  Jeff Jackowski
 */
void QueuePlaceShell(Uint8 pid);

/**
 * Removes the flag that indicates the given player's shell needs to be redrawn.
 * @param   pid  The player's ID.
 * @author  Jeff Jackowski
 */
void RemoveShell(Uint8 pid);

/**
 * Sets the flag that indicates the given player's name label needs to be
 * redrawn.
 * @param   pid  The player's ID.
 * @author  Jeff Jackowski
 */
void QueuePlaceName(Uint8 pid);

/** @todo  Keep this here?  */
void RenderGame();

/** @} */
