/**
 * @file gameconfig.h
 * Game configuration interface, including options, areas, etc.
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "message.h"
// SDL is included only for its SDL_Rect type. It is used to define spawn
// points because it fits well with the rest of the graphics and collision code.
#include <SDL.h>

/**
 * Stores information on the current game's options.
 *
 * @todo  Put this in a different file?
 *
 * @author Jeff Jackowski
 */
struct GameOptions_t {
	/**
	 * The maximum length of the game in milliseconds, or -1 to indicate
	 * no imposed length.
	 */
	Uint32 duration;
	/**
	 * The score for each team.
	 */
	Sint16 teamScore[TEAM_MAX];
	/**
	 * The winning score. This is modified by the isWinScoreAbs flag.
	 */
	Sint16 winScore;
	/**
	 * The width of the battle field in logical pixels (the kind that make up
	 * the tanks).
	 */
	Uint16 width;
	/**
	 * The height of the battle field in logical pixels (the kind that make up
	 * the tanks).
	 */
	Uint16 height;
	/**
	 * The number of obstacles on the battle field.
	 */
	Uint16 obstacles;
	/**
	 * The total number of network messages that describe the game configuration.
	 */
	Uint16 totalMsgs;
	/**
	 * The number of network messages describing the game configuration that
	 * have been received. On the server, this should be set to the same value
	 * as totalMsgs.
	 */
	Uint16 rcvdMsgs;
	/**
	 * The number of spawn areas on the battle field for each team.
	 */
	Uint8 spawnAreas[TEAM_MAX];
	/**
	 * The number of goal areas on the battle field for each team.
	 */
	Uint8 goalAreas[TEAM_MAX];
	/**
	 * A set of bit flags used to modify the game. The valid flags are defined
	 * in ::GAME_OPT_FLAGS.
	 * @note  Since this field is sent over the network, a bit field struct
	 *        cannot be used without breaking big to little endian
	 *        compatibility.
	 */
	Uint8 flags;
};

typedef struct GameOptions_t  GameOptions;

enum GAME_OPT_FLAGS {
	/**
	 * When true, indicates that the game configuration has been set
	 * and is now valid. The configuration will not be sent to clients
	 * until after this flag is set in LoadConfig().
	 */
	OPT_ISSET            = 1,
	/**
	 * True if an error occured in the last attempt to load a game
	 * configuration. Used on the server to show that a battle field
	 * file could not be loaded becuase if errors within it.
	 */
	OPT_ERROR            = 2,
	/**
	 * True to allow the use of guided shells.
	 */
	OPT_USEGUIDEDSHELL   = 4,
	/**
	 * True to allow the use of pong shells.
	 */
	OPT_USEPONGSHELL     = 8,
	/**
	 * True if the game is a capture-the-flag match, false if it is
	 * a deathmatch.
	 */
	OPT_ISCTF            = 16,
	/**
	 * True if the winning score is absolute -- the team that reaches
	 * it first wins. If false, the score is relative -- the team that
	 * scores the given amount above the other team's score wins.
	 */
	OPT_ISWINSCOREABS    = 32,
	/**
	 * Allow the game to end when neither team wins. If false, sudden
	 * death overtime will be used to determine a victor.
	 */
	OPT_ALLOWDRAW        = 64,
	/**
	 * Indicates that the team score has changed and needs to be sent to the
	 * clients. Only valid on the server.
	 */
	OPT_UPDATESCORE      = 128
};

/**
 * The current game's options.
 */
extern GameOptions gameOpts;

/**
 * The lists of spawn areas for the game organized by owning team.
 * @todo  Determine if this is the best spot for this declaration.
 */
extern SDL_Rect *spawnAreas[TEAM_MAX];

/**
 * The lists of CTF goal areas for the game organized by owning team.
 * @todo  Determine if this is the best spot for this declaration.
 */
extern SDL_Rect *goalAreas[TEAM_MAX];

#define BFLDERRLEN 120

/**
 * A description with human readible error data from the last error while
 * reading in a battlefield file.
 */
struct BattleFieldError_t {
	/**
	 * The line number where the error was encountered, or -1 if the error did
	 * not occur on a line. If the value is -1, all other fields should be
	 * ignored.
	 */
	int line;
	/**
	 * The file where the error was encountered.
	 */
	char file[80];
	/**
	 * The message describing the error.
	 */
	rtsa_char msg[BFLDERRLEN];
};

typedef struct BattleFieldError_t  BattleFieldError;

/**
 * The description of the last error to occur while reading a battlefield file.
 * If no error has occured, the data is invalid.
 */
extern BattleFieldError bfErr;

/**
 * @defgroup NetGConf Network Game Configuration
 *
 * Implements the process of sending the game configuration from the server to
 * all the clients, and receiving it on the client.
 *
 
@todo Complete this doc!

 Initialize by creating all messages with config data. Then number of messages
 and contents will be known prior to transmission.
 
 * On the server, ServiceGameConfig() is called every frame to handle sending
 * out more of the game's configuration. Clients are added to a list of clients
 * needing more information when they acknowledge the connection acceptence
 * message (see AckConnAccept()). When all the data is sent, the process
 * will begin again if there is at least one client still lacking some of the
 * configuration data. Sending the configuration happens as a global action
 * rather than on a per client basis to minimize the required bandwith on a LAN.
 *
 * @todo  Fix the above and below to match realtiy.
 *
 * The process of sending the configuration data once is:
 * - Send the game options (::MSG_TYPE_OPTION)
 * - Wait
 * - Loop through obstacles
 *   - Send the next ::MAX_OBSTACLES_PER_MSG obstacles, or the last remaining
 *     obstacles
 *   - Wait
 *   - If all clients have all configuration data, quit now.
 * - Loop through spawn areas
 *   - Send the next ::MAX_SPAWN_PER_MSG spawn points, or the last
 *     remaining points
 *   - Wait
 *   - If all clients have all configuration data, quit now.
 * - Determine if more configuration data must be sent in case more clients
 *   connected while configurartion data was being sent.
 *
 * The waits above are ment to limit the bandwith required by the server to
 * prevent joining players from disrupting a game in progress. Its
 * implementation performs the following actions in order:
 * - Grab the destination map after calling AddMessage()
 * - Wait at least 64ms in a non-blocking manner
 * - Check the destination clients to see if at least one has received the
 *   data.
 *   - If received, quit waiting.
 *   - If not received:
 *     - Wait for up to another 64ms while still checking for reception.
 *     - If not received by the destination, jump to seeing if other clients
 *       need the configuration data retransmitted.
 *
 * @author Jeff Jackowski
 * @ingroup Networking
 * @{
 */

/**
 * The maximum number of obstacles that can be transmitted in a single
 * ::MSG_TYPE_OBSTACLE message.
 * @note  There is no corresponding constant for the number of spawn points
 *        that a message may contain because the different message format makes
 *        the maximum a variable.
 */
#define MAX_OBSTACLES_PER_MSG  62

/**
 * Loads in a game configuration and prepares to send the configuration to all
 * clients. Prior to calling this function, set all the fields of ::gameOpts
 * that do not relate to the battle field. Clients may be added to the list of
 * those waiting to receive the game configuration prior to calling this
 * function.
 * @author Jeff Jackowski
 * @param fname  The name of the file that contains the battle field
 *               description. If the full path is not given, the working
 *               directory will be searched for the file.
 */
Bool InitGameConfig(const char *fname);

/**
 * Destroyes a loaded game configuration and deallocates all memory used to
 * store that configuration.
 * @author Jeff Jackowski
 */
void UninitGameConfig();

/**
 * Adds a client to the list of clients waiting for game configuration data.
 * Clients may be added to the list prior to calling InitGameConfig().
 * @param client  A pointer to the client to add.
 * @author Jeff Jackowski
 */
void AddToConfigList(ClientData *client);

/**
 * Removes a given client from the list of clients waiting to receive the game
 * configuration if the client is in the list.
 * @param client  The client to remove from the list. It may be a client that
 *                is not in the list.
 * @author Jeff Jackowski
 */
void RemoveFromConfigList(ClientData *client);

/**
 * The start of the list of clients that need to be given the game
 * configuration. Clients will normally be removed from this end.
 */
ClientData *configListStart;

/**
 * The end of the list of clients that need to be given the game
 * configuration. Clients will normally be added to this end.
 */
ClientData *configListEnd;

/**
 * Handles an iteration of sending game configuration data. Intended to be
 * called every frame by the server between the calls to ReceiveMessages()
 * and ServiceMessages().
 * @return  True on success, false if a message could not be transmitted.
 * @author  Jeff Jackowski
 */
Bool ServiceGameConfig();

/**
 * Handles an incomming message from the network describing the game options.
 * @internal
 * @param origin    The originating address of the message.
 * @param buffer    The buffer that contains the message to process.
 * @param unused    A required but unused part of the function prototype.
 * @param noid      A required but unused part of the function prototype.
 * @return          True when a valid disconnect is received.
 * @author  Jeff Jackowski
 */
Bool HandleGameOption(SocketAddr *origin, NetBuffer *buffer, Uint8 *unused,
	Uint16 noid);

/**
 * Handles an incomming message from the network describing the obstacles on
 * the battlefield.
 * @internal
 * @param origin    The originating address of the message.
 * @param buffer    The buffer that contains the message to process.
 * @param unused    A required but unused part of the function prototype.
 * @param noid      A required but unused part of the function prototype.
 * @return          True when a valid disconnect is received.
 * @author  Jeff Jackowski
 */
Bool HandleObstacle(SocketAddr *origin, NetBuffer *buffer, Uint8 *unused,
	Uint16 noid);

/**
 * Handles an incomming message from the network describing the goal and spawn
 * areas.
 * @internal
 * @param origin    The originating address of the message.
 * @param buffer    The buffer that contains the message to process.
 * @param unused    A required but unused part of the function prototype.
 * @param noid      A required but unused part of the function prototype.
 * @return          True when a valid disconnect is received.
 * @author  Jeff Jackowski
 */
Bool HandleArea(SocketAddr *origin, NetBuffer *buffer, Uint8 *unused,
	Uint16 noid);

/**
 * Handles an incomming message from the network with a updated scores.
 * @internal
 * @param origin    The originating address of the message.
 * @param buffer    The buffer that contains the message to process.
 * @param unused    A required but unused part of the function prototype.
 * @param noid      A required but unused part of the function prototype.
 * @return          True when a valid disconnect is received.
 * @author  Jeff Jackowski
 */
Bool HandleScoreUpdate(SocketAddr *origin, NetBuffer *buffer, Uint8 *unused,
	Uint16 noid);

/**
 * A store of score updates used to retain updates over a frame and send them
 * all out together.
 */
struct ScoreUpdate_t {
	/**
	 * The player ID of the killer.
	 */
	Uint8  killerId;
	/**
	 * The player ID of the killed.
	 */
	Uint8  killedId;
};

typedef struct ScoreUpdate_t  ScoreUpdate;

/**
 * Sends a score update to all connected clients.
 * @param scores  The new scores/kills to send.
 * @param len     The number of player updates to send.
 * @return True on sucess.
 * @author Jeff Jackowski
 */
Bool SendScoreUpdate(ScoreUpdate *scores, int len);

/** @} */

