/**
 * @file genconmac.h
 * The game's general constants and macros.
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#ifndef __genconmac_h__
#define __genconmac_h__

/**
 * The maximum number of characters allowed in a player's name.
 */
#define MAX_NAME_LEN  32

/**
 * The maximum number of players allowed to participate in a single game.
 * Increasing the value can cause an exponential increase in the iterations
 * through some loops.
 * @note  This value should be a power of two for best performance and use of
 *        memory.
 * @warning Setting this value to more than 256 will require changes to the
 *          code and to the newtork message formats.
 */
#define MAX_PLAYERS  32

/**
 * The intended duration of each frame in milliseconds.
 */
#define FRAME_DURATION  16

/**
 * The server's time offset of each frame in milliseconds relative to the
 * client's frame.
 */
#define SRV_FRAME_OFFSET  6

/**
 * The margin, in pixels, between the game's title and the top of the window.
 */
#define TOP_MARGIN       8

/**
 * The margin, in pixels, between the GPL or menu items and the bottom of the
 * window.
 */
#define BOTTOM_MARGIN    4

/**
 * The high-level event types used in the game and the menus to abstract the
 * different inputs that can trigger the events.
 */
enum EVENT_TYPES {
	/**
	 * The up input; a request to move the tank forward.
	 */
	EC_FORWARD,
	/**
	 * The down input; a request to move the tank backward.
	 */
	EC_BACKWARD,
	/**
	 * Request to turn the tank right.
	 */
	EC_ROTRIGHT,
	/**
	 * Request to turn the tank left.
	 */
	EC_ROTLEFT,
	/**
	 * Request to fire a shell.
	 */
	EC_FIRE,
	/**
	 * Request to toggle between full-screen and windowed display.
	 */
	EC_FS_TOGGLE,
	/**
	 * A character event status. The event does not include the character code,
	 * just information on the status of the key press.
	 */
	EC_CHARACTER,
	EC_MAX
};

/**
 * The event codes that are used to process user input in a way that separates
 * the exact input used from the code that responds to the input. The least
 * significant bit of an event is set when the condition exists, like when a
 * key is pressed. When the condition no longer exists, the value is set to
 * zero. This allows other bits to be used while a condition exists.
 *
 * The game code sets the second bit to indicate that the event has already been
 * handled; this bit is not cleared by the code reading in events until the
 * condition is cleared (key released).
 *
 * The array is indexed by the different events specified in ::EVENT_TYPES.
 */
extern Uint8 eventState[EC_MAX];

/**
 * The valid team codes.
 */
enum {
	/**
	 * The red team as seen in non-pong tank games in the original Combat.
	 */
	TEAM_RED,
	/**
	 * The blue team as seen in non-pong tank games in the original Combat.
	 */
	TEAM_BLUE,
	/**
	 * The green team as seen in pong tank games in the original Combat.
	 * @warning This team's use is not fully implemented.
	 */
	TEAM_GREEN,
	/**
	 * The purple team as seen in pong tank games in the original Combat.
	 * @warning This team's use is not fully implemented.
	 */
	TEAM_PURPLE,
	TEAM_MAX,
	/**
	 * The team value used for players who have connected to the server but
	 * have yet to select a team to join.
	 */
	TEAM_UNDEC = 0xFF
};

/**
 * The layer codes.
 */
enum LAYERS {
	/**
	 * Used for the background (image, solid color, ...).
	 */
	LAYER_BACKGROUND,
	/**
	 * Used for obstales on the play field.
	 */
	LAYER_OBSTACLES,
	/**
	 * Used for text messages.
	 */
	LAYER_TEXT,
	/**
	 * Used for player name labels on the tanks.
	 */
	LAYER_NAMES,
	/**
	 * Used for tanks and shells.
	 */
	LAYER_TANKS,
	LAYER_MAX
};

#define MOTION_0_0     0
#define MOTION_0_1   256
#define MOTION_22_0   98
#define MOTION_22_1  237
#define MOTION_45_0  181
#define MOTION_45_1  181

#define FMOTION_0_0   0.0f
#define FMOTION_0_1   1.0f
#define FMOTION_22_0  0.374606593f
#define FMOTION_22_1  0.927183855f
#define FMOTION_45_0  0.707106781f
#define FMOTION_45_1  0.707106781f

#define MOTION_DIV   4

/**
 * The constant multiplied to the values in the motion table to produce the
 * motion values for shells.
 */ 
#define MOTION_SHELL_FACTOR  5

/**
 * The maximum distance a guided shell can travel.
 */
#define SHELL_GUIDED_DIST   96
/**
 * The maximum distance a pong shell can travel. For these shells to be
 * effective, they must travel notably further than guided shells.
 */
#define SHELL_PONG_DIST    192

/**
 * The number of frames a player stays dead after being hit.
 */
#define DEAD_TIME   180

/**
 * Divides and rounds up if the fractional part of the result is greater than
 * 0.5. Only integer instructions are used. Implemented as a macro. Any
 * compiler should be able to reduce this to a constant if both
 * parameters are constants.
 * @param n  The numerator.
 * @param d  The denominator.
 * @return   The divided result after rounding.
 * @author   Jeff Jackowski
 */
#define rounddiv(n, d) \
     ((((n) % (d)) > ((int)(d) / 2)) ? ((int)(n) / (d) + 1) : ((int)(n) / (d)))

/**
 * Divides integers and rounds up any fractional part of the result. Any
 * compiler should be able to reduce this to a constant if both
 * parameters are constants.
 * @param n  The numerator.
 * @param d  The denominator.
 * @return   The divided result after rounding up.
 * @author   Jeff Jackowski
 */
#define rounddivup(n, d) \
        ((((n) % (d)) > 0) ? ((int)(n) / (d) + 1) : ((int)(n) / (d)))

/**
 * Finds the index into an array for a particular bit.
 * @param  bitid     The number of the bit in the array.
 * @param  datatype  The array element data type.
 * @author Jeff Jackowski
 */
#define bitindex(bitid, datatype) \
        (((unsigned)bitid) / ((unsigned)sizeof(datatype) * (unsigned)8))

/**
 * Produces the bit flag for a particular bit at its location in an array.
 * @param  bitid     The number of the bit in the array.
 * @param  datatype  The array element data type.
 * @author Jeff Jackowski
 */
#define bitflag(bitid, datatype) (1 << ((bitid) & (sizeof(datatype) * 8 - 1)))

/**
 * Computes the length of an array, in elements, that can store a minimum number
 * of bits. This macro is a constant when given parameters with constant values.
 * @param  bits      The minimum number of bits the array must hold.
 * @param  datatype  The array element data type.
 * @author Jeff Jackowski
 */
#define bitarraylen(bits, datatype)  rounddivup(bits, sizeof(datatype) * 8)

/**
 * A simple absolute value function. Avoids having to inculde math
 * libraries that would be underused by this program.
 * @param v  The value who's magnetude will be returned.
 * @return   The absolute value of v.
 * @author   Jeff Jackowski
 */
#define absval(v)  (((v) < 0) ? -(v) : (v))

#endif

