/**
 * @file hostlookup.h
 * Interface to host lookup functions.
 * @ingroup NetHostLookup
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#ifndef _hostlookup_h_
#define _hostlookup_h_

#include "bool.h"

/**
 * @defgroup NetHostLookup Host Lookup
 *
 * A set of functions to perform asynchronous host lookups with the help of the
 * @ref IOThread. Only one lookup can be in progress at any given time, and
 * there is no mechanism to queue lookup requests. Generally the results of
 * a lookup can only be retrieved once. These restrictions simply the lookup
 * implementation as well as its public interface. They should not impose
 * problems for most networked games.
 *
 * The implementation is ment to work with IPv6, but favors IPv4. IPv6 support
 * has not been tested, so this code may fail with IPv6. 
 * @note Some platforms, like Windows, do not have an implementation of
 *       inet_pton(). A simple implementation is provided for these platforms
 *       which uses the older inet_addr() function that does not support IPv6.
 *
 * Looking up the fully qualified name and address of the local system only
 * works when the system's hostname can be resolved to an address. This usually
 * works on Windows. On Linux, if there is no nameserver with this information,
 * or no such line in /etc/hosts, then the local address will only be resolved
 * to the loopback address. This problem is common on home LANs.
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 *
 * @author Jeff Jackowski
 * @ingroup Networking
 * @{
 */

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#ifdef _MSC_VER
// This isn't defined in Microsoft's ws2tcpip.h, but is defined in the mingw32
// equivalent.
#define IN6_ARE_ADDR_EQUAL  IN6_ADDR_EQUAL
#endif
#else
#include <netinet/in.h>
#endif

/**
 * The maximum hostname length supported.
 */
#define MAX_HOST_LEN  64

/**
 * Stores an IPv4 or IPv6 network address.
 */
struct INetAddr_t {
	union {
		/**
		 * The IPv4 address.
		 */
		struct in_addr  ipv4;
		/**
		 * The IPv6 address.
		 */
		struct in6_addr ipv6;
	};
	/**
	 * The address family of the stored address; either AF_INET or AF_INET6.
	 */
	int family;
};

typedef struct INetAddr_t INetAddr;

/**
 * Stores an IPv4 or IPv6 socket address. To obtain the family of the socket
 * address, use addr.sa_family.
 */
struct SocketAddr_t {
	union {
		/**
		 * The IPv4 socket address.
		 */
		struct sockaddr_in  ipv4;
		/**
		 * The IPv6 socket address.
		 */
		struct sockaddr_in6 ipv6;
		/**
		 * The generic socket address.
		 */
		struct sockaddr addr;
	};
	/**
	 * The length of the socket address structure that is in use.
	 */
	socklen_t length;
};

typedef struct SocketAddr_t SocketAddr;

/**
 * The socket address of the server that the client is in communication with.
 * The variable is declared in net.c, but mentioned here to minimize naming
 * conflict fun with Win32.
 */
extern SocketAddr outgoingAddr;

/**
 * True if two addresses, sans port, are equal.
 * @param    sa1  A pointer to a SocketAddr to compare.
 * @param    sa2  Another pointer to a SocketAddr to compare.
 * @relates  SocketAddr_t
 * @author   Jeff Jackowski
 */
#define EqualAddresses(sa1, sa2) \
	(((sa1)->addr.sa_family == (sa2)->addr.sa_family) &&               \
	((((sa2)->addr.sa_family == AF_INET)                               \
	&& ((sa2)->ipv4.sin_addr.s_addr == (sa1)->ipv4.sin_addr.s_addr))   \
	|| (((sa2)->addr.sa_family == AF_INET6) &&                         \
	IN6_ARE_ADDR_EQUAL(&((sa2)->ipv6.sin6_addr),                       \
	&((sa1)->ipv6.sin6_addr)))))

/**
 * Stores the name of the local system.
 */
extern char localhostName[MAX_HOST_LEN];

/**
 * Stores the address of the local system as a string.
 */
extern char localhostAddrStr[INET6_ADDRSTRLEN];

/**
 * Starts the process of finding the address of a host based on its name.
 * The lookup occurs on the separate @ref IOThread. Obtain the result by
 * calling LookupResult().
 * @warning Multiple lookup requests are not supported and are not queued.
 * @param   name  The hostname of the machine who's address is to be found.
 * @author  Jeff Jackowski
 */
void StartAddrLookup(char *name);

/**
 * Starts the process of finding the full hostname of a machine based on its
 * address. The lookup occurs on the separate @ref IOThread. Obtain the result
 * by calling LookupResult().
 * @warning Multiple lookup requests are not supported and are not queued.
 * @param addrstr A string containing the address in dotted decimal notation.
 * @warning       On Windows, only IPv4 address are supported.
 * @param name    The string where the hostname will be placed. It must be
 *                ::MAX_HOST_LEN bytes long.
 * @return  True on success, false if the address string is bad.
 * @author  Jeff Jackowski
 */
Bool StartStrNameLookup(char *addrstr, char *name);

/**
 * Starts the process of finding the full hostname of a machine based on its
 * address. The lookup occurs on the separate @ref IOThread. Obtain the result
 * by calling LookupResult().
 * @warning Multiple lookup requests are not supported and are not queued.
 * @param addr    The socket address of the host who's name is sought.
 * @warning       On Windows, only IPv4 address are supported.
 * @param name    The string where the hostname will be placed. It must be
 *                ::MAX_HOST_LEN bytes long.
 * @author  Jeff Jackowski
 */
void StartNameLookup(SocketAddr *addr, char *name);

/**
 * Provides the results of a host lookup started by a call to either
 * StartAddrLookup() or StartNameLookup(), or a notice that the operation is
 * in progress. After the results are retrived with this function, they will
 * no longer be availble through this interface.
 * @warning Multiple lookup requests are not supported and are not queued.
 * @param  name  For name lookups, this is ignored. For address lookups,
 *               (StartAddrLookup(), lookup by name) this provides a place to
 *               store the fully qualified name. The name pointer must point
 *               to an array ::MAX_HOST_LEN bytes long, or be NULL.
 * @param  addr  If non-NULL, the indicated address struct will be filled with
 *               the address in network byte order.
 * @return  -1 on error, 0 if the operation is still in progress, or 1 if
 *          the operation has completed and the results delivered.
 * @author  Jeff Jackowski
 */
int LookupResult(char *name, INetAddr *addr);

/**
 * An operation for the @ref IOThread that attempts to discover the local
 * system's hostname and address. Performs an address lookup of what the system
 * claims to be its name. If the lookup finds a name other than localhost, the
 * new name will be stored in ::localhostName. Otherwise, the name stored will
 * be the one returned by gethostname(). If the address is not for the
 * loopback interface, a string representation of the address will be stored in
 * ::localhostAddrStr. Otherwise, the string will have zero length.
 * @warning  On Windows, if the local system's IPv6 address is found first, the
 *           address string will have zero length.
 *
 * @warning  On Linux (and other UNIX-like syetems?), the address found on many
 *           home LANs, and some other LANs, will be the loopback address. To
 *           avoid this requires another address listed with the system's name
 *           in /etc/hosts, or (maybe?) an entry in a nameserver used by the
 *           system.
 *
 * @author   Jeff Jackowski
 */
void DoHostnameLookup(void *);

#ifdef WIN32
/**
 * A simple but not quite complete IPv4 only implementation of the standard
 * function inet_pton for platforms that lack it, like Windows.
 * @warning  The valid address 255.255.255.255 cannot be converted by this
 *           function.
 * @param  fam     The address family. Only AF_INET is supported.
 * @param  straddr The address as a string in doted decimal notation.
 * @param  addr    The address structure to fill. This should be a pointer to
 *                 an in_addr struct.
 * @return Positive on success, 0 if the string has a bad format, -1 on error
 *         or attempt to use IPv6.
 * @author Jeff Jackowski
 */
int inet_pton(int fam, char *straddr, void *addr);
#endif

/** @} */

#endif

