/**
 * @file iothread.c
 * Implementation for a thread used for blocking I/O operations.
 * @ingroup IOThread
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include <SDL.h>
#include <SDL_thread.h>
#include <assert.h>
#include "iothread.h"
#include "random.h"

/**
 * Used to notify the I/O thread that there is work to do.
 * @internal
 * @ingroup IOThread
 */
static SDL_cond *notify;

/**
 * A reference to information about the running I/O thread.
 * @internal
 * @ingroup IOThread
 */
static SDL_Thread *thread = NULL;

/**
 * Stores information on a queued operation.
 * @internal
 * @ingroup IOThread
 */
struct Operation_t {
	/**
	 * The function that will perform the operation.
	 */
	ioOperation op;
	/**
	 * Some data to pass to the function.
	 */
	void *data;
};

/**
 * The operation queue.
 * @internal
 * @ingroup IOThread
 */
static struct Operation_t queue[IO_QUEUE_SIZE];

/**
 * The number of operations in the queue.
 * @internal
 * @ingroup IOThread
 */
static int queueLen = 0;

/**
 * The index of the next operation in the queue.
 * @internal
 * @ingroup IOThread
 */
static int queueNext = 0;

/**
 * The semaphore used to control access to the operation queue.
 * @internal
 * @ingroup IOThread
 */
static SDL_mutex *queueMutex = NULL;

/**
 * The flags used to indicate when:
 * - the running operation should return early
 * - the I/O thread should terminate.
 * @internal
 * @ingroup IOThread
 */
static int opFlags = 0;

/**
 * Bit flag used with ::opFlags to indicate that the current operation should
 * return early.
 * @internal
 * @ingroup IOThread
 */
#define FLAG_CANCEL  1

/**
 * Bit flag used with ::opFlags to indicate that the I/O thread should
 * terminate.
 * @internal
 * @ingroup IOThread
 */
#define FLAG_TERM    2

/**
 * The thread body that waits for queued I/O requests and runs them.
 * @internal
 * @ingroup IOThread
 * @author Jeff Jackowski
 */
static int ioThread(void *ignore) {
	SDL_mutex *waitMutex;
	struct Operation_t oper;
	int next = 1;
	int lastLen;
	// initalize the random number generator
	RandInit();
	// create new mutex semaphores
	if ((queueMutex = SDL_CreateMutex()) == NULL) {
		// create failed!
		return -1;
	}
	if ((waitMutex = SDL_CreateMutex()) == NULL) {
		// create failed!
		SDL_DestroyMutex(queueMutex);
		queueMutex = NULL;
		return -1;
	}
	// lock it
	if (SDL_mutexP(waitMutex) == -1) {
		// lock failed!
		ioThread_die:
		SDL_DestroyMutex(waitMutex);
		SDL_DestroyMutex(queueMutex);
		queueMutex = NULL;
		return -1;
	}
	// wait for stuff to do
	while (SDL_CondWait(notify, waitMutex) == 0) {
		ioThread_noWait:
		// lock the queue
		if (SDL_mutexP(queueMutex) == -1) {
			// failed to use a mutex, so this isn't going to work
			goto ioThread_die;
		}
		// check for nothing in the queue
		if (queueLen-- == 0) {
			// time to quit
			break;
		}
		// store the number of items queued for a later check
		lastLen = queueLen;
		// clear the cancel request flag
		opFlags = 0;
		// copy the queue entry
		oper = queue[next];
		// advance the next counter
		next = (next + 1) % IO_QUEUE_SIZE;
		// release the queue
		SDL_mutexV(queueMutex);
		// run the operation
		(*oper.op)(oper.data);
		// need to terminate?
		if (opFlags & FLAG_TERM) {
			// leave the loop to perform cleanup
			break;
		}
		// check for more stuff queued
		if (queueLen) {
			// do not wait on the condition; continue with the next operation
			// now
			goto ioThread_noWait;
		}
	}
	SDL_DestroyMutex(waitMutex);
	SDL_DestroyMutex(queueMutex);
	queueMutex = NULL;
	return 0;
}

void StartIOThread() {
	// create the condition gizmo used to signal the I/O thread
	notify = SDL_CreateCond();
	// start the I/O thread
	thread = SDL_CreateThread(ioThread, NULL);
	// avoid possible race conditions by allowing the thread some time to
	// initalize itself
	SDL_Delay(2);
}

void StopIOThread() {
	// see if the thread was started and a queue mutex exists
	if (thread && queueMutex) {
		// The thread was started and should be running. Try to lock the queue.
		if (SDL_mutexP(queueMutex) == -1) {
			// kill the thread forcefully.
			EndIOThread_kill:
			SDL_KillThread(thread);
			//printf("Forced end of IO thread\n");
		}
		else {
			// remove the last item from the queue
			queueLen = 0;
			// request than any operation in progess please quit
			opFlags = FLAG_CANCEL | FLAG_TERM;
			// release the queue lock
			SDL_mutexV(queueMutex);
			// signal the thread
			if (SDL_CondSignal(notify) == -1) {
				// couldn't do it; assure a dead thread
				goto EndIOThread_kill;
			}
			// wait for the thread to finish
			SDL_WaitThread(thread, NULL);
			//printf("Normal end of IO thread\n");
		}
		thread = NULL;
	}
	SDL_DestroyCond(notify);
}

int QueueOperation(ioOperation op, void *data) {
	int ret;
	// the queue should already exist
	assert(queueMutex != NULL);
	// check for full queue
	if (queueLen == IO_QUEUE_SIZE)
		return -1;
	// attempt to lock the queue
	if (SDL_mutexP(queueMutex) == -1) {
		// failed
		return -2;
	}
	// compute new length
	ret = ++queueLen;
	// compute the index of the new item
	queueNext = (queueNext + 1) % IO_QUEUE_SIZE;
	// add the item
	queue[queueNext].op = op;
	queue[queueNext].data = data;
	// signal the thread
	SDL_CondSignal(notify);
	// release the lock
	SDL_mutexV(queueMutex);
	// send back the remaining queue length
	return ret;
}

Bool IsCancelationRequested() {
	return opFlags & FLAG_CANCEL;
}

Bool CancelOperations() {
	// see if the thread was started and a queue mutex exists
	if (thread && queueMutex) {
		// The thread was started and should be running. Try to lock the queue.
		if (SDL_mutexP(queueMutex) != -1) {
			// remove the last item from the queue
			queueLen = 0;
			// request than any operation in progess please quit
			opFlags = FLAG_CANCEL;
			// release the queue lock
			SDL_mutexV(queueMutex);
			// sucess
			return TRUE;
		}
	}
	// failed
	return FALSE;
}

