/**
 * @file iothread.h
 * Interface to a thread for blocking I/O operations.
 * @ingroup IOThread
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "bool.h"

/**
 * @defgroup IOThread I/O Thread
 *
 * A system to queue up blocking I/O operations to run on a separate thread.
 *
 * First, StartIOThread() must be called. Then asynchronous operations may be
 * queued by calling QueueOperation(). When finished, call StopIOThread() to
 * terminate the thread and cleanup gracefully. To cancel all queued operations
 * and the running operation, call CancelOperations(). Long running operations
 * should call IsCancelationRequested() to see if they should return early.
 *
 * The random number generator is gaurenteed to be ready for use by operations
 * running on the I/O thread.
 *
 * @author Jeff Jackowski
 * @{
 */

#ifndef IO_QUEUE_SIZE
/**
 * The size of the queue for I/O operations. For best performance, use a power
 * of two.
 */
#define IO_QUEUE_SIZE  4
#endif

/**
 * Starts the I/O thread. This must be called prior to queuing any operations.
 * The thread will initialize the random number generator to handle the case
 * where the generator uses therad local data.
 * @bug    On Linux with SDL 1.2.8, valgrind shows that when
 *         SDL_SYS_CreateThread() makes a call to pthread_create(), two things
 *         are allocated totaling about 10.5KB that are not deallocated.
 * @author Jeff Jackowski.
 */
void StartIOThread();

/**
 * Terminates the I/O thread.
 * @note   This function will not return until after any blocking operation
 *         in progress runs to completion.
 * @author Jeff Jackowski
 */
void StopIOThread();

/**
 * The function prototype of all functions called by the I/O thread to perform
 * blocking operations.
 * @param data  A pointer to arbitrary data used by the operation.
 */
typedef void (*ioOperation)(void *data);

/**
 * Adds an operation to the queue of pending blocking operations.
 * @param op    The function that will perform the operation.
 * @param data  Arbitrary datat that will be passed to the operation function.
 * @return  The number of operations that are queued. If the queue is
 *          already full, -1 is returned and the operation is not queued. If
 *          there is an error while locking the queue semaphore, -2 is returned.
 * @author  Jeff Jackowski
 */
int QueueOperation(ioOperation op, void *data);

/**
 * Tells if a request has been made to cancel the current operation on the
 * I/O thread through a call to CancelOperations(). Long running operations
 * should call this function periodically so they can terminate early on
 * request.
 * @return  True if the operation should quit now.
 * @author  Jeff Jackowski
 */
Bool IsCancelationRequested();

/**
 * Attempts to cancel all operations currently queued on the I/O thread. Sets
 * the cancel request flag in hopes the currently running operation will
 * terminate early, and removes all operations waiting in the queue.
 * @return  True on success, false on failure.
 * @author  Jeff Jackowski
 */
Bool CancelOperations();

/** @} */

