/**
 * @file main.c
 * The start of the program.
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#define _USE_MATH_DEFINES
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#include <fcntl.h>
#include <io.h>
#include <sys/stat.h>
#endif
#ifdef USE_AUDIO
#include "audio.h"
#endif
#include "color.h"
#include "iothread.h"
#include "tanksprite.h"
#include "game.h"
#include "gameconfig.h"
#include "net.h"
#include "random.h"
#include "states.h"
#include "menustate.h"
#include <SDL_ttf.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

extern Uint32 videoFlags;

SDL_Joystick *stick = NULL;

/**
 * The elapsed time of the frame. Required for fixed frame rate calculations.
 * @@ingroup StatePlay
 */
static Sint32 elapsedTime;

/**
 * The current simulation time in milliseconds.
 * @@ingroup StatePlay
 */
Uint32 frameTime;

/**
 * Stores the path to where the binary is located. The data files for the
 * program should also be here.
 */
char binPath[128];

/**
 * The entry point for Tank!
 * main() is responsible for:
 * - Parsing command line arguments (none yet)
 * - Delivering a GPL notice
 * - Initalizing and uninitalizing:
 *   - SDL
 *   - SDL_ttf
 *   - @ref RenderSys "The renderer"
 *   - @ref TextSys "The text system"
 *   - @ref IOThread "The blocking I/O thread"
 *   - @ref Networking "The networking layer"
 *   - The tank sprites
 *   - The random number generator
 * - Setting up the initial window
 * - Running the @ref OperStates "operating states" in a loop
 *
 * Command line parameters can be used to modify the program's operation.
 * At present, the command line parsing is very simple and will only parse
 * one argument. The arguments are:
 * - @a -s -- Run as a server rather than discover an existing one.
 * - @a -c &lt;server&gt; -- Run as a client and connect to the given server.
 * - @a -d -- Open a debug console on Windows. When used, stdout and stderr
 *      will not be written to the files SDL creates (stdout.txt and
 *      stderr.txt). Only supported on Windows with English-only builds.
 *
 * main()'s role has been expanded to include frame rate management, calling
 * Render(), and invoking the network module to send and receive messages.
 *
 * @author Jeff Jackowski
 */
int main(int argc, char *argv[]) {
	SDL_VideoInfo const *videoInfo;
	SDL_Surface *winIcon;
	
	#ifdef WIN32
	#ifndef USE_GETTEXT
	// support a debugging console on Windows
	Bool haveCon = FALSE;
	if (argc >= 2) {
		int loop;
		for (loop = 1; loop < argc; loop++) {
			if ((strcmp(argv[loop], "-d") == 0) && AllocConsole()) {
				// get the output file handle
				FILE *out =
				_fdopen(_open_osfhandle((intptr_t)GetStdHandle(STD_OUTPUT_HANDLE)
				, _O_TEXT), "w");
				// thought there was a better way to do this . . .
				*stdout = *stderr = *out;
				// flag the creation of the console for later cleanup
				haveCon = TRUE;
				break;
			}
		}
	}
	#elif defined(_MSC_VER) && (_MSC_VER > 1310)
	FILE oldout, olderr;
	oldout = *stdout;
	olderr = *stderr;
	{
		FILE *out;
		/**
		 * @bug On Windows with I18N support, all output to stdout and stderr
		 *      is lost. This is an improvement over the earlier condition
		 *      where the gettext implemenetation for {f|fw}printf() would
		 *      always crash, resulting in a crash on many error outputs.
		 */
		out = fopen("rtsa_status.txt", "w,ccs=UTF-16LE");
		*stdout = *out;
		out = fopen("rtsa_errors.txt", "w,ccs=UTF-16LE");
		*stderr = *out;
	}
	#endif
	#endif
	// output legal info in English before initializing much of anything,
	// including the gettext libraries
	#ifdef USE_GETTEXT
	wprintf(L"Retro Tank Super Attack, Copyright (C) 2007  Jeff Jackowski  jeffj@ro.com\n");
	wprintf(L"Retro Tank Super Attack comes with ABSOLUTELY NO WARRANTY. This\n");
	wprintf(L"is free software, and you are welcome to redistribute it under\n"); 
	wprintf(L"certain conditions as defined by the GPL. For more details, see\n");
	wprintf(L"the GPL included in gpl.txt with this program, or at\n");
	wprintf(L"http://www.gnu.org/copyleft/gpl.html\n\n");
	#else
	printf("Retro Tank Super Attack, Copyright (C) 2007  Jeff Jackowski  jeffj@ro.com\n");
	printf("Retro Tank Super Attack comes with ABSOLUTELY NO WARRANTY. This\n");
	printf("is free software, and you are welcome to redistribute it under\n"); 
	printf("certain conditions as defined by the GPL. For more details, see\n");
	printf("the GPL included in gpl.txt with this program, or at\n");
	printf("http://www.gnu.org/copyleft/gpl.html\n\n");
	#endif
	{
		char *end;
		// find the path of the binary
		#ifdef WIN32
		end = strrchr(argv[0], '\\');
		#else
		end = strrchr(argv[0], '/');
		#endif
		// if the character was found . . .
		if (end != NULL) {
			// set the length to include everything until the slash
			size_t len = end - argv[0] + 1;
			// copy over the path
			strncpy(binPath, argv[0], len);
			// terminate the result
			binPath[len] = 0;
		}
		else {
			// No usable path, so try to indicate the current directory.
			// This happens with Valgrind 2.4, but so far no where else.
			binPath[0] = '.';
			binPath[1] = '/';
			binPath[2] = binPath[3] = 0;
		}
		// printf("Binary path: %s\n", binPath);
	}

	#ifdef WIN32
	// initialize the winsock library
	fwprintf(stderr, L"Error output test\n");
	{
		WSADATA result;
		if (WSAStartup(0x202, &result) != 0) {
			fprintf(stderr, "Could not initalize Winsock\n");
			exit(31);
		}
		if ((result.wVersion & 0xFF) != 2) {
			fprintf(stderr,
			"Incompatible Winsock version %i.%i found. Need version 2.x\n",
			result.wVersion & 0xFF, result.wVersion >> 8);
			exit(32);
		}
	}
	/*	
	if (getenv("APPDATA")) {
		printf("Application data directory is\n\t%s\n",	getenv("APPDATA"));
	}
	*/
	#else
	/*
	if (getenv("HOME")) {
		printf("Home directory is %s\n", getenv("HOME"));
	}
	*/
	#endif
	/*
	else {
		printf("Could not determine the home directory\n");
	}
	*/
	
	RandInit();	
	StartIOThread();
	
	/*
	{  // host lookup test  ---  remove later
		char name[MAX_HOST_LEN];
		INetAddr addr;
		int ip = 0, res;
		gethostname(name, MAX_HOST_LEN);
		StartAddrLookup(name);
		while ((res = LookupResult(name, &addr)) == 0) SDL_Delay(1);
		if (res > 0) {
			ip = addr.ipv4.s_addr;
			printf("%s -> %u.%u.%u.%u\n", name, ip & 0xFF,
			(ip & 0xFF00) >> 8, (ip & 0xFF0000) >> 16, (ip & 0xFF000000) >> 24);
		}
		StartAddrLookup("owa.fullsail.com");
		while ((res = LookupResult(NULL, &addr)) == 0) SDL_Delay(1);
		if (res > 0) {
			ip = addr.ipv4.s_addr;
			printf("owa.fullsail.com -> %u.%u.%u.%u\n", ip & 0xFF,
			(ip & 0xFF00) >> 8, (ip & 0xFF0000) >> 16, (ip & 0xFF000000) >> 24);
		}
		StartStrNameLookup("216.180.54.1", name);
		while ((res = LookupResult(NULL, &addr)) == 0) SDL_Delay(1);
		if (res > 0) {
			ip = addr.ipv4.s_addr;
			printf("216.180.54.1 -> %s ", name);
			printf(" (%u.%u.%u.%u)\n", ip & 0xFF,
			(ip & 0xFF00) >> 8, (ip & 0xFF0000) >> 16, (ip & 0xFF000000) >> 24);
		}
	}
	*/

	dInfo.pw = 640;
	dInfo.ph = 480;
	dInfo.showLabels = TRUE;
	
	if (!TextInit()) {
		fprintf(stderr, "Unable to initialize the text system\n");
		exit(1);
	}

	// Initialize SDL for video output
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK /* | SDL_INIT_AUDIO */
	/* | SDL_INIT_NOPARACHUTE */) < 0) {
		// failed to initalize SDL
		prnfile(stderr, strings[STRING_ERR_SDL], SDL_GetError());
		exit(2);
	}
	// Fetch the video info
	videoInfo = SDL_GetVideoInfo();
	
	if (!videoInfo)	{
		prnfile(stderr, strings[STRING_ERR_VIDQ], SDL_GetError());
		SDL_Quit();
		exit(3);
	}
	
	// the flags to pass to SDL_SetVideoMode
	videoFlags  = SDL_ANYFORMAT | SDL_RESIZABLE | SDL_DOUBLEBUF;
	if ((argc >= 2) && (strcmp(argv[1], "-f") == 0)) {
		videoFlags |= SDL_FULLSCREEN;
	}
	
	// This checks to see if surfaces can be stored in memory
	if (videoInfo->hw_available)
		videoFlags |= SDL_HWSURFACE;
	else
		videoFlags |= SDL_SWSURFACE;
	
	// This checks if hardware blits can be done
	//if (videoInfo->blit_hw)
		videoFlags |= SDL_HWACCEL; // always try
	
	// generate all the tank sprites
	GenTankSprites();
	
	// attempt to get a surface for the window's icon
	if ((winIcon = dInfo.display = SDL_CreateRGBSurface(SDL_SWSURFACE, 32, 32,
	32, 0xFF, 0xFF00, 0xFF0000, 0xFF000000)) != NULL) {
		Player tank;
		// set the color values to match the icon's color setup
		SetupColorVals();
		// setup values to draw a red tank facing up filling a 32x32 surface
		dInfo.luw = dInfo.luh = 4;
		tank.tank.loc.w = tank.tank.loc.h = dInfo.luw << 3;
		tank.tank.physLocX = -2;
		tank.tank.physLocY = 0;
		tank.rot = 0;
		tank.team = 0;
		// draw the sprite onto the icon surface
		DrawTank(winIcon, (RenderItem*)&tank);
		// set the window icon
		SDL_WM_SetIcon(winIcon, NULL);
	}
	// Create a 640x480 SDL surface & window
	if ((dInfo.display = SDL_SetVideoMode(dInfo.pw, dInfo.ph, 32, videoFlags))
	== NULL) {
		prnfile(stderr, strings[STRING_ERR_VIDMODE], SDL_GetError());
		SDL_Quit();
		exit(4);
	}
	// if the window icon was created . . .
	if (winIcon) {
		// free it -- it was copied by SDL, so no reason to keep it
		SDL_FreeSurface(winIcon);
	}

	// setup the color values to match the display
	SetupColorVals();
	
	// show information on the actual display surface
	if (dInfo.display->flags & SDL_DOUBLEBUF) {
		// double buffered video
		prnout(strings[STRING_STAT_DOUBLEBUFF]);
		// No mouse! Can cause problems with double buffered graphics.
		SDL_ShowCursor(SDL_DISABLE);
	}
	else {
		// single buffered video
		prnout(strings[STRING_STAT_SINGLEBUFF]);
	}

	if (dInfo.display->flags & SDL_HWACCEL) {
		// hardware accelerarion
		prnout(strings[STRING_STAT_HWACCEL]);
	}
	else {
		// all on CPU
		prnout(strings[STRING_STAT_NOHWACCEL]);
	}
	
	if (dInfo.display->flags & SDL_HWSURFACE) {
		// output surface in graphics memory
		prnout(strings[STRING_STAT_OUTGRPMEM]);
	}
	else {
		// output surface in system memory
		prnout(strings[STRING_STAT_OUTSYSMEM]);
	}
	
	{
		char driver[32];
		SDL_VideoDriverName(driver, 32);
		// show the video drive's name
		prnout(strings[STRING_STAT_VIDDRNAME], driver);
	}
	
	// Initalize the true type font engine
	if (TTF_Init() != 0) {
		// font library failed to initialize
		prnfile(stderr, strings[STRING_ERR_TTFINIT], TTF_GetError());
		SDL_Quit();
		exit(5);
	}
	
	// initalize the rendering system
	if (!RenderInit(128, (dInfo.display->flags & SDL_DOUBLEBUF) ? 2 : 1)) {
		// failed to initialize rendering
		prnfile(stderr, strings[STRING_ERR_RENDER]);
		SDL_Quit();
		exit(6);
	}
	
	// initalize the graphical part of the text system
	if (!FontInit()) {
		// failed to load fonts
		prnfile(stderr, strings[STRING_ERR_FNTINIT]);
		SDL_Quit();
		exit(7);
	}
	
	// initalize joystcks, if any
	if (SDL_NumJoysticks() > 0) {
		// open the first joystck and check for success
		if ((stick = SDL_JoystickOpen(0)) != 0) {
			// show joystick info
			prnout(strings[STRING_STAT_JOYSTICK],
			SDL_JoystickName(0), SDL_JoystickNumAxes(stick),
			SDL_JoystickNumButtons(stick));
			SDL_JoystickEventState(SDL_ENABLE);
		}
	}
	
	// Set the title bar in environments that support it.
	TextSetWindowTitle();
		
	#ifdef USE_AUDIO
	if (!InitAudio()) {
		// no audio output
		prnout(strings[STRING_STAT_NOAUDIO]);
		UninitAudio();
	}
	#endif

	// check for a command line argument with the hostname of the server to
	// use
	if ((argc >= 3) && (strcmp(argv[1], "-c") == 0)) {
		// connect to the specified server
		NetInit(argv[2]);
	}
	else if ((argc >= 2) && (strcmp(argv[1], "-s") == 0)) {
		// be a server
		NetInitSrv();
	}
	else {
		// discover a server on the LAN
		NetInit(NULL);
	}
	
	// start of state loop notification
	prnout(strings[STRING_STAT_ENTERSTATE]);
	
	// setup the initial frame timer state
	elapsedTime = SDL_GetTicks();
	frameTime = elapsedTime - (elapsedTime % FRAME_DURATION) +
	FRAME_DURATION;
	elapsedTime %= FRAME_DURATION;
	
	// run the states until finished
	for (; RunState(); ) {
		
		#ifndef RENDER_EVERY_FRAME
		// check for on-time execution
		if (elapsedTime <= FRAME_DURATION) {
		#endif
	
			Render();
			
			// see how much time was used
			elapsedTime = SDL_GetTicks() - frameTime;
			// adjust for weirdness
			if (elapsedTime < 0) {
				elapsedTime = 0;
			}
	
		#ifndef RENDER_EVERY_FRAME
			// wait
			if (elapsedTime < FRAME_DURATION) {
				SDL_Delay(FRAME_DURATION - elapsedTime);
			}
		}
		else {
			// record that we are closer to being at the correct time
			// index
			elapsedTime -= FRAME_DURATION;
		}
		#else
		SDL_Delay(FRAME_DURATION -
		((elapsedTime > FRAME_DURATION) ? FRAME_DURATION : elapsedTime));
		#endif
		// advance the time reference once for this trip through the event loop
		frameTime += FRAME_DURATION;
	}
	
	//rtsa_char *testStr = strings[STRING_STAT_EXITSTATE];
	// notification that the program has left the loop and is shutting down
	//int written = 
	prnout(strings[STRING_STAT_EXITSTATE]);
	
	if (menuSettings) {
		free(menuSettings);
	}
	if (stick) {
		SDL_JoystickClose(stick);
	}
	RandUninit();
	#ifdef USE_AUDIO
	UninitAudio();
	#endif
	StopIOThread();
	NetUninit();
	TextUninit();
	RenderUninit();
	#ifdef WIN32
	WSACleanup();
	#ifndef USE_GETTEXT
	if (haveCon) {
		FreeConsole();
	}
	#elif defined(_MSC_VER) && (_MSC_VER > 1310)
	fclose(stdout);
	*stdout = oldout;
	*stderr = olderr;
	#endif
	#endif
	TTF_Quit();
	SDL_Quit();
	return 0;
}

