/**
 * @file menu.c
 * Implements the menu system.
 * @ingroup MenuSys
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "menu.h"

#ifdef WIN32
// Avoid a name conflict with a function in the win32 API
#undef  RemoveMenu
#define RemoveMenu  _RemoveMenu_win32_
#endif

#include "genconmac.h"
#include "color.h"
#include "states.h"
#include "message.h"
#include "net.h"
#include <stdlib.h>
#include <assert.h>

#ifdef WIN32
// Avoid multiply defined symbols
#undef  RemoveMenu
#define RemoveMenu _RemoveMenu_tank_
#endif


/**
 * The margin between the sides of the window and the menu.
 */
#define MENU_MARGIN  96

/**
 * The thickness of the focus markers in pixels
 */
#define MARKER_THICKNESS  3

/**
 * The length of the focus markers in pixels
 */
#define MARKER_LENGTH  24

extern RenderItem background;

/**
 * @internal
 * A pointer to the start of an array of items that will comprise the menu.
 * @ingroup MenuSys
 */
static MenuItem *items;

/**
 * @internal
 * The index of the menu item that has input focus.
 * @ingroup MenuSys
 */
static int focus;

/**
 * @internal
 * The number of elements in the items array.
 * @ingroup MenuSys
 */
static int numItems;

/**
 * @internal
 * Used for the rectangle edge markers around the menu item with input focus.
 */
static RenderItem *markers;

static void PlaceMarkers() {
	int loop;
	RenderItem *mark;
	for (mark = markers, loop = 8; loop > 0; mark++, loop--) {
		// on the left
		if (loop & 2) {
			mark->loc.x = items[focus].region.x - MARKER_THICKNESS;
		}
		// on the right
		else {
			mark->loc.x = items[focus].region.x + items[focus].region.w
			+ MARKER_THICKNESS - mark->loc.w;
		}
		// on the top
		if (loop & 4) {
			mark->loc.y = items[focus].region.y - MARKER_THICKNESS;
		}
		// on the bottom
		else {
			mark->loc.y = items[focus].region.y + items[focus].region.h
			+ MARKER_THICKNESS - mark->loc.h;
		}
		// make the mark visible
		PlaceRenderItem(mark);
	}
}

/**
 * Attempts to resize the menu items to fit the requested size window.
 * @param  winW  The new width of the window.
 * @param  winH  The new height of the window.
 * @return True if the menu items could fit and were resized. False if the
 *         items could not fit and were not altered.
 * @author Jeff Jackowski
 */
static Bool ResizeMenu(int winW, int winH) {
	MenuItem *item;
	int loop;
	int w, h = MARKER_THICKNESS;  // stores the proposed menu size
	int y;         // the y coordinate of the last placed item
	int x;         // the new x coordinate for all items
	// propose value for w
	if ((w = winW - MENU_MARGIN * 2) < (MENU_MARGIN * 2)) {
		// inadequate size -- do not resize
		return FALSE;
	}
	// loop to find good values for the width and height before modifying
	// the menu items
	for (item = items, loop = numItems; loop > 0; item++, loop--) {
		// all menu items must have an update function
		assert(item->update != NULL);
		// all menu items must have their minimum sizes set
		assert((item->minWidth > 0) && (item->minHeight > 0));
		// check for inadequate width
		if (item->minWidth > (winW - MARKER_THICKNESS * 2)) {
			// use maximum width
			w = winW - MARKER_THICKNESS * 2;
		}
		// check for a need for more width
		if (item->minWidth > w) {
			// record the greater width
			w = item->minWidth;
		}
		// accumulate the height
		h += item->minHeight + MARKER_THICKNESS;
	}
	// compute the x coordinate
	x = (winW - w) >> 1;
	// check for a height that won't fit
	if (h > (loop = winH - TOP_MARGIN - BOTTOM_MARGIN * 2 - 
	textItems[TEXTITEM_TITLE].loc.h)) {
		// remove the title's hieght
		loop += textItems[TEXTITEM_TITLE].loc.h + TOP_MARGIN - BOTTOM_MARGIN;
		// don't show the title
		ClearRenderItem(&(textItems[TEXTITEM_TITLE]));
		// see if it will fit with the normal bottom margin
		if (h <= loop) {
			// set the height with the bottom margin up from the bottm
			y = winH - h - BOTTOM_MARGIN;
		}
		else if ((y = winH - (h >> 1)) < 0) {
			y = 0;
		}
		// no remaining height
		h = 0;
	}
	// the items have adequate vertical room
	else {
		// place the first item below the title
		y = TOP_MARGIN + BOTTOM_MARGIN + textItems[TEXTITEM_TITLE].loc.h;
		// compute the remaining height
		h = winH - y - BOTTOM_MARGIN - h;
		// assure the title is drawn at the correct location
		textItems[TEXTITEM_TITLE].loc.x =
		(winW >> 1) - (textItems[TEXTITEM_TITLE].loc.w >> 1);
		PlaceRenderItem(&(textItems[TEXTITEM_TITLE]));
	}
	// loop through the items to modify their size
	for (item = items, loop = numItems; loop > 0; item++, loop--) {
		// set x
		item->region.x = x;
		// set y
		item->region.y = y;
		// set width
		item->region.w = w;
		// set height
		item->region.h = item->minHeight;
		// check for expandable height
		if (item->expandHeight) {
			// give it the the remaining height, too
			item->region.h += h;
		}
		// advance the y coordinate for the next item
		y += item->region.h + MARKER_THICKNESS;
		// clear the placed flag
		item->isPlaced = FALSE;
	}
	// update the markers
	PlaceMarkers();
	// resize successful
	return TRUE;
}

void SetupMenu(MenuItem *menu, int first, int count) {
	int loop;
	RenderItem *mark;
	// assure the item that had input focus initially is within the array of
	// items
	assert((first >= 0) && (first < count));
	// setup the internal data
	items = menu;
	numItems = count;
	focus = first;
	// set the focus flag of the focused item
	items[focus].hasFocus = TRUE;
	// put the backgound image in place
	background.renderer = SolidFillRenderer;
	background.data = &(colorVals[COLOR_REGBACK]);
	background.loc.w = dInfo.pw;
	background.loc.h = dInfo.ph;
	AddRenderItem(&background, LAYER_BACKGROUND);
	PlaceRenderItem(&background);
	// place the title text in the top; it will be cenetered by resize
	textItems[TEXTITEM_TITLE].loc.y = TOP_MARGIN;
	// initalize the markers
	markers = calloc(8, sizeof(RenderItem));
	for (mark = markers, loop = 8; loop > 0; mark++, loop--) {
		// tall marker?
		if (loop & 1) {
			mark->loc.w = MARKER_THICKNESS;
			mark->loc.h = MARKER_LENGTH;
		}
		// short marker
		else {
			mark->loc.w = MARKER_LENGTH;
			mark->loc.h = MARKER_THICKNESS;
		}
		// set the render function
		mark->renderer = SolidFillRenderer;
		// set the color
		mark->data = &(colorVals[COLOR_OBSTACLE]);
		// put the mark in a layer
		AddRenderItem(mark, LAYER_OBSTACLES);
	}
	// size the menu items
	ResizeMenu(dInfo.pw, dInfo.ph);
	// get key Unicode values (slightly broken -- assumes all chars are 2 bytes)
	SDL_EnableUNICODE(TRUE);
}

void RemoveMenu() {
	int loop;
	// get rid of the markers
	for (loop = 0; loop < 8; loop++) {
		RemoveRenderItem(&(markers[loop]));
	}
	// free up the memory for the markers
	free(markers);
	RemoveRenderItem(&background);
	// no longer need the key values, so avoid the overhead
	SDL_EnableUNICODE(FALSE);
}

int MenuRun() {
	int loop, state = 0;
	MenuItem *item;
	SDL_Event event;
	rtsa_char keyval = 0;
	
	// clear the event storing a typed character
	eventState[EC_FORWARD] = eventState[EC_BACKWARD] = 0;
	// loop through all waiting events
	while (SDL_PollEvent(&event)) {
		// figure event type
		switch (event.type) {
			case SDL_QUIT:
				return STATE_QUIT;
				break;
			case SDL_ACTIVEEVENT:
				// some strange focus thing
				/*
				if (event.active.gain == 0)
					eventState[EC_INACTIVE] = TRUE;
				else
					eventState[EC_INACTIVE] = FALSE;
				*/
				break;
			case SDL_KEYDOWN:
				// handle key presses
				switch (event.key.keysym.sym) {
					case SDLK_ESCAPE:
						return STATE_QUIT;
						break;
					case SDLK_UP:
						eventState[EC_FORWARD] |= 1;
						break;
					case SDLK_DOWN:
						eventState[EC_BACKWARD] |= 1;
						break;
					case SDLK_TAB:
						if (event.key.keysym.mod & KMOD_SHIFT) {
							eventState[EC_FORWARD] |= 1;
						}
						else {
							eventState[EC_BACKWARD] |= 1;
						}
						break;
					case SDLK_LEFT:
						eventState[EC_ROTLEFT] |= 1;
						break;
					case SDLK_RIGHT:
						eventState[EC_ROTRIGHT] |= 1;
						break;
					case SDLK_SPACE:
						keyval = ' ';
						eventState[EC_CHARACTER] |= 1;
					case SDLK_RETURN:
						eventState[EC_FIRE] |= 1;
						break;
					default:
						#ifdef USE_GETTEXT
						eventState[EC_CHARACTER] |= 1;
						// store the key code
						keyval = event.key.keysym.unicode;
						// modify for upper/lower case
						/*  was once useful, but I *think* it isn't needed now
						if (event.key.keysym.mod & KMOD_SHIFT) {
							eventState[EC_CHARACTER] -= 'a' - 'A';
						}
						*/
						#else
						// only accept keys that fit within ASCII
						if ((event.key.keysym.unicode & 0xFF80) == 0) {
							eventState[EC_CHARACTER] |= 1;
							keyval = (rtsa_char)event.key.keysym.unicode;
						}
						#endif
						break;
				}
				break;
			case SDL_KEYUP:
				switch (event.key.keysym.sym) {
					case SDLK_LEFT:
						eventState[EC_ROTLEFT] = 0;
						break;
					case SDLK_RIGHT:
						eventState[EC_ROTRIGHT] = 0;
						break;
					case SDLK_SPACE:
						eventState[EC_CHARACTER] = 0;
					case SDLK_RETURN:
						eventState[EC_FIRE] = 0;
						break;
					case SDLK_BACKSPACE:
						return STATE_SPLASH;
					default:
						if (event.key.keysym.unicode) {
							eventState[EC_CHARACTER] = 0;
						}
						break;
				}
				break;
			case SDL_JOYAXISMOTION:
				if (!(SDL_GetAppState() & SDL_APPINPUTFOCUS)) break;
				// x-axis
				if (event.jaxis.axis == 0) {
					// right
					if (event.jaxis.value > 16384) {
						eventState[EC_ROTRIGHT] |= 1;
						eventState[EC_ROTLEFT] = 0;
					}
					// left
					else if (event.jaxis.value < -16384) {
						eventState[EC_ROTLEFT] |= 1;
						eventState[EC_ROTRIGHT] = 0;
					}
					// center
					else {
						eventState[EC_ROTLEFT] = 0;
						eventState[EC_ROTRIGHT] = 0;
					}
				}
				// y-axis
				else if (event.jaxis.axis == 1) {
					// down
					if (event.jaxis.value > 16384) {
						eventState[EC_BACKWARD] |= 1;
						eventState[EC_FORWARD] = 0;
					}
					// up
					else if (event.jaxis.value < -16384) {
						eventState[EC_FORWARD] |= 1;
						eventState[EC_BACKWARD] = 0;
					}
					// center
					else {
						eventState[EC_FORWARD] = 0;
						eventState[EC_BACKWARD] = 0;
					}
				}
				break;
			case SDL_JOYBUTTONDOWN:
				if (!(SDL_GetAppState() & SDL_APPINPUTFOCUS)) break;
				eventState[EC_FIRE] |= 1;
				break;
			case SDL_JOYBUTTONUP:
				if (!(SDL_GetAppState() & SDL_APPINPUTFOCUS)) break;
				eventState[EC_FIRE] = 0;
				break;
			case SDL_VIDEORESIZE:
				// attempt to resize the menu
				if (ResizeMenu(event.resize.w, event.resize.h)) {
					// attempt to resize the window (shouldn't ever fail)
					ResizeWindow(event.resize.w, event.resize.h);
					// assure good color values
					SetupColorVals();
					// put the backgound image into its new place
					background.loc.w = dInfo.pw;
					background.loc.h = dInfo.ph;
					PlaceRenderItem(&background);
				}
			default:
				break;
		}
	}
	// ----- run network code ----
	ProcessNetworkMessages();
	// see if a focus change is possible
	if ((numItems > 1) && !items[focus].lockFocus) {
		// check for the down/next item input
		if (eventState[EC_BACKWARD]) {
			// loop through the next items
			for (loop = focus + 1; loop < numItems; loop++) {
				// check for one that can aquire the focus
				if (items[loop].active) {
					// the old focus is lost
					items[focus].hasFocus = FALSE;
					if (items[focus].focus) {
						(*(items[focus].focus))(&(items[focus]));
					}
					// set the focus to this one
					items[focus = loop].hasFocus = TRUE;
					if (items[focus].focus) {
						(*(items[focus].focus))(&(items[focus]));
					}
					// update the marker locations
					PlaceMarkers();
					break;
				}
			}
		}
		// check for the up/previous item input
		else if (eventState[EC_FORWARD]) {
			// loop through the previous items
			for (loop = focus - 1; loop >= 0; loop--) {
				// check for one that can aquire the focus
				if (items[loop].active) {
					// the old focus is lost
					items[focus].hasFocus = FALSE;
					if (items[focus].focus) {
						(*(items[focus].focus))(&(items[focus]));
					}
					// set the focus to this one
					items[focus = loop].hasFocus = TRUE;
					if (items[focus].focus) {
						(*(items[focus].focus))(&(items[focus]));
					}
					// update the marker locations
					PlaceMarkers();
					break;
				}
			}
		}
	}
	// pass the input on to the focused item
	if (items[focus].input) {
		// run input processing and check for a state change request
		if ((state = (*(items[focus].input))(&(items[focus]), eventState,
		keyval)) != 0) {
			return state;
		}
	}
	// update all the menu items
	for (item = items, loop = numItems; loop > 0; item++, loop--) {
		// call the update
		if ((state = (*(item->update))(item)) != 0) {
			return state;
		}
	}
	// stay in this state
	return STATE_MENUROOT;
}

