/**
 * @file menu.h
 * Defines the data type and functions used by the menu system.
 * @ingroup MenuSys
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#ifndef _menu_h_
#define _menu_h_

#ifdef WIN32
// Avoid a name conflict with a function in the win32 API
#define RemoveMenu  _RemoveMenu_win32_
#endif

#include "render.h"
#include "text.h"

#ifdef WIN32
// Avoid multiply defined symbols
#undef RemoveMenu
#define RemoveMenu _RemoveMenu_tank_
#endif

/**
 * @defgroup MenuSys Menu System
 * A generalized system for opertating simple menus.
 *
 * @{
 */

typedef struct MenuItem_t  MenuItem;

/**
 * Responds to a change in focus for an item. An item will be notified about
 * losing focus before another item is notified about gaining focus. When a
 * menu is initalized, the item that has focus will receive a call to this
 * function. Check MenuItem::hasFocus for the new focus status of the item.
 * @param item  The MenuItem to operate upon.
 */
typedef void (*MenuItemFocus)(MenuItem *item);

/**
 * Responds to player input for the menu item. Called once per frame only on
 * the item with input focus.
 * @warning  If the player presses the space bar, @a keyval will be
 *           set to the character value for a space, <I>and</I> the flag for
 *           the fire event will be set.
 * @param item  The MenuItem to operate upon.
 * @param ec    A pointer to event codes that describe the player's input.
 * @param keyval  The character value of a key pressed, or zero. Generally,
 *                the values in @a ec should be inspected first.
 * @return    If non-zero, the operating state (::STATE_CODES) that should
 *            run next. If zero, no state change will occur.
 */
typedef int (*MenuItemInput)(MenuItem *item, Uint8 *ec, rtsa_char keyval);

/**
 * Called once per frame to update the graphics on a menu item. This is called
 * after the focus change function and after the input function. The function
 * is responsible for placing the menu item's render items when the flag
 * MenuItem::isPlaced is clear, which can occur when the program's window is
 * resized, and when the update function is called for the first time on a
 * given menu item.
 * @param item  The MenuItem to operate upon.
 * @return      If non-zero, the operating state (::STATE_CODES) that should
 *              run next. If zero, no state change will occur.
 */
typedef int (*MenuItemUpdate)(MenuItem *item);

/**
 * Represents a user interface component on a menu.
 * @author  Jeff Jackowski
 */
struct MenuItem_t {
	/**
	 * The minimum width and height the menu item requires to draw itself
	 * properly.
	 */
	int minWidth, minHeight;
	/**
	 * Defines the region over which this menu item is drawn. Used by the menu
	 * system to place markers around the menu item to denote it has input
	 * focus.
	 */
	SDL_Rect region;
	/**
	 * The visible things that are related to this menu item. This is an
	 * array.
	 */
	RenderItem *graphics;
	/**
	 * The number of items in graphics.
	 */
	Uint32 numGraphics;
	/**
	 * The function used to modify the menu item when it gains and loses focus.
	 * The value may be NULL to indicate that there is no function to call.
	 */
	MenuItemFocus focus;
	/**
	 * The function that responds to user input for the menu item.
	 * The value may be NULL to indicate that there is no function to call.
	 */
	MenuItemInput input;
	/**
	 * The function that keeps the display of the menu item up to date.
	 * This function must be defined so that after window resizes and
	 * resolution changes the menu item will be drawn in the proper place.
	 */
	MenuItemUpdate update;
	union {
		/**
		 * A set of flags. Only a few are defined by the menu system; the
		 * remaining bits may be used for user-defined states. The default
		 * value of flags is zero.
		 */
		Uint32 flags;
		struct {
			/**
			 * When true, the menu item may be given the input focus.
			 */
			Bool active      : 1;
			/**
			 * When true, the item has the input focus.
			 */
			Bool hasFocus    : 1;
			/**
			 * When true, the player is not allowed to switch focus away from
			 * this item. Has no effect when this item does not have input
			 * focus.
			 */
			Bool lockFocus   : 1;
			/**
			 * Enables expanding the height of the menu item when resized to
			 * fill the availble height.
			 * @warning  Only one item on a menu can have this flag set.
			 */
			Bool expandHeight : 1;
			/**
			 * True when the menu item has placed all of its graphics within
			 * its assigned region. This flag is cleared when the window is
			 * resized. The menu item's update funtion (see MenuItemUpdate) is
			 * responsible for placing the render items and clearing this flag.
			 */
			Bool isPlaced     : 1;
		};
	};
	/**
	 * Additional application specific information about this option.
	 */
	void *data;
};

/**
 * Initalizes the menu system's internal data for a new menu.
 * @param menu  A pointer to an array of menu items that will make up the menu.
 *              The items must already be initalized with correct data,
 *              all RenderItems must already be added to a layer, and the
 *              RenderItems should already be placed if they need to be
 *              visible. When the player provides a down input, the focus will
 *              move to an item with a greater index.
 * @param first The index of the menu item that will be initially given input
 *              focus.
 * @param count The number of items in the menu array.
 * @author  Jeff Jackowski
 */
void SetupMenu(MenuItem *menu, int first, int count);

/**
 * Removes intenal menu system data and prepares for a new menu.
 * @author  Jeff Jackowski
 */
void RemoveMenu();

/**
 * Runs a menu operating state. The corresponding initalization and
 * uninitalization must be cusotm made for each menu.
 * @pre     textItems[::TEXTITEM_TITLE] has been initalized to the
 *          game title. It will be displayed above the menu.
 * @return  The next operating state to run.
 * @author  Jeff Jackowski
 */
int MenuRun();

/** @} */

#endif

