/**
 * @file menustate.h
 * Defines the functions used for STATE_CODES::STATE_MENUROOT.
 * @ingroup StateMenu
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "player.h"

/**
 * @defgroup StateMenu Menu States
 *
 * Implements the @ref MenuRoot and @ref MenuNet.
 *
 * @ingroup OperStates
 * @{
 */

/**
 * A collection of information entered by the user at the menu that is moved
 * to other data structures for gameplay. None of the values are finalized
 * until the user decides to start playing.
 */
struct MenuSettings_t {
	/**
	 * The tank icons used to denote the teams for selection in the 
	 * @ref MenuTeamSel.
	 */
	Player teams[TEAM_MAX];
	/**
	 * The selected color for each team. This should be the regular team color
	 * used during gameplay.
	 */
	int selColor[TEAM_MAX];
	/**
	 * The deselected color for each team. This should make the tank and text
	 * visible against the background color, ::COLOR_REGBACK, and provide a
	 * ghostly appearance.
	 */
	int deselColor[TEAM_MAX];
	/**
	 * The team ID the user is currently selecting to join.
	 */
	int team;
	/**
	 * An opaque field containing the name the user will have in the game.
	 * Instead of accessing this field directly, use GetLocalPlayerName() and
	 * SetLocalPlayerName().
	 */
	rtsa_char localPlayerName[MAX_NAME_LEN];
};

typedef struct MenuSettings_t MenuSettings;

/**
 * The settings enetered by the user that are maintained until the game state
 * begins but not during the game state. The pointer must be NULL when
 * deallocated.
 */
extern MenuSettings *menuSettings;

/**
 * Initializes the menu items used to make the @ref MenuRoot.
 * @todo    Load any persistant game settings.
 * @post    If successful, static menu text has been rendered and
 *          ::menuSettings points to allocated memory and is initalized to
 *           zero.
 * @return  True on sucess, false if there was an error.
 * @author  Jeff Jackowski
 */
int MenuRootInit();

/**
 * Removes the root menu.
 * @post    ::menuSettings is not deallocated so that its settings can be used
 *          in GameInit().
 * @todo    Save any persistant game settings.
 * @return  True on sucess, false if there was an error.
 * @author  Jeff Jackowski
 */
int MenuRootUninit();

/** @} */
