/**
 * @file message.c
 * Implementation of the @ref NetMsgSys.
 * @ingroup NetMsgSys
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

// sanity check
#if defined(MSG_INF_TRIES) && defined(NDEBUG)
#error Infinite message retries (MSG_INF_TRIES) may only be specified for debugging builds
#endif

#include "aatree.h"
#include "vector.h"
#include "game.h"
#include "gameconfig.h"
#include "playernet.h"
#include "net.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <SDL.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#endif

/**
 * The root of the AA-tree used to reference in-flight messages by ID number.
 * @internal
 */
static AANode *inflightId = AA_ROOT_NODE;
/**
 * The priotity queue that references in-flight messages sorted by the next
 * resend time.
 * @internal
 */
static Vector inflightTime;
/**
 * The vector of unused but allocated messages (MsgDescr).
 * @internal
 */
static Vector available;
/**
 * A vector of allocated but unused buffers that are for storing message
 * data passed to or recived from the network.
 * @internal
 */
static Vector buffers;
/**
 * A counter containing the ID number to use for the next transmitted message.
 * The tree used to store the messages gives special meaning to 0xFFFF (-1).
 * Because of this, the message ID must not be 0xFFFF.
 * @internal
 */
static Uint16 nextMessageId = 0;

extern Uint32 frameTime;

/**
 * The initial resend interval in milliseconds.
 * @todo  Make this 64ms plus round-trip latency.
 * @internal
 */
#define INIT_RESEND  192

/**
 * The initial number of allocated network buffers. More will be allocated as
 * needed.
 * @internal
 */
#define INIT_BUFFERS   16

/**
 * The initial number of allocated messages (MsgDescr). More will be allocated
 * as needed.
 * @internal
 */
#define INIT_MESSAGES 32

/**
 * The initial number of tries left (MsgDescr::triesLeft) for new messages.
 * @internal
 */
#define INIT_TRIES 4

/**
 * The minimum number of pointers to store in the ::inflightTime and
 * ::available vectors, and twice the minumum number of pointers used in
 * ::buffers. The value can, and probably should, be greater than the initial
 * number of allocated objects.
 * @internal
 */
#define MIN_POINTERS  64 // half size for buffers

/**
 * The maximum number of unallocated buffers that are maintained.
 * @internal
 */
#define MAX_BUFFERS   (Uint32)((networkState == NETSTATE_SERVER) ? 32 : 16)

/**
 * The maximum number of unallocated messages (MsgDescr) that are maintained.
 * @internal
 */
#define MAX_MESSAGES  (Uint32)((networkState == NETSTATE_SERVER) ? 128 : 64)

/**
 * A comparison function used with Vector to maintain a binary heap sorted by
 * resend time.
 * @internal
 * @param lesser   The message that will need to be resent sooner if this
 *                 function returns true.
 * @param greater  The message that will need to be resent later if this
 *                 function returns true.
 * @return  True if lesser needs to be resent before greater.
 * @author  Jeff Jackowski
 */
static Bool TimeCompare(MsgDescr **lesser, MsgDescr **greater) {
	return (*lesser)->nextSend < (*greater)->nextSend;
}

/**
 * A destructor function used to deallocate message descriptors; intended for
 * use with Vector.
 * @internal
 * @param msg  A double pointer to the message to destroy.
 * @author     Jeff Jackowski
 */
static void MsgDescrDestructor(MsgDescr **msg) {
	// check for an allocated buffer
	if ((*msg)->buffer != NULL) {
		// free the buffer
		BufferFree((*msg)->buffer);
	}
	// deallocate the message descriptor
	free(*msg);
	#ifndef NDEBUG
	// give the pointer a bogus value
	*msg = (MsgDescr*)0xCDCDCDCD;
	#endif
}

void CancelAllMessages() {
	// alter the message vectors so that they will free all the message data
	inflightTime.destructOp = (DestructorOp)MsgDescrDestructor;
	// remove the messages that describe in-flight messages
	if (inflightId != AA_ROOT_NODE) {
		// destroy the tree
		AATreeDestroy(inflightId);
		// keep a valid tree
		inflightId = AA_ROOT_NODE;
	}
	VectorClear(&inflightTime);
	// reset the destructors
	inflightTime.destructOp = NULL;
}

int GetNumOutgoingMessages() {
	return inflightTime.items;
}

MsgDescr *MessageNew() {
	MsgDescr *msg;
	int loop;
	// check for an available message
	if ((msg = (MsgDescr*)VectorLast(&available)) != NULL) {
		// pointer fun to get the real value
		msg = *(MsgDescr**)msg;
		// got a message from the available list; now it is not available
		available.items--;  // the shortened version of VectorPop()
	}
	// no available messages, so make a new one
	else {
		msg = malloc(sizeof(MsgDescr));
	}
	#ifndef NDEBUG
	// fill the message with crud to force proper initialization of the data
	memset(msg, 0xCD, sizeof(MsgDescr));
	#endif
	// initalize the flags and the interval
	msg->flags = 0;
	/**
	 * @todo  Set the initial resend time to twice the average two-way latency
	 *        of network commincations as determined by the time sync operation
	 *        with some constant minimum (32ms?).
	 */
	msg->interval = INIT_RESEND;
	msg->triesLeft = INIT_TRIES;
	// clear the destination map
	loop = DESTMAPLEN;
	do {
		msg->destMap[--loop] = 0;
	} while (loop > 0);
	// give back the message
	return msg;
}

void MessageFree(MsgDescr *msg) {
	#ifndef NDEBUG
	int loop;
	// grab the first message
	MsgDescr **ptr = (MsgDescr**)(available.array);
	assert(msg != NULL);
	// loop through all the messages
	for (loop = available.items; loop > 0; loop--, ptr++) {
		// assure that the freed buffer is not already in the list (double free)
		assert(msg != *ptr);
	}
	#endif
	// check for an allocated buffer that needs to be reclaimed
	if ((msg->buffer != NULL) && !msg->buffLifeUnmanaged) {
		// free the buffer
		BufferFree(msg->buffer);
	}
	// check for a full list of available messages
	if (available.items >= MAX_MESSAGES) {
		// deallocate the message descriptor
		free(msg);
	}
	else {
		// clear the buffer pointer
		msg->buffer = NULL;
		// put it at the end of the available list
		VectorPush(&available, &msg);
	}
}

NetBuffer *BufferNew() {
	long *spot;
	NetBuffer *buff;
	// try to get an already allocated buffer from the list of buffers
	if ((spot = (long*)VectorLast(&buffers)) == NULL) {
		// no free buffers, so make one
		if ((buff = (NetBuffer*)malloc(sizeof(NetBuffer))) == NULL) {
			// no memory
			return NULL;
		}
	}
	else {
		// Adjust buff to point to the buffer and not a spot within the vector.
		// The buffers themselves must never move, so they cannot be within
		// the vector.
		buff = *(NetBuffer**)spot;
		// clear the item in the vector
		*spot = 0; //NULL;
		// remove the buffer from the list
		buffers.items--;
	}
	// initialize the length to show no data
	buff->length = 0;
	return (NetBuffer*)buff;
}

void BufferFree(NetBuffer *buff) {
	#ifndef NDEBUG
	int loop;
	// grab the first buffer
	NetBuffer **ptr = (NetBuffer**)(buffers.array);
	assert(buff != NULL);
	// loop through all the buffers
	for (loop = buffers.items; loop > 0; loop--, ptr++) {
		// assure that the freed buffer is not already in the free list
		assert(buff != *ptr);
	}
	#endif
	// check for a full list of available messages
	if (buffers.items >= MAX_BUFFERS) {
		// deallocate the message
		free(buff);
	}
	else {
		// put it at the end of the available list
		VectorPush(&buffers, &buff);
	}
}

void BufferVecDestr(void *buff) {
	BufferFree(*(NetBuffer **)buff);
}

extern int sock;

/*
 * Constructs a message in a buffer in preperation for sending it over the
 * network.
 * @internal
 * @deprecated
 * @dot
 * digraph ConstrFlow {
 *   node [shape=record, fontname=Helvetica, fontsize=7, height=0.1];
 *   edge [fontname=Helvetica, fontsize=7];
 *   Start [ label="Start", shape="Mrecord" ];
 *   BuffCheck [ label="Allocated buffer with message?", shape="diamond",
 *     URL="\ref MsgDescr::buffer" ];
 *   Start -> BuffCheck;
 *   BufferObt [ label="Obtain a buffer" ];
 *   BuffCheck -> BufferObt [ label="No" ];
 *   ErrChk [ label="Success getting buffer?", shape="diamond" ];
 *   BufferObt ->ErrChk;
 *   Constr [ label="Construct the message by calling constrFunc",
 *     URL="\ref MsgDescr::constrFunc" ];
 *   ErrChk -> Constr [ label="Yes" ];
 *   BuffCheck -> Constr [ label="Yes" ];
 *   ConstrErrChk [ label="Was there an error?", shape="diamond" ];
 *   Constr -> ConstrErrChk;
 *   SetFlag [ label="Set the in-flight flag" ];
 *   ConstrErrChk -> SetFlag [ label="No" ];
 *   End [ label="End", shape="Mrecord" ];
 *   SetFlag -> End;
 *   EndErr [ label="End with error", shape="Mrecord" ];
 *   ConstrErrChk -> EndErr [ label="Yes" ];
 *   ErrChk -> EndErr [ label="No" ];
 * }
 * @enddot
 * @param msg  The message descriptor for the message to construct.
 * @return     True on success, false if either a buffer could not be allocated
 *             or the message's constructor failed to build the message.
 * @ingroup    NetMsgSys
 * @author     Jeff Jackowski
 */
 /*
static Bool ConstructMessage(MsgDescr *msg) {
	// if there is no persistant buffer, or one has not been set,
	// obtain a buffer from the free list, if possible
	if (msg->buffer == NULL) {
		// if no buffer is set, there must be a constructor function
		assert(msg->constrFunc != NULL);
		// get a buffer from the list and check for failure
		if ((msg->buffer = BufferNew()) == NULL) {
			// no memory; can't send
			return FALSE;
		}
	}
	// construct the message
	if (msg->constrFunc) {
		(*(msg->constrFunc))(msg);
	}
	// check for an error
	if (msg->buffer->length < 0) {
		// cancel the message
		return FALSE;
	}
	// set the in-flight flag
	msg->inflight = TRUE;
	// success!
	return TRUE;
}
*/

/**
 * Transmits the given message to the other end(s) and/or cancels the message.
 * @internal
 * @pre  The message is not in the ::inflightTime priority queue, but is in
 *       the ::inflightId tree.
 *
 * Flowchart:
 * @todo  Update the flowchart. Messages are now constructed in a buffer just
 *        before sending them the first time rather than at the start of this
 *        function. The construction now happens in ConstructMessage().
 *        Also added in functionality to call failure function on message when
 *        not acknowledged and to cancel message after a set number of retries.
 *        Implemented the functionality for the MsgDescr::noResend flag. Added
 *        message approvals to client-side sending.
 *
 * @dot
 * digraph SendFlow {
 *   node [shape=record, fontname=Helvetica, fontsize=7, height=0.1];
 *   edge [fontname=Helvetica, fontsize=7];
 *   Start [ label="Start", shape="Mrecord" ];
 *   ClOrSv [ label="Is this process a client?", shape="diamond",
 *     URL="\ref ::networkState" ];
 *   Start -> ClOrSv;
 *   SendToSv [ label="Send to the server" ];
 *   ClOrSv -> SendToSv [ label="Yes" ];
 *   SvLoop [ label="Loop through clients in destination list",
 *     URL="\ref MsgDescr::destMap" ];
 *   ClOrSv -> SvLoop [ label="No" ];
 *   SvBroadCheck [ label="Need to broadcast to this client?",
 *     shape="diamond" ];
 *   SvLoop -> SvBroadCheck;
 *   SvBroadCheck2 [ label="Broadcast it already?", shape="diamond" ];
 *   SvBroadCheck -> SvBroadCheck2 [ label="Yes" ];
 *   SvBroadCheck2 -> SvEndLoop [ label="Yes" ];
 *   SvApprove [ label="Approve sending the message",
 *     URL="\ref MsgDescr::approveFunc" ];
 *   SvUnapproved [ label="Remove client from destination list",
 *     URL="\ref MsgDescr::destMap" ];
 *   SvApprove -> SvUnapproved [ label="Not approved" ];
 *   SvBroadCheck2 -> SvApprove [ label="No" ];
 *   SvBroadcast [ label="Broadcast once" ];
 *   SvApprove -> SvBroadcast [ label="Broadcast" ];
 *   SvErrCheck [ label="Send error?", shape="diamond" ];
 *   SvBroadcast -> SvErrCheck;
 *   SvKill [ label="Remove client from destination list",
 *     URL="\ref MsgDescr::destMap" ];
 *   SvErrCheck -> SvKill [ label="Yes" ];
 *   SvFail [ label="Call failure function", URL="\ref MsgDescr::failFunc" ];
 *   SvKill-> SvFail;
 *   SvUnicast [ label="Unicast to client" ];
 *   SvBroadCheck -> SvApprove [ label="No" ];
 *   SvApprove -> SvUnicast [ label="Unicast" ];
 *   SvUnicast -> SvErrCheck;
 *   SvEndLoop [ label="More clients?", shape="diamond" ];
 *   SvUnapproved -> SvEndLoop;
 *   SvErrCheck -> SvEndLoop [ label="No" ];
 *   SvFail -> SvEndLoop;
 *   SvEndLoop -> SvLoop [ label="Yes" ];
 *   SvProcessCheck [ label="Sent to any client(s)?", shape="diamond" ];
 *   SvEndLoop -> SvProcessCheck [ label="No" ];
 *   Cancel [ label="Cancel message" ];
 *   SvProcessCheck -> Cancel [ label="No" ];
 *   ConstrErrCheck -> Cancel [ label="Yes" ];
 *   BuffKeepCheck [ label="Persistant buffer?", shape="diamond",
 *     URL="\ref MsgDescr::buffLifeUnmanaged" ];
 *   SendToSv -> BuffKeepCheck;
 *   SvProcessCheck -> BuffKeepCheck [ label="Yes" ];
 *   BuffKeep [ label="Remove buffer from available list" ];
 *   BuffKeepCheck -> BuffKeep [ label="Yes" ];
 *   BuffReclaim [ label="Reclaim the buffer" ];
 *   BuffKeepCheck -> BuffReclaim [ label="No" ];
 *   UpdateTime [ label="Update resend time", URL="\ref MsgDescr::nextSend" ];
 *   BuffKeep -> UpdateTime;
 *   BuffReclaim -> UpdateTime;
 *   End [ label="End", shape="Mrecord" ];
 *   UpdateTime -> End;
 *   Cancel -> End;
 * }
 * @enddot
 * @param msg   The message to transmit.
 * @param inSys True if the message has been added to the heap and tree
 *              tracking messages in the system. This value will be false when
 *              called from AddMessage() because the resend time will be set
 *              while in this function.
 * @return True on success, false if the message was not sent because of error
 *         or no recipients.
 * @ingroup NetMsgSys
 * @author Jeff Jackowski
 */
static Bool Send(MsgDescr *msg, Bool inSys) {
	Uint32 now = SDL_GetTicks();
	// ----- Transmit the message -----
	// for clients (easy way)
	if ((networkState == NETSTATE_RCVCFG) ||
	(networkState == NETSTATE_CLIENT) || (networkState == NETSTATE_CLIDISCONN)) {
		// check for an acknowledgement
		if (msg->destMap[0] == 0)
			// done with this message
			goto Send_cancel;
		// one less try; see if this is the last try
		if (msg->triesLeft == 0) {
			// failed -- no acknowledgement
			if (msg->failFunc != NULL) {
				(*(msg->failFunc))(msg, NULL);
			}
			goto Send_cancel;
		}
		// check for need to resend
		if (!msg->noResend) {
			// check for lack of approval to send this message
			if ((msg->approveFunc != NULL) && 
			!(*(msg->approveFunc))(msg, 0)) {
				// no approval -- cancel this message
				goto Send_cancel;
			}
			// send the message to the server and check for failure
			else if (sendto(sock, msg->buffer->data, msg->buffer->length, 0,
			&outgoingAddr.addr, outgoingAddr.length) != msg->buffer->length) {
				// failed to construct or send message
				if (msg->failFunc != NULL) {
					(*(msg->failFunc))(msg, NULL);
				}
				goto Send_cancel;
			}
			msg->inflight = TRUE;
		}
	}
	// for servers (complex way)
	else {
		ClientData *client = clients;
		DESTMAPTYPE *sendFlags = &(msg->destMap[0]), sendFlagMask;
		int mappos = 0;  // index into MsgDescr::destMap
		int cid = 0;     // the client ID
		int processedClients = 0;
		Bool bcastSent = FALSE;  // records if a broadcast message was sent
		for (; mappos < DESTMAPLEN; mappos++, sendFlags++) {
			// see if there are no destination clients
			if (*sendFlags == 0) {
				// advanve to the next block of clients
				cid += sizeof(DESTMAPTYPE) * 8;
				client += sizeof(DESTMAPTYPE) * 8;
				continue;
			}
			sendFlagMask = 1;
			// cid through all the clients
			for (; sendFlagMask && (cid < MAX_PLAYERS);
			client++, cid++, sendFlagMask <<= 1) {
				// check for a disconnecting client
				if (!client->contact || (client->disconn &&
				((now - client->disconnTime) >= 4096))) {
					// clear the flags
					client->flags = 0;
					// remove client from destination list
					*sendFlags &= ~sendFlagMask;
					// don't send to this client
					continue;
				}
				else if (client->disconn &&
				(msg->buffer->data[0] != MSG_TYPE_CONN_DISCONN))
					continue;
				// first check for an inactive client record or a client that
				// does not need the message
				if (!(*sendFlags & sendFlagMask) ||
				// avoid sending to self
				(localPlayer == cid) ||
				// next, check for a client that receives broadcasts, and if a
				// broadcast has already been sent
				(bcastSent && client->bcast)) {
					// no need to send to this client
					continue;
				}
				// see if this is the last chance for a client to acknowledge
				if (msg->triesLeft == 0) {
					// this client is a destination for this message, but did
					// not acknowledge the message, so call the failure handler
					if (msg->failFunc != NULL) {
						(*(msg->failFunc))(msg, client);
					}
					continue;
				}
				// for messages that can be re-transmitted . . .
				if (!msg->noResend) {
					// check for lack of approval to send this message
					if ((msg->approveFunc != NULL) &&
					!(*(msg->approveFunc))(msg, cid)) {
						// no approval -- remove client from destination list
						*sendFlags ^= sendFlagMask;
						// do not send
						continue;
					}
					// see if a broadcast should be sent to this client
					if (msg->bcast && client->bcast) {
						assert(outgoingAddr.ipv4.sin_addr.s_addr ==
						htonl(INADDR_BROADCAST));
						// send out the data and check for error
						if (sendto(sock, msg->buffer->data, msg->buffer->length,
						0, &outgoingAddr.addr,	outgoingAddr.length)
						!= msg->buffer->length) {
							// failed to send a message to this client
							goto Send_failure;
						}
						// flag that a broadcast was sent so it isn't
						// rebroadcast to other clients that will receive
						// broadcasts
						bcastSent = TRUE;
						msg->inflight = TRUE;
					}
					// a direct, unicast message needs to be sent to this client
					else {
						// set the socket to not broadcast
						// needed?
						// send out the data
						if (sendto(sock, msg->buffer->data, msg->buffer->length,
						0, &(client->addr.addr), client->addr.length)
						!= msg->buffer->length) {
							Send_failure:
							// Can't send to this client? Remove it from the
							// destiniation list
							*sendFlags ^= sendFlagMask;
							// failed to send a message to this client
							if (msg->failFunc != NULL) {
								(*(msg->failFunc))(msg, client);
							}
							// don't count sending this message
							continue;
						}
						msg->inflight = TRUE;
						// set the socket to broadcast
					}
				}
				// processed another client
				processedClients++;
			}
		}
		// check for no processed clients
		if (processedClients == 0) {
			// cancel this message
			goto Send_cancel;
		}
		#ifdef MSG_DEBUG
		else if (msg->destMap[0] <= 1) {
			printf("Message %i has no destination but went somewhere\n",
			msg->msgId);
		}
		#endif
	}
	// update the resend time
	msg->nextSend = SDL_GetTicks() + msg->interval;
	#ifdef MSG_INF_TRIES
	// check for what is normally the maximum number of attemps
	if (--msg->triesLeft == 0) {
		// do not let the value drop below zero
		msg->triesLeft++;
		// set the interval to a high value
		msg->interval = 16384;
	}
	else {
		// increase the resend time
		msg->interval <<= 1;
	}
	#else
	// one less attempt left
	msg->triesLeft--;
	// wait longer until the next send
	msg->interval <<= 1;
	#endif
	return TRUE;
	Send_cancel:
	// see if the buffer should not be maintained
	if (!msg->buffLifeUnmanaged) {
		// reclaim the buffer
		BufferFree(msg->buffer);
		msg->buffer = NULL;
	}
	if (inSys) {
		#ifdef MSG_DEBUG
		printf("Canceling message %i\n", msg->msgId);
		#endif
		// remove the message from the tree
		AATreeRemove(inflightId, msg->msgId);
	}
	MessageFree(msg);
	return FALSE;
}

Bool AddMessage(MsgDescr *msg) {
	AANode *node;
	// the message must include a buffer that is long enough to contain
	// something
	assert((msg->buffer != NULL) && (msg->buffer->length > 3));
	// assure the network state is valid for sending a message
	assert(networkState <= NETSTATE_SERVER);
	// assure that messages sent to the server have the correct destination
	assert((networkState != NETSTATE_CLIENT) || (msg->destMap[0] == 1));
	// assure there are tries left
	assert(msg->triesLeft > 0);
	// if the message needs an ID . . .
	if (!msg->doNotAddId) {
		// advance the next message ID -- 0xFFFF is an invalid value
		if (++nextMessageId == 0xFFFF) {
			// advance again
			nextMessageId++;
		}
		// set the message ID
		msg->msgId = nextMessageId;
		// write it into the buffer
		MSG_WRITE_16C(msg->buffer->data, 1, msg->msgId);
	}
	// first send out the message and check for success
	if (Send(msg, FALSE)) { 
		// put the message into the AA-tree
		node = AATreeAdd(&inflightId, msg->msgId);
		assert(node != NULL);
		node->data = msg;
		// put the message into the heap
		VectorHeapPush(&inflightTime, &msg);
		// success
		return TRUE;
	}
	// failed
	return FALSE;
}

void ServiceMessages() {
	MsgDescr *msg;
	Uint32 now = SDL_GetTicks();
	// assure the network state is valid for servicing messages
	assert((networkState >= NETSTATE_CONNECTING) ||
	(networkState <= NETSTATE_SERVER));
	// loop until all messages awaiting resend have been serviced
	while (1) {
		// check for no messages in the system or for messages not yet ready to
		// be resent
		if (((msg = VectorFirst(&inflightTime)) == NULL) ||
		(msg = *(MsgDescr**)msg, msg->nextSend > now)) {
			// nothing left to do
			return;
		}
		// take the message out of the priority queue
		VectorHeapPop(&inflightTime);
		#ifdef MSG_DEBUG
		printf("Send message %d, ", msg->msgId);
		
		{
			AANode *node;
			node = AATreeFind(inflightId, msg->msgId);
			if (node == NULL) {
				printf("***** Message not in tree *****\n");
			}
			else if (node->data != msg) {
				printf("***** Different message in tree, ID = %i *****\n",
				node->key);
			}
			else if (node->key != msg->msgId) {
				printf("***** Message in tree had different ID: %i *****\n",
				node->key);
			}
		}
		
		#endif
		// send it again and check for success
		if (Send(msg, TRUE)) {
			// put it back into the priority queue
			VectorHeapPush(&inflightTime, &msg);
			#ifdef MSG_DEBUG
			printf("keeping for later\n");
			#endif
		}
		#ifdef MSG_DEBUG
		else printf("done with it\n");
		#endif
	}
}

Bool ValidateClient(SocketAddr *origin, Uint8 clientId) {
	ClientData *client;
	// the client is not this client
	if ((clientId == localPlayer) ||
	#if (MAX_PLAYERS < 256)
	// assure that the client ID is within bounds
	(clientId > MAX_PLAYERS) ||
	#endif
	// see if the client is inactive
	(!(client = &(clients[clientId]))->contact) ||
	// check for non-matching address
	!EqualAddresses(origin, &(client->addr)))  {
		return FALSE;
	}
	// matching address
	return TRUE;
}

Bool ValidateServer(SocketAddr *origin) {
	// only validate a server when connected to one, and make sure the address
	// is the same now as when the connection was made
	if (((networkState == NETSTATE_RCVCFG) ||
	(networkState == NETSTATE_CLIENT) || (networkState == NETSTATE_CLIDISCONN)) &&
	EqualAddresses(origin, &outgoingAddr)) {
		// looks like the server
		return TRUE;
	}
	return FALSE;
}

/**
 * Handles incomming acknowledgment messages (::MSG_TYPE_ACK) by invoking
 * the acknowledgment function on the corresponding messages
 * (MsgDescr::ackFunc).
 * @internal
 * Flowchart:
 * @dot
 * digraph HandleAckFlow {
 *   node [shape=record, fontname=Helvetica, fontsize=7, height=0.1];
 *   edge [fontname=Helvetica, fontsize=7];
 *   Start [ label="Start", shape="Mrecord" ];
 *   GID [ label="Get client ID" ];
 *   Start -> GID;
 *   Validate [ label="Is the client valid?", URL="\ref ValidateClient()",
 *     shape="diamond" ];
 *   GID -> Validate;
 *   Loop [ label="Loop through acks" ];
 *   Validate -> Loop [ label="Yes" ];
 *   GetMsg [ label="Get next ack'ed message" ];
 *   Loop -> GetMsg;
 *   Exist [ label="Does message exist?", shape="diamond" ];
 *   GetMsg -> Exist;
 *   Exist -> Loop [ label="No" ];
 *   DestChk [ label="Is ack'ing client a dest for the message?",
 *     shape="diamond" ];
 *   Exist -> DestChk [ label="Yes" ];
 *   DestChk -> Loop [ label ="No" ];
 *   Call [ label="Call the message's ack function",
 *     URL="\ref MsgDescr::ackFunc" ];
 *   DestChk -> Call [ label ="Yes" ];
 *   Rmov [ label="Remove client from message's dest map",
 *    URL="\ref MsgDescr::destMap" ];
 *   Call -> Rmov;
 *   Rmov -> Loop;
 *   End [ label="End", shape="Mrecord" ];
 *   Validate -> End [ label="No" ];
 *   Loop -> End [ label="Done" ];
 * }
 * @enddot
 *
 * @param origin    The originating address of the message.
 * @param buff      The buffer that contains the message to process.
 * @param unused    A required but unused part of the function prototype.
 * @param noid      A required but unused part of the function prototype.
 * @return          Always false; never acknowledge acknowledgments.
 * @author  Jeff Jackowski
 */
static Bool HandleAck(SocketAddr *origin, NetBuffer *buff, Uint8 *unused,
Uint16 noid) {
	AANode *node;
	Uint8 *ptr = &(buff->data[3]);
	int loop;
	DESTMAPTYPE *map, flag;
	Uint16 msgId;
	Uint8 clientId;
	// get the client id
	clientId = buff->data[1];
	// validate the client or server
	if (((((networkState == NETSTATE_RCVCFG) ||
	(networkState == NETSTATE_CLIENT) || (networkState == NETSTATE_CLIDISCONN)) 
	&& !ValidateServer(origin)) || (((networkState == NETSTATE_SERVER) ||
	(networkState == NETSTATE_SRVDISCONN)) &&
	!ValidateClient(origin, clientId))) ||
	// validate the length of the message
	(buff->length != (3 + buff->data[2] * 2))) {
		return FALSE;        // invalid origin or length
	}
	// loop through the acks
	for (loop = buff->data[2]; loop > 0; loop--) {
		// read in the ID of the acknowledged message
		MSG_READ_16(ptr, msgId);
		// find the corresponding message descriptor
		if ((node = AATreeFind(inflightId, msgId)) != NULL) {
			#ifdef MSG_DEBUG
			if (node->key != msgId) {
				printf("***** Find got a node for a different key *****\n");
				printf("\tLooking for %i, got %i\n", msgId, node->key);
			}
			if (((MsgDescr*)(node->data))->msgId != msgId) {
				printf("***** Node's key differs from message ID *****\n");
				printf("\tNode key is %i, ID is %i\n", node->key,
				((MsgDescr*)(node->data))->msgId);
			}
			#endif
			// get the location of the destination flag
			map = &(((MsgDescr*)(node->data))->destMap[
			GetDestMapIndex(clientId)]);
			flag = GetDestMapFlag(clientId);
			// see if this client is on the destination list
			if (*map & flag) {
				// call the ack function
				if ((*(((MsgDescr*)(node->data))->ackFunc)) != NULL) {
					(*(((MsgDescr*)(node->data))->ackFunc))((MsgDescr*)
					(node->data), clientId);
				}
				// remove the client from the destination list
				*map &= ~flag;
				#ifdef MSG_DEBUG
				printf("Client #%d ack'ed message %d\n",
				clientId, msgId);
				#endif
			}
			#ifdef MSG_DEBUG
			else {
				int ind;
				DESTMAPTYPE bit;
				printf("Client #%d RE-ack'ed message %d\nDest: ",
				clientId, msgId);
				for (ind = 0; ind < DESTMAPLEN; ind++) {
					for (bit = 1; bit > 0; bit <<= 1) {
						if (((MsgDescr*)(node->data))->destMap[ind] & bit)
							putchar('1');
						else
							putchar('0');
					}
				}
				putchar('\n');
			}
			#endif
		}
		#ifdef MSG_DEBUG
		else {
			printf("Client #%d ack'ed CANCELED message %d\n",
			clientId, msgId);
		}
		#endif
	}
	// never acknowledge acknowledgments
	return FALSE;
}

/**
 * Handles an acknowledgment of connection acceptance by bringing a client into
 * the game. This is used only on the server.
 * @internal
 * @param msg  The descriptor for the connection acceptance message.
 * @param cid  The client's ID.
 * @author     Jeff Jackowski
 */
static void AckConnAccept(MsgDescr *msg, int cid) {
	// mark the client as active
	clients[cid].active = TRUE;
	// allow the same random ID to be reused
	clients[cid].randId = 0;
	// add the client to the config list
	AddToConfigList(&(clients[cid]));
	#ifndef NDEBUG
	// A debugging status message (console only) on the server that tells of a
	// newly connected client. Because this is only used in debug builds and
	// not on the GUI, translating it should be considered a low priority.
	// The first parameter (%1$d) is the ID number assigned to the client.
	// The second parameter (%2$d) is the broadcast capability flag which is
	// generally true for clients on the same LAN as the server.
	#ifdef USE_GETTEXT
	wprintf(L"Accepted new client, ID = %d, broadcast = %d\n", cid,
	clients[cid].bcast);
	#else
	printf("Accepted new client, ID = %d, broadcast = %d\n", cid,
	clients[cid].bcast);
	#endif
	#endif
}

/**
 * Handles failure of a client to respond to a connection acceptance message by
 * making available the spot that was reserved for the client. A connection
 * denial message is sent to the client. This is used only on the server.
 * @internal
 * @param msg     The descriptor for the connection acceptance message.
 * @param client  The information on the client that tried to connect.
 * @author        Jeff Jackowski
 */
static void FailConnAccept(MsgDescr *msg, ClientData *client) {
	// Send a denial message.
	char outbuff[5];
	// write the type ID
	outbuff[0] = MSG_TYPE_CONN_DENY;
	// write the random ID
	MSG_WRITE_32C(outbuff, 1, client->randId);
	// send it to the requestor and don't worry about errors
	sendto(sock, outbuff, 5, 0, &(client->addr.addr), client->addr.length);
	// mark the client as inactive
	client->flags = 0;
	#ifndef NDEBUG
	// A debugging status message (console only) on the server that provides
	// notification of a failed connection attempt. Because this is only used
	// in debug builds and not on the GUI, translating it should be considered
	// a low priority. The only parameter (%d) is the ID number assigned to the
	// client.
	#ifdef USE_GETTEXT
	wprintf(L"Attempted connection failed for client, ID = %d\n", client->cid);
	#else
	printf("Attempted connection failed for client, ID = %d\n", client->cid);
	#endif
	#endif
}

/**
 * Handles incomming connection request messages from clients. Finds an
 * available spot for a new client while checking for a previous connection
 * attempt by the same client. Used only on the server side.
 * @internal
 * Flowchart:
 * @dot
 * digraph AddMessageFlow {
 *   node [shape=record, fontname=Helvetica, fontsize=7, height=0.1];
 *   edge [fontname=Helvetica, fontsize=7];
 *   Start [ label="Start", shape="Mrecord" ];
 *   GetRand [ label="Read in the random ID" ];
 *   Start -> GetRand;
 *   Loop [ label="Loop through client array" ];
 *   GetRand -> Loop;
 *   ChkCli [ label="Is there an address and random ID match?",
 *     shape="diamond" ];
 *   Loop -> ChkCli;
 *   FirstSpot [ label="Is this client the first avalible spot?",
 *     shape="diamond" ];
 *   ChkCli -> FirstSpot [ label="No" ];
 *   FirstSpot -> Loop [ label="No" ];
 *   Record [ label="Record this client" ];
 *   FirstSpot -> Record [ label="Yes" ];
 *   Record -> Loop;
 *   Good [ label="Valid client, no errors?", shape="diamond" ];
 *   Loop -> Good [ label="Done" ];
 *   Deny [ label="Send the client a denial" ];
 *   Good -> Deny [ label="No" ];
 *   Reserve [ label="Reserve the recorded client spot" ];
 *   Good -> Reserve [ label="Yes" ];
 *   Build [ label="Build the accept message" ];
 *   Reserve -> Build;
 *   AddMsg [ label="Add the accept message", URL="\ref AddMessage()" ];
 *   Build -> AddMsg;
 *   End [ label="End", shape="Mrecord" ];
 *   ChkCli -> End [ label="Yes" ];
 *   Deny -> End;
 *   AddMsg -> End;
 * @enddot
 *
 * @param origin    The originating address of the message.
 * @param buff      The buffer that contains the message to process.
 * @param unused    A required but unused part of the function prototype.
 * @param noid      A required but unused part of the function prototype.
 * @return          Always false; a new connection acceptance message is
 *                  created to acknowledge the request message.
 * @author  Jeff Jackowski
 */
static Bool HandleConnReq(SocketAddr *origin, NetBuffer *buff, Uint8 *unused,
	Uint16 noid) {
	MsgDescr *msg;
	ClientData *client = NULL, *spot = clients;
	int clientId, loop;
	Uint32 randId, now = SDL_GetTicks();
	// read in the random ID
	MSG_READ_32C(buff->data, 2, randId);
	// find the next available client position
	for (loop = 0; loop < MAX_PLAYERS; loop++, spot++) {
		// see if this client has already attempted to connect
		if ((spot->randId == randId) && EqualAddresses(&(spot->addr), origin)) {
			// do nothing -- it was handled already
			return FALSE;
		}
		// see if this is an available spot
		if ((client == NULL) && (!spot->contact || (spot->disconn &&
		((now - spot->disconnTime) >= 4096)))) {
			// clear the client flags
			spot->flags = 0;
			// record the client spot for later use
			client = spot;
			client->cid = clientId = loop;
		}
	}
	// if a client spot was not found, or if a new descriptor cannot be made,
	// or if a buffer for the message cannot be made . . .
	if ((client == NULL) || ((msg = MessageNew()) == NULL) ||
	((msg->buffer = BufferNew()) == NULL)) {
		// . . . send a denial message
		char outbuff[5];
		// write the type ID
		outbuff[0] = MSG_TYPE_CONN_DENY;
		// write the random ID
		outbuff[1] = buff->data[2];
		outbuff[2] = buff->data[3];
		outbuff[3] = buff->data[4];
		outbuff[4] = buff->data[5];
		// send it to the requestor and don't worry about errors
		sendto(sock, outbuff, 5, 0, &(origin->addr), origin->length);
		// if a message descriptor was allocated, give it back
		if (msg != NULL) {
			MessageFree(msg);
		}
		// do not acknowledge this message
		return FALSE;
	}
	// mark the client as having contacted the server
	client->contact = TRUE;
	// record the broadcast reception flag
	client->bcast = 0;   // buff->data[1] > 0;
	// record the random identification value
	client->randId = randId;
	// record the originating address
	client->addr = *origin;
	// setup the message descriptor
	msg->approveFunc = NULL;
	msg->ackFunc =  AckConnAccept;
	msg->failFunc = FailConnAccept;
	msg->interval <<= 1;
	// build the response message
	msg->buffer->length = 8;
	msg->buffer->data[0] = MSG_TYPE_CONN_ACCEPT;
	msg->buffer->data[3] = clientId;
	// write the random ID
	MSG_WRITE_32C(msg->buffer->data, 4, client->randId);
	// set the destination
	msg->destMap[GetDestMapIndex(clientId)] = GetDestMapFlag(clientId);
	// one last check
	if (networkState == NETSTATE_SERVER) {
		// add the message to the system
		AddMessage(msg);
	}
	// do not acknowledge this message
	return FALSE;
}

/**
 * Handles a connection denied message from the server by setting the network
 * state to ::NETSTATE_SRVDENY.
 * @internal
 * @param origin    The originating address of the message.
 * @param buff      The buffer that contains the message to process.
 * @param unused    A required but unused part of the function prototype.
 * @param noid      A required but unused part of the function prototype.
 * @return          Always false.
 * @author  Jeff Jackowski
 */
static Bool HandleConnDeny(SocketAddr *origin, NetBuffer *buff, Uint8 *unused,
Uint16 noid) {
	// see if this data is valid
	if ((networkState == NETSTATE_CONNECTING) && (buff->length == 5) &&
	ValidateServer(origin)) {
		// the connection has been denied
		networkState = NETSTATE_SRVDENY;
	}
	return FALSE;
}

/**
 * Handles a connection acceptance message from a server. Checks the random
 * ID in the message to assure the acceptance is for this process.
 * @internal
 *
 * @pre   The client's random ID is stored in ::clientRandId.
 *
 * @post  If successful, the ::networkState is set to ::NETSTATE_RCVCFG,
 *        ::localPlayer is set to the player's ID, ::outgoingAddr is set to
 *        the server's address, and ::sock is modified to no longer broadcast.
 *
 * @param origin    The originating address of the message.
 * @param buff      The buffer that contains the message to process.
 * @param unused    A required but unused part of the function prototype.
 * @param noid      A required but unused part of the function prototype.
 * @return          True if the acceptance was for this client.
 * @author  Jeff Jackowski
 */
static Bool HandleConnAccept(SocketAddr *origin, NetBuffer *buff, Uint8 *unused,
Uint16 noid) {
	Uint32 randId;
	Uint8 cid;
	// this message is only valid when this process is trying to find or
	// connect to a server
	if ((networkState == NETSTATE_INIT) || (networkState >= NETSTATE_RCVCFG) ||
	// the message length must also be correct
	(buff->length != 8)) {
		// ignore this message
		return FALSE;
	}
	// read in the client ID
	cid = buff->data[3];
	#if (MAX_PLAYERS < 256)
	// check for an ID out of bounds
	if (cid >= MAX_PLAYERS) {
		// bad data; ignore it
		return FALSE;
	}
	#endif
	// read in the random ID
	MSG_READ_32C(buff->data, 4,randId);
	// check for a matching random ID
	if (clientRandId == randId) {
		int bcast = 0;
		// set the new ID value
		localPlayer = cid;
		// change the network status
		networkState = NETSTATE_RCVCFG;
		// do not broadcast
		setsockopt(sock, SOL_SOCKET, SO_BROADCAST, (const char*)&bcast,
		sizeof(int));
		// send future messages to the server
		outgoingAddr = *origin;
		// acknowledge the message
		return TRUE;
	}
	// message not for this host
	return FALSE;
}

extern int timeSyncState;

void MsgFailDropConnection(MsgDescr *descr, ClientData *client) {
	// see if this process is a client or is disconnecting
	if ((networkState == NETSTATE_RCVCFG) ||
	(networkState == NETSTATE_CLIENT) || (networkState == NETSTATE_CLIDISCONN)) {
		// move this process to the initial networking state
		CloseSocket(sock);
		#ifndef NDEBUG
		if (networkState < NETSTATE_CLIDISCONN) {
			#ifndef NDEBUG
			#ifdef USE_GETTEXT
			wprintf(L"Lost connection to the server\n");
			#else
			printf("Lost connection to the server\n");
			#endif
			#endif
		}
		#endif
		networkState = timeSyncState = NETSTATE_INIT;
	}
	// see if this process is the server
	if (networkState == NETSTATE_SERVER) {
		// is this a disconnect message, and is
		// the failing client the disconnecting one?
		if ((descr->buffer->length == 4) &&
		(descr->buffer->data[0] == MSG_TYPE_CONN_DISCONN) &&
		(descr->buffer->data[3] == client->cid)) {
			// remove the offending client
			client->flags = 0;
			players[client->cid].flags = 0;
			// take the client out of the config list
			RemoveFromConfigList(client);
			#ifndef NDEBUG
			#ifdef USE_GETTEXT
			wprintf(L"Client #%d disconnected and stopped responding\n",
			client->cid);
			#else
			printf("Client #%d disconnected and stopped responding\n",
			client->cid);
			#endif
			#endif
		}
		// this client was told of another that disconnected
		else {
			#ifndef NDEBUG
			#ifdef USE_GETTEXT
			wprintf(L"Client #%d did not respond to a disconnect message\n",
			client->cid);
			#else
			printf("Client #%d did not respond to a disconnect message\n",
			client->cid);
			#endif
			#endif
			// send this client a disconnect since it isn't responding
			SendDisconnectMessage(client->cid);
			// take the client out of the config list
			RemoveFromConfigList(client);
		}
	}
	// see if this process is a disconnecting server
	//else if (networkState == NETSTATE_SRVDISCONN) {
		// change the network state to show that the process is complete
		
}

void MsgFailDisconnect(MsgDescr *descr, ClientData *client) {
	// for clients . . .
	if ((networkState == NETSTATE_RCVCFG) ||
	(networkState == NETSTATE_CLIENT)) {
		#ifndef NDEBUG
		#ifdef USE_GETTEXT
		wprintf(L"The server stopped responding\n");
		#else
		printf("The server stopped responding\n");
		#endif
		#endif
		// the server failed to respond, so leave the game
		SendDisconnectMessage(localPlayer);
		networkState = NETSTATE_CLIDISCONN;
	}
	else if (networkState == NETSTATE_SERVER) {
		#ifndef NDEBUG
		#ifdef USE_GETTEXT
		wprintf(L"Client #%d stopped responding; no ack to message %d\n",
		client->cid, descr->msgId);
		#else
		printf("Client #%d stopped responding; no ack to message %d\n",
		client->cid, descr->msgId);
		#endif
		#endif
		// flag the player and client as inactive
		RemovePlayer(&players[client->cid]);
		client->active = FALSE;
		// flag the client as disconnecting and record the time of the event
		client->disconn = TRUE;
		client->disconnTime = SDL_GetTicks();
		assert(client->contact == TRUE);
		// take the client out of the config list
		RemoveFromConfigList(client);
		// a client failed to respond, so eject it from the game
		SendDisconnectMessage(client->cid);
	}
}

/**
 * Handle an incomming disconnection message. Clients must handle the case of
 * another client leaving the game and of the server killing the game. The
 * server removes disconnecting clients from the game and sends a message to
 * all other clients about the disconnection.
 * @internal
 * @param origin    The originating address of the message.
 * @param buff      The buffer that contains the message to process.
 * @param clientId  A pointer to the client ID that needs an acknowledgment.
 * @param noid      A required but unused part of the function prototype.
 * @return          True when a valid disconnect is received.
 * @author  Jeff Jackowski
 */
static Bool HandleConnDisconn(SocketAddr *origin, NetBuffer *buff,
Uint8 *clientId, Uint16 noid) {
	// see if this process is a client
	if (((networkState == NETSTATE_RCVCFG) || (networkState == NETSTATE_CLIENT))
	&& ValidateServer(origin)) {
		// see if the disconnecting host is the server
		if (buff->data[3] == 0) {
			#ifndef NDEBUG
			#ifdef USE_GETTEXT
			wprintf(L"Server disconnected\n");
			#else
			printf("Server disconnected\n");
			#endif
			#endif
			// change the network state
			networkState = NETSTATE_DISCONN;
			// acknowledge the message
			return TRUE;
		}
		// see if this client was disconnected (communication troubles)
		if (buff->data[3] == localPlayer) {
			#ifndef NDEBUG
			#ifdef USE_GETTEXT
			wprintf(L"Server terminated connection\n");
			#else
			printf("Server terminated connection\n");
			#endif
			#endif
			// change the network state
			networkState++;
			// acknowledge the message
			return TRUE;
		}
		#if (MAX_PLAYERS < 256)
		// assure that the client ID is within bounds
		if (buff->data[3] >= MAX_PLAYERS) return FALSE;
		#endif
		// remove the corresponding player from the game
		RemovePlayer(&players[buff->data[3]]);
		// acknowledge the message
		return TRUE;
	}
	// see if this process is a server
	if ((networkState == NETSTATE_SERVER) &&
	ValidateClient(origin, buff->data[3])) {
		// flag the player and client as inactive
		RemovePlayer(&players[buff->data[3]]);
		clients[buff->data[3]].active = FALSE;
		// flag the client as disconnecting and record the time of the event
		clients[buff->data[3]].disconn = TRUE;
		clients[buff->data[3]].disconnTime = SDL_GetTicks();
		assert(clients[buff->data[3]].contact == TRUE);
		// acknowledge the message
		*clientId = buff->data[3];
		// take the client out of the config list
		RemoveFromConfigList(&clients[buff->data[3]]);
		#ifndef NDEBUG
		#ifdef USE_GETTEXT
		wprintf(L"Client #%d disconnected\n", buff->data[3]);
		#else
		printf("Client #%d disconnected\n", buff->data[3]);
		#endif
		#endif
		// send the repeated disconnect
		SendDisconnectMessage(buff->data[3]);
		// acknowledge the message
		return TRUE;
	}
	return FALSE;
}

Bool SendDisconnectMessage(Uint8 id) {
	MsgDescr *msg;
	// Send a message to all other clients to inform them of the disconnect.
	// get the descriptor
	if ((msg = MessageNew()) == NULL) {
		// failed; nothing more can be done
		return FALSE;
	}
	// get a buffer
	if ((msg->buffer = BufferNew()) == NULL) {
		// failed; relase the message
		MessageFree(msg);
		return FALSE;
	}
	// setup the message descriptor
	msg->approveFunc = NULL;
	msg->ackFunc = NULL;
	msg->failFunc = MsgFailDropConnection;
	msg->buffer->length = 4;
	//msg->interval = INIT_RESEND << 2;
	msg->triesLeft = INIT_TRIES << 1;
	// build the response message
	msg->buffer->data[0] = MSG_TYPE_CONN_DISCONN;
	msg->buffer->data[3] = id;
	msg->bcast = TRUE;  // broadcast OK
	// set the destination
	if (networkState == NETSTATE_SERVER) {
		// tell everyone of the departure, including the failed client in hopes
		// of making it more obvious to that client
		SetDestToAll(msg);
	}
	else {
		// this must be set to indicate the server is the destination
		msg->destMap[0] = 1;
		#ifndef NDEBUG
		#ifdef USE_GETTEXT
		wprintf(L"Disconnecting from server\n");
		#else
		printf("Disconnecting from server\n");
		#endif
		#endif
	}
	// add the message to the system
	return AddMessage(msg);
}

/**
 * The function pointer table of functions that respond to incomming messages.
 * @internal
 */
static const HandleMsg messageHandlers[MSG_TYPE_MAX] = {
	HandleAck,               // MSG_TYPE_ACK
	HandleConnReq,           // MSG_TYPE_CONN_REQ
	HandleConnDeny,          // MSG_TYPE_CONN_DENY
	HandleConnAccept,        // MSG_TYPE_CONN_ACCEPT
	HandleConnDisconn,       // MSG_TYPE_CONN_DISCONN
	HandleGameOption,        // MSG_TYPE_OPTION
	HandleObstacle,          // MSG_TYPE_OBSTACLE
	HandleArea,              // MSG_TYPE_SPAWN
	HandleArea,              // MSG_TYPE_GOAL
	// NULL,                    // MSG_TYPE_GAME_START
	// NULL,                    // MSG_TYPE_GAME_OVER
	HandleScoreUpdate,       // MSG_TYPE_SCORE
	HandlePlayerUpdate,      // MSG_TYPE_PLAYER_UPDATE
	HandlePlayerSpawn        // MSG_TYPE_PLAYER_SPAWN
};

/**
 * The maximum number of messages that are acknowledged in a single message.
 * @internal
 */
#define MSG_MAX_ACK 15

/**
 * The size of a single acknowledge message in bytes.
 * @internal
 */
#define MSG_ACK_SIZE  ((MSG_MAX_ACK + 1) << 1)

/**
 * The buffer used to receive incomming messages.
 * @internal
 */
static NetBuffer inbuff;

/**
 * The buffer used to store acknowledgment message(s). For clients, its inital
 * size should be ::MSG_ACK_SIZE. For servers, it should be ::MSG_ACK_SIZE *
 * ::MAX_PLAYERS. In both cases, all bytes must be set to zero.
 * @internal
 */
static Uint8 *ackbuff = NULL;

/**
 * Transmits an acknowledgment message.
 * @internal
 * @param ack  The start of the buffer that holds the acknowledge message to
 *             send.
 * @author     Jeff Jackowski
 */
static void SendAck(Uint8 *ack) {
	SocketAddr *dest;
	// find the destination
	if (networkState == NETSTATE_SERVER) {
		dest = &(clients[ack[1]].addr);
	}
	else {
		dest = &outgoingAddr;
	}
	// construct the rest of the ack message
	ack[0] = MSG_TYPE_ACK;
	ack[1] = localPlayer;
	// send it
	sendto(sock, ack, (ack[2] << 1) + 3, 0, &(dest->addr), dest->length);
	// reset the message counter
	ack[2] = 0;
}

Bool ReceiveMessages() {
	Uint8 *ack = ackbuff;
	SocketAddr origin;
	int error = 0;
	Uint16 msgId;
	Uint8 list[rounddivup(MAX_PLAYERS, 8)] = { 0 };
	Uint8 msgType, clientId = 0;
	// assure the network state is valid for receiving messages
	assert(networkState <= NETSTATE_SRVDISCONN);
	// loop while there are incomming messages to process
	while ((origin.length = sizeof(struct sockaddr_in6),  // set max length
	// receive data
	(inbuff.length = recvfrom(sock, inbuff.data, MSG_MAX_SIZE, 0, &origin.addr, 
	&origin.length)) > 0) || ((error = GetNetErrorVal()) == ECONNREFUSED)) {
		// check for an error caused by previously sending data somewhere
		if (error == ECONNREFUSED) {
			// clear the error value
			error = 0;
			// ignore the problem and try to receive more data
			continue;
		}
		// read in the type
		msgType = inbuff.data[0];
		// assure a valid type
		if (msgType < MSG_TYPE_MAX) {
			assert(messageHandlers[msgType] != NULL);
			// message includes an ID for itself?
			if (msgType >= MSG_TYPE_CONN_ACCEPT) {
				// read in the message ID
				MSG_READ_16C(inbuff.data, 1, msgId);
			}
			// branch based on type and see if acknowledgment is requested
			if ((*(messageHandlers[msgType]))(&origin, &inbuff, &clientId,
			msgId)) {
				// if a server . . .
				if (networkState == NETSTATE_SERVER) {
					// assure the client ID is valid
					assert(clientId < MAX_PLAYERS);
					// get a reference to the right spot in the ack buffer
					ack = &(ackbuff[clientId * MSG_ACK_SIZE]);
				}
				// assure the client ID has not changed when acting as a client
				assert((networkState != NETSTATE_CLIENT) || (clientId == 0));
				// add a message
				MSG_WRITE_16C(ack, 3 + (ack[2] << 1), msgId);
				// set the client ID
				ack[1] = clientId;
				// check for maximum stored messages
				if (++ack[2] == MSG_MAX_ACK) {
					// send it
					SendAck(ack);
					// the client no longer needs acknowledgment
					list[clientId >> 3] &= ~(1 << (clientId & 7));
				}
				else {
					// mark the client as needing an ack
					list[clientId >> 3] |= 1 << (clientId & 7);
				}
			}
			#ifdef MSG_DEBUG
			else if (msgType >= MSG_TYPE_CONN_ACCEPT) {
				printf("Not ack'ing message #%i, type %i\n", msgId, msgType);
			}
			#endif
		}
		// if an invlaid type, ignore the whole message
		#ifdef MSG_DEBUG
		else {
			printf("Got bad message type\n");
		}
		#endif
	}
	// if a server . . .
	if (networkState == NETSTATE_SERVER) {
		int byteLoop = 0, bitLoop;
		clientId = 0;
		// loop through the bytes
		for (byteLoop = 0; byteLoop < rounddivup(MAX_PLAYERS, 8); byteLoop++) {
			// check for no acks in this block of 8
			if (list[byteLoop] == 0) {
				// advance the client count past this block
				clientId += 8;
				// move to the next block
				continue;
			}
			// loop through the bits
			for (bitLoop = 0; list[byteLoop] && (bitLoop < 8);
			bitLoop++, list[byteLoop] >>= 1, clientId++) {
				// see if an ack should be sent to this client 
				if (list[byteLoop] & 1) {
					// send the ack message
					SendAck(&(ackbuff[clientId * MSG_ACK_SIZE]));
				}
			}
		}
	}
	// if a client and there are messages to acknowledge . . .
	else if (ack[2]) {
		// acknowledge the server
		SendAck(ack);
	}
	// check for a real error, not a blocking condition
	if ((error > 0) && (error != EWOULDBLOCK)) {
		#ifndef NDEBUG
		// output an error message
		#ifdef USE_GETTEXT
		printf("Error while trying to recive data:\n\t%ls\n",
		GetNetErrorString(error));
		#else
		printf("Error while trying to recive data:\n\t%s\n",
		GetNetErrorString(error));
		#endif
		#endif
		return FALSE;
	}
	return TRUE;
}

Bool MessageInit() {
	int loop;
	// setup the vectors for their intended use
	available.destructOp = inflightTime.destructOp = NULL;
	buffers.destructOp = GenericPointerFree; // not BufferVecDestr()!
	inflightTime.lessOp = (LessThanOp)TimeCompare;
	available.itemSize = inflightTime.itemSize =
	buffers.itemSize = sizeof(void *);
	available.minSize = inflightTime.minSize = MIN_POINTERS;
	buffers.minSize = MIN_POINTERS >> 1;
	// initalize a vector and check for failure
	if (!VectorInit(&inflightTime, MIN_POINTERS))
		return FALSE;
	// do it again with a different vector
	if (!VectorInit(&available, MIN_POINTERS))
		return FALSE;
	// one more time
	if (!VectorInit(&buffers, MIN_POINTERS >> 1))
		return FALSE;
	// allocate memory for the initial set of buffers
	for (loop = INIT_BUFFERS; loop > 0; loop--) {
		char **buff = VectorAdd(&buffers);
		if ((*buff = malloc(sizeof(NetBuffer))) == NULL) {
			return FALSE;
		}
	}
	// allocate memory for the initial set of messages
	for (loop = INIT_MESSAGES; loop > 0; loop--) {
		MsgDescr **msg = VectorAdd(&available);
		if ((*msg = malloc(sizeof(MsgDescr))) == NULL) {
			return FALSE;
		}
		// clear the flags
		(*msg)->flags = 0;
		// clear the buffer pointer
		(*msg)->buffer = NULL;
	}
	// allocate an acknowledge buffer for the client
	if ((ackbuff = calloc(1, MSG_ACK_SIZE)) == NULL) {
		return FALSE;
	}
	return TRUE;
}

void MessageUninit() {
	// alter the message vectors so that they will free all the message data
	available.destructOp = inflightTime.destructOp =
	(DestructorOp)MsgDescrDestructor;
	// remove the messages
	if (inflightId != AA_ROOT_NODE) {
		// destroy the tree
		AATreeDestroy(inflightId);
		// keep a valid tree
		inflightId = AA_ROOT_NODE;
	}
	VectorDestroy(&available);
	VectorDestroy(&inflightTime);
	// remove the buffers
	VectorDestroy(&buffers);
	free(ackbuff);
}

/**
 * Transitions the messaging system from a client state (really anything other
 * than the server state) to the server state.
 * @internal  Do not call this function directly; instead, call NetCliToSrv().
 * @return  True on sucess, false on failure.
 * @author Jeff Jackowski
 */
Bool MessageCliToSrv() {
	Uint8 *buff;
	// change the size of the acknowledge buffer to handle multiple clients
	if ((buff = realloc(ackbuff, MSG_ACK_SIZE * MAX_PLAYERS)) == NULL) {
		return FALSE;
	}
	// clear the acknowledge buffer
	memset(ackbuff = buff, 0, MSG_ACK_SIZE * MAX_PLAYERS);
	//  remove all messages in the system
	CancelAllMessages();
	return TRUE;
}

/**
 * Transitions the messaging system from the server state to the client state.
 * @internal  Do not call this function directly; instead, call NetSrvToCli().
 * @return  True on sucess, false on failure.
 * @author Jeff Jackowski
 */
Bool MessageSrvToCli() {
	Uint8 *buff;
	// change the size of the acknowledge buffer to be just large enough for
	// one other system (the server)
	if ((buff = realloc(ackbuff, MSG_ACK_SIZE)) == NULL) {
		return FALSE;
	}
	// clear the acknowledge buffer
	memset(ackbuff = buff, 0, MSG_ACK_SIZE);
	//  remove all messages in the system
	CancelAllMessages();
	// no game option messages have been received yet
	gameOpts.totalMsgs = -1;
	gameOpts.rcvdMsgs = 0;
	return TRUE;
}

void ProcessNetworkMessages() {
	// if the network state uses the messaging system . . .
	if ((networkState >= NETSTATE_RCVCFG) &&
	(networkState <= NETSTATE_SERVER)) {
		// process incomming messages
		ReceiveMessages();
		// run through the game loop
		UpdatePlayers(frameTime);
		if (networkState == NETSTATE_SERVER) {
			// handle configuration messages
			ServiceGameConfig();
		}
		// process outgoing messages
		ServiceMessages();
	}
}

