/**
 * @file message.h
 * Interface to the @ref NetMsgSys.
 * @ingroup NetMsgSys
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#ifndef _message_h_
#define _message_h_

#include "client.h"

/**
 * @defgroup NetMsgSys Messaging System
 *
 * A system used to send data, receive acknowledgment of sent data, and
 * responding to acknowledgments or lack thereof.
 *
 * - When messages are sent but not yet acknowledged, they are in-flight.
 * - All message descriptors (::MsgDescr) are dynamically allocated.
 * - Pointers to message descriptors for in-flight messages are stored in a
 *   @ref ::inflightTime "priority queue" by resend time, and in an
 *   @ref ::inflightId "AA-Tree" by message ID.
 * - Pointers to unused but allocated message descriptors are stored in a
 *   @ref ::available "list". If the list has the
 *   @ref ::MAX_MESSAGES "maximum number of descriptors", then freed 
 *   descriptots are deallocated. If the list is empty, new descriptors are
 *   allocated on demand.
 * - Pointers to unused buffers are stored in a @ref ::buffers "list". There is
 *   no list of buffers that are in use, but the message descriptors contain
 *   a @ref MsgDescr::buffer "pointer" to the buffer it owns.
 *
 * Diagram of message data relationships:
 * @dot
 * digraph MessageUse {
 *   node [shape=record, fontname=Helvetica, fontsize=7, height=0.1];
 *   edge [fontname=Helvetica, fontsize=7];
 *   // in-flight messages
 *   InFlight [ label="In-flight message descriptor", URL="\ref MsgDescr" ];
 *   BinHeap [ label="Binary heap", URL="\ref ::inflightTime" ];
 *   AATree [ label="AA-Tree", URL="\ref ::inflightId" ];
 *   BinHeap -> InFlight [ label="Sort by resend time",
 *     URL="\ref MsgDescr::nextSend" ];
 *   AATree -> InFlight [ label="Sort by ID", URL="\ref MsgDescr::msgId" ];
 *   // available messages
 *   Avail [ label="Available message desctiptor", URL="\ref MsgDescr" ];
 *   AvailList [ label="List of available descriptors",
 *     URL="\ref ::available" ];
 *   AvailList -> Avail;
 *   InFlight -> Avail [ label="Finished with message", style="dashed" ];
 *   Avail -> InFlight [ label="Add new message", style="dashed",
 *     URL="\ref AddMessage()" ];
 *   // buffers
 *   UsedBuff [ label="Used buffer" ];
 *   AvailBuff [ label="Available buffer" ];
 *   UsedBuff -> AvailBuff [ label="Free buffer", style="dashed",
 *     URL="\ref BufferFree()" ];
 *   AvailBuff -> UsedBuff [ label="Allocate buffer", style="dashed" ];
 *   InFlight -> UsedBuff [ label="MsgDescr::buffer",
 *     URL="\ref MsgDescr::buffer" ];
 *   BuffList [ label="List of available buffers", URL="\ref ::buffers" ];
 *   BuffList -> AvailBuff;
 * }
 * @enddot
 * @bug In the above dot digraph, all the links must not use the Javadoc style
 *      @ symbol to denote tags, although the Javadoc style is valid everywhere
 *      else.
 * @bug It would be nice to add line breaks to the labels in the above graph,
 *      but Doxygen wrecks the line break tags that the dot tool recognizes.
 *
 * In the diagram, arrows point in the same direction as the actual pointer.
 * Dashed lines indicate how an item's state can change at run-time.
 *
 * The following macros affect how the message system operates:
 * - MSG_INF_TRIES -- Messages never leave the system unless all destinations
 *   acknowledge the message, or an error occured while attempting to send.
 *   This can cause problems when clients disconnect followed by other clients
 *   connecting to the server. It is intended to assist with debugging
 *   client-server communication. When a message's MsgDescr::triesLeft value
 *   drops to zero, its value will not be further decerased and the resend
 *   interval will be set to a constant 16 seconds.
 *
 * @warning  The messaging system is not thread safe.
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/ 
 *
 * @author Jeff Jackowski
 * @ingroup Networking
 * @{
 */

typedef struct MsgDescr_t  MsgDescr;

/**
 * The message type identifiers that are always sent as the first byte of every
 * message. The ID's must be put in the following order to simplify
 * ReceiveMessages():
 * - Messages that lack an ID
 * - Messages that include an ID.
 */
enum {
	/**
	 * Message acknowledgment.
	 */
	MSG_TYPE_ACK,
	/**
	 * Connection request; see @ref NetDesMsgConReq "design" and
	 * @ref NetMsgConReq "message format".
	 */
	MSG_TYPE_CONN_REQ,
	/**
	 * Connection denial.
	 */
	MSG_TYPE_CONN_DENY,
	/**
	 * Connection acceptance; see @ref NetDesMsgConAccept "design"
	 * and @ref NetMsgConAccept "message format".
	 */
	MSG_TYPE_CONN_ACCEPT,
	/**
	 * Disconnection notice. This can be sent by client and server.
	 */
	MSG_TYPE_CONN_DISCONN,
	/**
	 * Game configuration options; only sent by the server.
	 */
	MSG_TYPE_OPTION,
	/**
	 * A set of obstacle data; only sent by the server.
	 */
	MSG_TYPE_OBSTACLE,
	/**
	 * A set of spawn areas; only sent by the server.
	 */
	MSG_TYPE_SPAWN,
	/**
	 * A set of goal areas for CTF; only sent by the server.
	 */
	MSG_TYPE_GOAL,
	/**
	 * Notice that the game is about to begin.
	 */
	MSG_TYPE_GAME_START,
	/**
	 * Notice that the game has concluded.
	 */
	MSG_TYPE_GAME_OVER,
	/**
	 * An update to the scores of the teams.
	 */
	MSG_TYPE_SCORE = MSG_TYPE_GOAL + 1,
	/**
	 * Contains updated state data for multiple players.
	 */
	MSG_TYPE_PLAYER_UPDATE,
	/**
	 * A request from a player to spwan in the game.
	 */
	MSG_TYPE_PLAYER_SPWAN,
	MSG_TYPE_MAX
};

/**
 * The maximum size of a message in bytes. See @ref NetBuffer for details on
 * appropriate values.
 */
#define MSG_MAX_SIZE  510

/**
 * The buffer type used to store messages sent and received over the network.
 * The size of this struct is a power of two for efficient use of dynamically
 * allocated memory.
 */
struct NetBuffer_t {
	/**
	 * The array of arbitrary data.
	 */
	Uint8  data[MSG_MAX_SIZE];
	/**
	 * The length of the message in bytes. A negative value is interpreted as
	 * an error.
	 */
	Sint16 length;
};

typedef struct NetBuffer_t  NetBuffer;

/**
 * The function prototype used to construct the message in a buffer. When the
 * function is called for the first time for a given message, the
 * MsgDescr::inflight flag will be cleared; all subsequent calls will see the
 * flag set. The system is
 * responsible for transmitting the message to all the recipiants.
 * @deprecated
 * @param descr  The message descriptor data.
 */
typedef void (*ConstructMsg)(MsgDescr *descr);

/*
 * The simplest message contructor -- writes the message ID to a pre-existing
 * pre-filled buffer. Useful for messages that are constructed in buffers
 * prior to being added to the message system because the message ID is not
 * known until the message is put into the system.
 * @deprecated
 * @pre  The buffer for the message exists and already has message data prior
 *       to calling AddMessage().
 * @post The message ID is written to the buffer starting a byte 1, and the
 *       MsgDescr::constrFunc is set to NULL.
 * @param descr  The message descriptor of the message to modify.
 * @author Jeff Jackowski
 */
//void MsgSetId(MsgDescr *descr);

/**
 * The function prototype used to react to the recipient acknowledging a
 * message. After this function is called, the bit corresponding to the
 * acknowledging host in MsgDescr::destMap will be cleared.
 *
 * @pre   The client has been validated with ValidateClient().
 *
 * @param descr  The message descriptor data.
 * @param cid    The client's ID.
 */
typedef void (*AcknowledgeMsg)(MsgDescr *descr, int cid);

/**
 * The function prototype used to react to a recipient that did not acknowledge
 * a message, or if an error occured when trying to send a message.
 * @param descr  The message descriptor data.
 * @param client A pointer to the client data that failed to respond, or NULL
 *               if this process is a client.
 */
typedef void (*FailMsg)(MsgDescr *descr, ClientData *client);

/**
 * A message failure function that immediately terminates all communication.
 * It does not send out a disconnect message unless it is on the server and the
 * message is not a disconnect message or the failed client is the one
 * specified in the disconnect message.
 * @param descr  The message descriptor data.
 * @param client A pointer to the client data that failed to respond, or NULL
 *               if this process is a client.
 * @author  Jeff Jackowski
 */
void MsgFailDropConnection(MsgDescr *descr, ClientData *client);

/**
 * A message failure function that disconnects from the remote end in the most
 * reliable manner possible. It sends out the disconnect message to all
 * possible destinations. Because a failure has occured, the disconnect
 * message may not be received.
 * @param descr  The message descriptor data.
 * @param client A pointer to the client data that failed to respond, or NULL
 *               if this process is a client.
 * @author  Jeff Jackowski
 */
void MsgFailDisconnect(MsgDescr *descr, ClientData *client);

/**
 * Approves sending a message to a given recipient. If approval is not given,
 * the recipient will be removed from the message's destination list and the
 * corresponding FailMsg function will not be called.
 * @param descr  The message descriptor data for the message to approve.
 * @param cid    The destination client ID.
 * @return       True if the message should be sent to the indicated player.
 */
typedef Bool (*ApproveSendMsg)(MsgDescr *descr, int cid);

/**
 * The data type used for MsgDescr::destMap.
 */
#define DESTMAPTYPE  Uint16

/**
 * The number of elements in the array MsgDescr::destMap.
 */
#define DESTMAPLEN  bitarraylen(MAX_PLAYERS, DESTMAPTYPE)

/**
 * Returns the index to the MsgDescr::destMap array that contains the flag for
 * the given client.
 * @param  id  The ID of the client who's array position is sought.
 * @return     The index.
 * @author     Jeff Jackowski
 */
#define GetDestMapIndex(id)  bitindex(id, DESTMAPTYPE)

/**
 * Returns the bit flag that indicates the given client in the
 * MsgDescr::destMap array at the correct index (see GetDestMapIndex()).
 * @param  id  The ID of the client who's bit flag is sought.
 * @return     The bit flag. Only the client's bit is set.
 * @author     Jeff Jackowski
 */
#define GetDestMapFlag(id)  bitflag(id, DESTMAPTYPE)

/**
 * Marks a message as being destined to all clients.
 * @param  msg  A pointer to the message descriptor to modify.
 * @author      Jeff Jackowski
 */
#define SetDestToAll(msg) \
{                                                                \
	int loop;                                                    \
	DESTMAPTYPE *map = &((msg)->destMap[0]);                     \
	*map = ~0;                                                   \
	for (loop = DESTMAPLEN - 1; loop > 0; loop--) *(++map) = ~0; \
}
/*
	DESTMAPTYPE *map = &((msg)->destMap[0]);                  \
	for (loop = DESTMAPLEN; loop > 0; loop--) *(map++) = ~0;  \
*/

/**
 * Describes a message, including who should receive it and data needed to
 * handle acknowledgments.
 * @author Jeff Jackowski
 */
struct MsgDescr_t {
	/*
	 * The function called to construct the message.
	 * @warning Either this function pointer must not be NULL, or buffer and
	 *          length must already be set and the buffer must contain the
	 *          message data. See BufferNew() for the required steps to enter
	 *          a pre-made message (this pointer NULL) into the system.
	 */
	//ConstructMsg constrFunc;
	/**
	 * The function called to react to acknowledgment of recipt of the message.
	 * May be NULL if there is no action upon acknowledgment.
	 */
	AcknowledgeMsg ackFunc;
	/**
	 * The function called to react to a recipient that failed to acknowledge
	 * the message. May be NULL if there is no action upon failure.
	 */
	FailMsg failFunc;
	/**
	 * Approves a message to be sent to each recipient, or NULL to always
	 * approve sending the message.
	 */
	ApproveSendMsg approveFunc;
	/**
	 * @warning  Changed to always be persistent, but will be freed after the
	 *           last (re)send potentially before the message is freed.
	 *
	 * A buffer given to this descriptor to write out a message.
	 * The lifespan of the message is one transmission unless either one of the
	 * flags buffLifeMsg or buffLifeUnmanaged is set. If the buffer should be
	 * dropped while the message is still in-flight, then call BufferFree()
	 * with this pointer and clear buffLifeMsg. To regain a persistent buffer,
	 * set buffLifeMsg in the constrFunc function.
	 *
	 * The length of the buffer should be set by constrFunc or be set before
	 * the call to AddMessage(). A negative value is interpreted as an error
	 * and will cause the message to be canceled.
	 */
	NetBuffer *buffer;
	/**
	 * A pointer to arbitrary data for use by the message's handling functions.
	 * @bug  If the message is canceled, any dynamically allocated data
	 *       referenced here will not be deallocated.
	 */
	void *data;
	/**
	 * A bitmap of which clients the message is destined for who have not yet
	 * acknowledged the message. All elements of the array are set to zero in
	 * MessageNew(). Using this provides greater flexibility than storing a
	 * single address and uses less memory than an INetAddr struct. Use
	 * ::DESTMAPLEN as the array's size.
	 * @warning  The data type is configurable, so do not assume its size.
	 * @note     On the client, the least significant bit of the first array
	 *           element is used to denote the destination is the server. It
	 *           must be cleared to indicate the server has acknowledged the
	 *           message, and it must be set either by the constrFunc or before
	 *           AddMessage() is called.
	 */
	DESTMAPTYPE destMap[DESTMAPLEN];
	/**
	 * The common game time when the message will be resent if not all
	 * recpients have acknowledged the message. This value is used to sort
	 * messages in the ::inflightTime priority queue and is set at the end
	 * of the internal function Send().
	 */
	Uint32 nextSend;
	/**
	 * The identifier for the message. This identifier is sent with all copies
	 * of this message, including re-transmissions. The value is set by
	 * AddMessage().
	 */
	Uint16 msgId;
	/**
	 * The time in milliseconds between sends. A default value is set by
	 * MessageNew() and it is doubled after nextSend is updated. The value may
	 * be changed before calling AddMessage() and by the message's constrFunc.
	 */
	Uint16 interval;
	/**
	 * The number of attempts remaining before giving up on this message. When
	 * value reaches zero, the message remains in the system for one more
	 * interval to allow incomming acknowledgements more time. After the
	 * interval, notifictaion of any non-acknowledging destinations will be
	 * made through calling failFunc just before canceling the message.
	 */
	Sint8 triesLeft;
	union {
		/**
		 * Bit flags that indicate the message's state and affect how the
		 * message is handled. The value is set to zero by MessageNew().
		 */
		Uint8 flags;
		struct {
			/**
			 * True while this struct has valid data and is being used to send
			 * and track acknowledgments to a message.
			 * @deprecated
			 */
			Bool inflight   : 1;
			/**
			 * Indicates that the message may be broadcast to clients.
			 * @note  Only valid for servers.
			 */
			Bool bcast      : 1;
			/**
			 * Indicates that the message may be multicast to clients.
			 * @warning  This is not yet implemented.
			 * @note  Only valid for servers.
			 */
			Bool mcast      : 1;
			/**
			 * When true, the message will not be resent. This still allows the
			 * message to be acknowledged. The message will be canceled once
			 * its triesLeft count has dropped to zero, or the next retry after
			 * all destinations have acknowledged
			 */
			Bool noResend   : 1;
			/**
			 * Indicates the message buffer will not be destroyed by the messaging
			 * system until BufferFree() is called outside of the system.
			 */
			Bool buffLifeUnmanaged : 1;
			/**
			 * True when the message format does not need a message ID number
			 * added. Normally messages will need an ID. They will be added when
			 * this flag is cleared.
			 */
			Bool doNotAddId : 1;
		};
	};
};

/**
 * Initializes the network messaging system.
 * @return True if the operation succeeded.
 * @author Jeff Jackowski
 */
Bool MessageInit();

/**
 * Uninitializes the network messaging system.
 * @author Jeff Jackowski
 */
void MessageUninit();

/**
 * Gets a new uninitialized message so that it can be put to use describing
 * a new message.
 * @post    The resulting message will have default values for:
 *          - The resend interval (MsgDescr::interval), set to ::INIT_RESEND.
 *          - The remaining tries (MsgDescr::triesLeft), set to ::INIT_TRIES.
 *          - The option flags (MsgDescr::flags), set to zero.
 * @return  A pointer to the message data, or NULL if there is inadequate
 *          memory.
 * @author  Jeff Jackowski
 */
MsgDescr *MessageNew();

/**
 * Returns an unused message to the queue of available messages, or deallocates
 * it if the maximum number of messages are in the available queue
 * (::MAX_MESSAGES). When the macro MSG_PTR_CHK is defined, the process will
 * fail on an assert if the given message is already in the available message
 * list.
 * @warning Do not call this on messages passed to AddMessage().
 * @pre     The message is not referenced in either the ID sorted tree
 *          (::inflightId) or the resend time sorted priority queue
 *          (::inflightTime).
 * @return  msg  The unused message.
 * @author  Jeff Jackowski
 */
void MessageFree(MsgDescr *msg);

/**
 * Obtains a new message buffer from the internal
 * list of buffers. Intended to allow a buffer to be added to a message prior
 * to passing its descriptor to AddMessage() rather than use a message
 * constructor function (MsgDescr::constrFunc). For this to work, do the
 * following:
 * - Set MsgDescr::buffer to the non-NULL return value of this function.
 * - Put the desired data in the buffer.
 * - Set MsgDescr::buffer::length to the length in bytes of the message in the
 *   buffer.
 * - Set the MsgDescr::buffLifeMsg flag.
 *
 * @return  The address of the buffer, or NULL if no buffers were available
 *          and a new buffer could not be allocated. The new buffer's
 *          NetBuffer::length field will be initialized to zero.
 * @author  Jeff Jackowski.
 */
NetBuffer *BufferNew();

/**
 * Returns a no longer used network buffer to the queue of available buffers.
 * When the macro MSG_PTR_CHK is defined, the process will fail on an assert
 * if the given buffer is already in the available buffer list.
 * @param buff  The buffer to reclaim.
 * @author Jeff Jackowski
 */
void BufferFree(NetBuffer *buff);

/**
 * A Vector @ref DestructorOp "destructor" that returns a no longer used
 * network buffer to the queue of available buffers.
 * @param buff  A double pointer to the NetBuffer to reclaim.
 *        @warning  The parameter is a single pointer to fit the Vector
 *                  interface for a DestructorOp. The function really expects
 *                  a double pointer. Since the available buffer vector stores
 *                  pointers, the vector code will pass in a pointer into the
 *                  list where a pointer is stored, so a double pointer is
 *                  correct.
 * @author Jeff Jackowski
 */
void BufferVecDestr(void *buff);

/**
 * Adds a new message to the messaging system and sends it out immediately.
 * The system will handle acknowledgments and re-transmissions as needed, and
 * will assume the responsibility of deallocating the message.
 *
 * Flowchart:
 * @dot
 * digraph AddMessageFlow {
 *   node [shape=record, fontname=Helvetica, fontsize=7, height=0.1];
 *   edge [fontname=Helvetica, fontsize=7];
 *   Start [ label="Start", shape="Mrecord" ];
 *   MsgId [ label="Set the message ID" ];
 *   Start -> MsgId;
 *   SendFunc [ label="Call internal Send() function", URL="\ref Send()" ];
 *   MsgId -> SendFunc;
 *   ErrCheck [ label="Did Send() succeed?", shape="diamond" ];
 *   SendFunc -> ErrCheck
 *   AddTree [ label="Add the message to the AA-Tree" ];
 *   ErrCheck -> AddTree [ label="Yes" ];
 *   AddHeap [ label="Add the message to the priority queue" ];
 *   AddTree -> AddHeap;
 *   End [ label="End with success", shape="Mrecord" ];
 *   AddHeap -> End;
 *   EndErr [ label="End with error", shape="Mrecord" ];
 *   ErrCheck -> EndErr [ label="No" ];
 * }
 * @enddot
 *
 * @param msg  The message to add to the system. The pointer should have been
 *             returned by a previous call to MessageNew(). A message can only
 *             be added once, after which it will represent the transmission
 *             state of the message.
 * @return  True on success, false if the message failed to send. On failure,
 *          the message is deallocated.
 * @author  Jeff Jackowski
 */
Bool AddMessage(MsgDescr *msg);

/**
 * Services the outgoing messages that are already in the system. This includes
 * removing messages that have been acknowledged by all recipients, and
 * resending messages that haven't.
 * @author Jeff Jackowski
 */
void ServiceMessages();

/**
 * Handles all incomming messages.
 * @return True if successful, false on network error.
 * @author Jeff Jackowski
 */
Bool ReceiveMessages();

/**
 * Cancels all in-flight messages effectively killing communication with the
 * remote end(s). Prepares the messaging system for use with another host.
 * There is no need to call this function if the message system is about to be
 * uninitialized with a call to MessageUninit().
 * @author  Jeff Jackowski
 */
void CancelAllMessages();

/**
 * Returns the number of messages that have been sent but not acknowledged.
 * @return  The number of messages.
 * @author  Jeff Jackowski
 */
int  GetNumOutgoingMessages();

/**
 * The function prototype used to process incomming messages.
 * @param origin    The originating address of the message as reported by
 *                  recvfrom().
 * @param buffer    The buffer that contains the whole message including the
 *                  message type byte.
 * @param clientId  A pointer used to return the ID of the client sending the
 *                  message. This must be set for the message to be
 *                  acknowledged by the server. For client side operation,
 *                  this value should always be zero.
 * @param messageId The message ID that this function is processing. This is
 *                  useful to prevent re-updating player data that has been
 *                  received twice because the server didn't get an
 *                  acknowledgment in time. Some message types lack a message
 *                  ID; see the @ref NetMsgConAccept "message formats" for
 *                  which ones.
 * @return          True to acknowledge the message, false otherwise.
 */
typedef Bool (*HandleMsg)(SocketAddr *origin, NetBuffer *buffer,
	Uint8 *clientId, Uint16 messageId);

/**
 * Sends a disconnect message (see ::MSG_TYPE_CONN_DISCONN and the
 * @ref NetMsgConDis "data format") indicating the specified client is leaving
 * the game. The message is sent using the messaging system.
 * @param  id  The client ID of the host leaving the game. For servers, the
 *             value should be zero.
 * @return  True if the message was transmitted.
 * @author  Jeff Jackowski
 */
Bool SendDisconnectMessage(Uint8 id);

/**
 * Determines if incomming client data came from a valid source. This function
 * relies on data returned from socket functions of the operating system, and
 * as such is suseptible to things like spoofing.
 * @param origin    The originating address of the data.
 * @param clientId  The reported identity of the client.
 * @return  True if it appears the data came from the indicated client, false
 *          if the data seems to have come from another source.
 * @author  Jeff Jackowski
 */
Bool ValidateClient(SocketAddr *origin, Uint8 clientId);

/**
 * Determines if the inomming server data came from the server that is running
 * the game, and if this process is a client.
 * @param origin    The originating address of the data.
 * @return  True if it appears the data came from the server and should not
 *          be ignored.
 * @author  Jeff Jackowski
 */
Bool ValidateServer(SocketAddr *origin);

/**
 * Handles all network I/O for the client or server. Receives incomming
 * messages, services outgoing messages, and updates player data. If running
 * as a server, also services the game configuration messages.
 * @author Jeff Jackowski
 */
void ProcessNetworkMessages();

/** @} */

#endif
