/**
 * @file name.c
 * Implements the interface to store and change the player's name. Intended to
 * be extended with a menu item for the user to change the name. 
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "menustate.h"
//#include "text.h"
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#if defined(USE_GETTEXT) || defined(DOXYGEN)
#include <wchar.h>
#include <libintl.h>
#include <iconv.h>
#endif

/**
 * Set the player name to the  user name stored in an enviornment variable.
 * Known to work on Linux and Windows; will probably work on other
 * Unix-like systems.
 * @post   ::localPlayerName contains the user name as a null terminated string
 *         up to ::MAX_NAME_LEN bytes long.
 * @author Jeff Jackowski.
 */
static void SetUsername() {
	const char *str = NULL;
	// find the user's name from the enviornment settings
	if ((str = getenv("USER")) == NULL) {   // for unix
		// for windows (may be defined but blank on unix systems)
		str = getenv("USERNAME"); 
	}
	// check for no enviornment setting or zero length name
	if ((str == NULL) || (str[0] == 0)) {
		// use the default name
		str = (char*)strings[STRING_DEFAULTNAME];
	}
	#ifdef USE_GETTEXT
	else {
		// a name was found, but it needs to be converted
		#ifndef WIN32
		iconv_t descr = iconv_open("wchar_t", "");
		#else
		iconv_t descr = iconv_open("UTF-16LE", "");
		#endif
		// no error?
		if (descr != (iconv_t)-1) {
			const char *src = str;
			wchar_t *dest = menuSettings->localPlayerName;
			size_t srcLen = strlen(str);
			size_t destLen = sizeof(menuSettings->localPlayerName) *
			sizeof(wchar_t);
			iconv(descr, (char**)&src, &srcLen, (char**)&dest, &destLen);
			*dest = 0;
			iconv_close(descr);
			// all done
			return;
		}
	}
	#endif
	// copy over the name
	rtsa_strncpy(menuSettings->localPlayerName, (rtsa_char*)str,
	MAX_NAME_LEN - 1);
	// assure null termination
	menuSettings->localPlayerName[MAX_NAME_LEN - 1] = 0;
}

void GetLocalPlayerName(rtsa_char *dest) {
	assert(menuSettings != NULL);
	// see if the name has not already been set
	if (menuSettings->localPlayerName[0] == 0) {
		// get the user name
		SetUsername();
	}
	// copy the name over -- name length guaranteed within bounds
	rtsa_strcpy(dest, menuSettings->localPlayerName);
}

void SetLocalPlayerName(rtsa_char *name) {
	assert(menuSettings != NULL);
	// check for request to use the user name from the OS
	if ((name == NULL) || (name[0] == 0)) {
		// use the user name
		SetUsername();
	}
	else {
		// copy the requested name
		rtsa_strncpy(menuSettings->localPlayerName, name, MAX_NAME_LEN - 1);
		// assure null termination
		menuSettings->localPlayerName[MAX_NAME_LEN - 1] = 0;
	}
}

