/**
 * @file net.h
 * Networking stuff.
 * @ingroup Networking
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "bool.h"
#ifndef WIN32
#include <errno.h>
#endif

/**
 * @defgroup Networking Networking
 *
 * Implements the networking portion of the game.
 *
 * @ref NetMsg
 *
 * @{
 */

/**
 * The port used by the server to receive everything but time sync messages.
 */
#define PORT_SRV     26000
/**
 * The port used by the server to send receive time sync messages.
 */
#define PORT_STSYNC  26001
/**
 * The port used by the client to receive everything but time sync messages.
 */
#define PORT_CLI     26002
/**
 * The port used by the client to send and receive time sync messages. It could
 * have the same value as ::PORT_STSYNC, but that would prevent running a
 * client and server on the same computer which is very convienient for
 * development.
 */
#define PORT_CTSYNC  26003

/**
 * A macro used to close sockets. If the socket is -1, it will not be closed.
 * This makes it safe to use the macro even when it is not known if the
 * socket is already open.
 * @param s  The socket to close.
 * @author Jeff Jackowski
 */
#ifdef WIN32
#define CloseSocket(s)  if (s != -1) { closesocket(s); s = -1; }
#else
#define CloseSocket(s)  if (s != -1) { close(s); s = -1; }
#endif

/**
 * A macro to write a byte to a character buffer.
 * @param  ptr  The character pointer to the next writable spot in the buffer.
 *              The pointer will be incremented.
 * @param  val  The value to write to the buffer.
 * @author Jeff Jackowski
 */
#define MSG_WRITE_8(ptr, val)   *((ptr)++) = (val)

/**
 * A macro to write a byte to a character buffer at a constant position within
 * the buffer.
 * @param  ptr  The character pointer to the buffer.
 *              The pointer will not be altered.
 * @param  index  The index that of the buffer that will recieve the first
 *                byte of val. The index will not be altered.
 * @param  val  The value to write to the buffer.
 * @author Jeff Jackowski
 */
#define MSG_WRITE_8C(ptr, index, val)   (ptr)[index] = (val)

/**
 * A macro to write a two byte numerical value in binary form to a character
 * buffer.
 * @param  ptr  The character pointer to the next writable spot in the buffer.
 *              The pointer will be incremented.
 * @param  val  The value to write to the buffer.
 * @author Jeff Jackowski
 */
#define MSG_WRITE_16(ptr, val)  *((ptr)++) = ((val) >> 8); \
                                *((ptr)++) = ((val) & 0xFF)

/**
 * A macro to write a two byte to a value character buffer at a constant
 * position within the buffer.
 * @param  ptr  The character pointer to the buffer.
 *              The pointer will not be altered.
 * @param  index  The index that of the buffer that will recieve the first
 *                byte of val. The index will not be altered.
 * @param  val  The value to write to the buffer.
 * @author Jeff Jackowski
 */
#define MSG_WRITE_16C(ptr, index, val)   (ptr)[index] = ((val) >> 8);\
                                         (ptr)[index + 1] = ((val) & 0xFF)

/**
 * A macro to write a four byte numerical value in binary form to a character
 * buffer.
 * @param  ptr  The character pointer to the next writable spot in the buffer.
 *              The pointer will be incremented.
 * @param  val  The value to write to the buffer.
 * @author Jeff Jackowski
 */
#define MSG_WRITE_32(ptr, val)  *((ptr)++) = ((val) >> 24); \
                                *((ptr)++) = ((val) >> 16) & 0xFF; \
                                *((ptr)++) = ((val) >> 8)  & 0xFF; \
                                *((ptr)++) = ((val) & 0xFF)

/**
 * A macro to write a four byte value to a character buffer at a constant
 * position within the buffer.
 * @param  ptr  The character pointer to the buffer.
 *              The pointer will not be altered.
 * @param  index  The index that of the buffer that will recieve the first
 *                byte of val. The index will not be altered.
 * @param  val  The value to write to the buffer.
 * @author Jeff Jackowski
 */
#define MSG_WRITE_32C(ptr, index, val)  (ptr)[index] = ((val) >> 24); \
                                    (ptr)[index + 1] = ((val) >> 16) & 0xFF; \
                                    (ptr)[index + 2] = ((val) >> 8) & 0xFF; \
                                    (ptr)[index + 3] = ((val) & 0xFF)

/**
 * A macro to read a byte from a character buffer.
 * @param  ptr  The character pointer to the next readable spot in the buffer.
 *              The pointer will be incremented.
 * @param  val  The variable that will contain the value read from the buffer.
 * @author Jeff Jackowski
 */
#define MSG_READ_8(ptr, val)   (val) = *((ptr)++)

/**
 * A macro to read a two byte number in binary form from a character buffer.
 * @param  ptr  The character pointer to the next readable spot in the buffer.
 *              The pointer will be incremented.
 * @param  val  The variable that will contain the value read from the buffer.
 * @author Jeff Jackowski
 */
#define MSG_READ_16(ptr, val)  (val) =  (Uint8)*((ptr)++) << 8; \
                               (val) |= (Uint8)*((ptr)++)

/**
 * A macro to read a two byte value from a character buffer at a constant
 * position within the buffer.
 * @param  ptr  The character pointer to the buffer.
 *              The pointer will not be altered.
 * @param  index  The index that of the buffer with the first
 *                byte of val. The index will not be altered.
 * @param  val  The variable that will contain the value read from the buffer.
 * @author Jeff Jackowski
 */
#define MSG_READ_16C(ptr, index, val) (val) =  ((Uint8*)(ptr))[index] << 8; \
                                      (val) |= ((Uint8*)(ptr))[index + 1]

/**
 * A macro to read a four byte number in binary form from a character buffer.
 * @param  ptr  The character pointer to the next readable spot in the buffer.
 *              The pointer will be incremented.
 * @param  val  The variable that will contain the value read from the buffer.
 * @author Jeff Jackowski
 */
#define MSG_READ_32(ptr, val)  (val) =  (Uint8)*((ptr)++) << 24; \
                               (val) |= (Uint8)*((ptr)++) << 16; \
                               (val) |= (Uint8)*((ptr)++) << 8; \
                               (val) |= (Uint8)*((ptr)++)

/**
 * A macro to read a four byte value from a character buffer at a constant
 * position within the buffer.
 * @param  ptr  The character pointer to the buffer.
 *              The pointer will not be altered.
 * @param  index  The index that of the buffer with the first
 *                byte of val. The index will not be altered.
 * @param  val  The variable that will contain the value read from the buffer.
 * @author Jeff Jackowski
 */
#define MSG_READ_32C(ptr, index, val) \
        (val) =  ((Uint8*)(ptr))[index] << 24; \
        (val) |= ((Uint8*)(ptr))[index + 1] << 16; \
        (val) |= ((Uint8*)(ptr))[index + 2] << 8; \
        (val) |= ((Uint8*)(ptr))[index + 3]

/**
 * Codes for the operating state of the network module used with ::networkState.
 */
enum {
	/**
	 * The inital state of the networking system indicating that it is ready
	 * to act as a client or transition to the server state.
	 */
	NETSTATE_INIT,
	/**
	 * Indicates that a server is being sought on the LAN. This operation uses
	 * the @ref IOThread.
	 */
	NETSTATE_SEARCHING,
	/**
	 * Indicates that an attempt is underway to connect to a server. This is
	 * used during the acknowledgment process. If the player selected a
	 * specific server to use, this code is also used during the initial
	 * request.
	 */
	NETSTATE_CONNECTING,
	/**
	 * Indicates that process is acting as a game client and is receiving game
	 * configuation information in preperation for game play.
	 */
	NETSTATE_RCVCFG,
	/**
	 * Indicates this process is acting as a game client that is already
	 * connected to a server and has the game state data.
	 */
	NETSTATE_CLIENT,
	/**
	 * Indicates that the client process is attempting to disconnect from the
	 * server in a reliable way. The operation uses the messaging system on the
	 * @ref IOThread to implement the disconnect in a reliable way.
	 */
	NETSTATE_CLIDISCONN,
	/**
	 * Indicates this process is acting as a game server.
	 */
	NETSTATE_SERVER,
	/**
	 * Indicates that the server process is attempting to disconnect from all
	 * clients in a reliable way. The operation uses the messaging system on the
	 * @ref IOThread to implement the disconnect in a reliable way.
	 */
	NETSTATE_SRVDISCONN,
	/**
	 * Indicates that the server contacted on the LAN has denied a connection.
	 * Only valid with servers on the LAN because only one server per LAN is
	 * presently supported. The process will not transition to NETSTATE_SERVER
	 * while in this state. After a connection error with a server beyond the
	 * LAN, the state will be NETSTATE_CONERR instead of this state.
	 */
	NETSTATE_SRVDENY,
	/**
	 * Indicates that there is no connection. This state will persist for some
	 * length of time to inform the user of the event, and so that a process
	 * that sends the disconnect message has some time to reliably inform the
	 * other host(s).
	 */
	NETSTATE_DISCONN,
	/**
	 * Indicates an error occured during the attempt to locate a server.
	 */
	NETSTATE_SEARCHERR,
	/**
	 * Indicates an error occured that prevented the process from acting as
	 * a server.
	 */
	NETSTATE_SRVERR,
	/**
	 * Indicates an error occured while attempting to connect to a server.
	 */
	NETSTATE_CONERR
};

/**
 * Contains the current operating state of the networking module.
 */
extern unsigned int networkState;

/**
 * Initalizes the networking system. Starts the process of searching for a
 * server on the @ref IOThread. Before it can be used as a server, a call
 * must be made to NetCliToSrv(). After initialzing the networking system, it
 * should be uninitialized before the process terminates.
 * @pre  The @ref IOThread has already been started.
 *
 * @post The @ref ::networkState "network state" is set to ::NETSTATE_INIT.
 *
 * @param  srv  The string with the hostname of the server to connect to,
 *              or NULL to discover a server on the LAN.
 * @return True if successful, false on failure.
 * @author Jeff Jackowski
 */
Bool NetInit(char *srv);

/**
 * Initalizes the networking system and attempts to be a server without
 * searching for one on the network. After initialzing the networking system,
 * it should be uninitialized before the process terminates.
 * @pre  The @ref IOThread has already been started.
 *
 * @post The @ref ::networkState "network state" is set to ::NETSTATE_SERVER.
 *
 * @return True if successful, false on failure.
 * @author Jeff Jackowski
 */
Bool NetInitSrv();

/**
 * Uninitalizes the networking system. No networking functions in this module
 * may be called again until after a successful call to NetInit().
 * @author Jeff Jackowski
 */
void NetUninit();

/**
 * Starts the process of disconnecting this client from the server, or
 * disconnecting all clients from this server.
 *
 * @post     The ::networkState is set to ::NETSTATE_CLIDISCONN or 
 *           ::NETSTATE_SRVDISCONN and the
 *           disconnect task is queued on the @ref IOThread.
 *
 * @warning  Since this uses the messaging system on another thread, the
 *           messaging system should not be used on any other thread until
 *           the network state changes.
 * @author   Jeff Jackowski
 */
void NetDisconnect();

/**
 * Changes the network state from a client state to server.
 * @post   The @ref ::networkState "network state" is set to ::NETSTATE_SERVER.
 *
 * @todo   Add IPv6 support.
 *
 * @return True on success, false on failure
 * @author Jeff Jackowski
 */
Bool NetCliToSrv();

/**
 * Changes the network state from server state to client.
 * @post The @ref ::networkState "network state" is set to ::NETSTATE_INIT.
 * @return True on success, false on failure
 * @author Jeff Jackowski
 */
Bool NetSrvToCli();

/**
 * Creates a new unicast ready UDP socket that is bound to the given address.
 * @param  addr   A pointer to a SockAddr, sockaddr_in, or sockaddr_in6 that
 *                contains the address to which the socket will be bound.
 *         @warning  The address family must be set correctly.
 * @param  block  True to create a blocking socket.
 * @return The configured socket, or -1 on error.
 * @author Jeff Jackowski
 */
int UnicastSocket(void *addr, Bool block);

// Make error handing more consistent across Unix-like systems and Windows

/**
 * Retrives the error value from the last network operation.
 * @warning  To work with Windows, winsock2.h must be included. This can cause
 *           name conflicts with some functions, like RemoveMenu(), so
 *           winsock2.h is not included here.
 * @author Jeff Jackowski
 */
#ifdef WIN32
#define GetNetErrorVal()  WSAGetLastError()
#define EWOULDBLOCK  WSAEWOULDBLOCK
#define EBADF        WSAEBADF
#define EFAULT       WSAEFAULT
#define EINTR        WSAEINTR
#define EINVAL       WSAEINVAL
#define ENOTSOCK     WSAENOTSOCK
#define EADDRINUSE   WSAEADDRINUSE
#define ENETDOWN     WSAENETDOWN
#define ENETUNREACH  WSAENETUNREACH
#define ECONNRESET   WSAECONNREFUSED
#define ECONNREFUSED WSAECONNRESET 
#define EHOSTUNREACH WSAEHOSTUNREACH
#else
#define GetNetErrorVal()  errno
#endif

/**
 * A gobal variable used to store network errors after cetrain events, such as
 * during connection errors. Most useful to pass an error condition from one
 * thread to another since the operating system's error is thread local.
 */
extern int netError;

/**
 * Returns a pointer to a string that briefly explains the last network error.
 * @param  err  The error code who's description is to be returned.
 * @return A pointer to a constant string with the error description in terse
 *         English. The strings need to be changed to support other languages.
 * @author Jeff Jackowski
 */
const rtsa_char *GetNetErrorString(int err);

/** @} */

