/**
 * @file netstat.c
 * Implements the network status menu item.
 * @ingroup MenuNetStat
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include <string.h>
#include <stdlib.h>
#include "color.h"
#include "genconmac.h"
#include "hostlookup.h"
#include "netstat.h"
#include "text.h"
#include "timesync.h"
#include "net.h"
#ifndef WIN32
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

char statHost[MAX_HOST_LEN];

/**
 * The data needed by the network status menu item while it is in use. The
 * pointer is maintained in MenuItem::data.
 * @internal
 */
struct NetStatData_t {
	/**
	 * The network state from the previous frame, or -1 if no previous frame.
	 */
	int prevNetStat;
	/**
	 * The time synchronization state from the previous frame, or -1 if no
	 * previous frame.
	 */
	int prevSyncStat;
	/**
	 * True when a hostname for the server has been found and can be rendered.
	 */
	Bool gotHost;
};

typedef struct NetStatData_t  NetStatData;

/**
 * A convienience macro for accessing data maintained for the life of the
 * network status menu item.
 * @internal
 * @author Jeff Jackowski
 */
#define netdat  ((NetStatData*)(item->data))

Bool NetStatInit(MenuItem *item) {
	// allocate memory for the persistant state
	if ((item->data = malloc(sizeof(NetStatData))) == NULL) {
		// failed
		return FALSE;
	}
	netdat->prevNetStat = netdat->prevSyncStat = -1;
	// setup the menu item fields
	item->minWidth = 480;
	item->minHeight = (textItems[TEXTITEM_FIREOPT].loc.h << 1) + 2;
	item->focus = NetStatFocus;
	item->input = NetStatInput;
	item->update = NetStatUpdate;
	item->active = TRUE;
	// put the item containing the network status message into a layer
	if (textItems[TEXTITEM_NETSTAT].renderer == NULL) {
		// the text item will be initialized later with a call to
		// TextRenderString() before it is drawn
		textItems[TEXTITEM_NETSTAT].renderer = (ItemRenderer)1;
	}
	AddRenderItem(&(textItems[TEXTITEM_NETSTAT]), LAYER_TEXT);
	if (textItems[TEXTITEM_NETOPTS].renderer == NULL) {
		textItems[TEXTITEM_NETOPTS].renderer = (ItemRenderer)1;
	}
	AddRenderItem(&(textItems[TEXTITEM_NETOPTS]), LAYER_TEXT);
	// TEXTITEM_FIREOPT is added by the text system for the menu state because
	// it is a constant string.
	return TRUE;
}

void NetStatUninit(MenuItem *item) {
	free(netdat);
}

extern SocketAddr outgoingAddr;

/**
 * Renders the local machine's address to ::textItems[::TEXTITEM_NETOPTS].
 * @internal
 * @pre    The ::localhostAddrStr must not be an empty string.
 * @param  item  The network status menu item to work with.
 * @author Jeff Jackowski
 */
static void RenderLocalAddr(MenuItem *item) {
	// render the address
	TextRenderString(&(textItems[TEXTITEM_NETOPTS]), FONT_SCORE, 0,
	colorFmtInd[COLOR_TEXT], strings[STRING_NETADDR], localhostAddrStr);
	// place it
	textItems[TEXTITEM_NETOPTS].loc.x = item->region.x +
	((item->region.w - textItems[TEXTITEM_NETOPTS].loc.w) >> 1);
	textItems[TEXTITEM_NETOPTS].loc.y = item->region.y +
	textItems[TEXTITEM_NETOPTS].loc.h + 2;
}

/**
 * Informs the user that time sync is in progress.
 * @internal
 * @param  item  The network status menu item to work with.
 * @author Jeff Jackowski
 */
static void RenderTSync(MenuItem *item) {
	// render the address
	TextRenderString(&(textItems[TEXTITEM_NETOPTS]), FONT_SCORE, 0,
	colorFmtInd[COLOR_TEXT], strings[STRING_NETSYNC]);
	// place it
	textItems[TEXTITEM_NETOPTS].loc.x = item->region.x +
	((item->region.w - textItems[TEXTITEM_NETOPTS].loc.w) >> 1);
	textItems[TEXTITEM_NETOPTS].loc.y = item->region.y +
	textItems[TEXTITEM_NETOPTS].loc.h + 2;
}

/**
 * Renders the latency to the server.
 * @internal
 * @param  item  The network status menu item to work with.
 * @author Jeff Jackowski
 */
static void RenderLatency(MenuItem *item) {
	// render the address
	TextRenderString(&(textItems[TEXTITEM_NETOPTS]), FONT_SCORE, 0,
	colorFmtInd[COLOR_TEXT], strings[STRING_NETLAT], timeSyncResults.latency);
	// place it
	textItems[TEXTITEM_NETOPTS].loc.x = item->region.x +
	((item->region.w - textItems[TEXTITEM_NETOPTS].loc.w) >> 1);
	textItems[TEXTITEM_NETOPTS].loc.y = item->region.y +
	textItems[TEXTITEM_NETOPTS].loc.h + 2;
}

int NetStatUpdate(MenuItem *item) {	
	// update the data with the latest information from the server
	// see if the visible components need to be moved
	if (!item->isPlaced) {
		// set the placed flag
		item->isPlaced = TRUE;
		// set the host flag
		netdat->gotHost = FALSE;
		
		// Move the components to the proper place.
		// Place the fire for options prompt
		textItems[TEXTITEM_FIREOPT].loc.x = item->region.x +
		((item->region.w - textItems[TEXTITEM_FIREOPT].loc.w) >> 1);
		textItems[TEXTITEM_FIREOPT].loc.y = item->region.y +
		textItems[TEXTITEM_FIREOPT].loc.h + 2;
		
		// check for a found address
		if (localhostAddrStr[0]) {
			// render the address
			RenderLocalAddr(item);
		}
		
		// see if the options prompt should be visible now
		if (item->hasFocus) {
			// display it at its new location
			PlaceRenderItem(&(textItems[TEXTITEM_FIREOPT]));
		}
		else if (localhostAddrStr[0]) {
			// diplay the game options
			PlaceRenderItem(&(textItems[TEXTITEM_NETOPTS]));
		}
		goto NetStatUpdate_statStr;
	}
	// see if the network state has changed since the text was rendered
	if ((netdat->prevNetStat != networkState) ||
	(!netdat->gotHost && (LookupResult(NULL, NULL) != 0))) {
		rtsa_char const 	*str;
		#ifndef WIN32
		char addrstr[INET6_ADDRSTRLEN];
		#endif
		NetStatUpdate_statStr:
		// make the current network state the previous one
		netdat->prevNetStat = networkState;
		// check for need to lookup name of remote server
		if ((networkState == NETSTATE_CONNECTING) ||
		(networkState == NETSTATE_RCVCFG) ||
		(networkState == NETSTATE_CLIENT) ||
		(networkState == NETSTATE_SRVDENY)) {
			// if there is no host name and an address can be put into a string
			if ((statHost[0] == 0) &&
			#ifdef WIN32
			// use the older IPv4 functions
			((str = (rtsa_char*)inet_ntoa(outgoingAddr.ipv4.sin_addr)) != NULL)) {
				// do nothing to maintain the same logic as on other systems
			}
			#else
			(inet_ntop(outgoingAddr.addr.sa_family,
			(outgoingAddr.addr.sa_family == AF_INET) ?
			(void*)&(outgoingAddr.ipv4.sin_addr) : 
			(void*)&(outgoingAddr.ipv6.sin6_addr),
			addrstr, INET6_ADDRSTRLEN) != NULL)) {
				// use the address
				str = (rtsa_char*)addrstr;
			}
			#endif
			else {
				// use the host name
				str = (rtsa_char*)statHost;
				netdat->gotHost = TRUE;
			}
		}
		else if ((netdat->prevNetStat >= NETSTATE_SEARCHERR) &&
		(netdat->prevNetStat <= NETSTATE_CONERR)) {
			str = GetNetErrorString(netError);
		}
		else {
			str = (rtsa_char*)localhostName;
		}
		// render status text
		TextRenderString(&(textItems[TEXTITEM_NETSTAT]), FONT_SCORE, 0,
		colorFmtInd[COLOR_TEXT],
		strings[STRING_NETLOOKING + netdat->prevNetStat - 1], str);
		// place the text
		textItems[TEXTITEM_NETSTAT].loc.x = item->region.x + 2;
		textItems[TEXTITEM_NETSTAT].loc.y = item->region.y;
		PlaceRenderItem(&(textItems[TEXTITEM_NETSTAT]));
		
		if (networkState == NETSTATE_DISCONN) {
			ClearRenderItem(&(textItems[TEXTITEM_NETOPTS]));
		}
	}
	// check for a change in the time sync state
	if (netdat->prevSyncStat != timeSyncState) {
		// store the current state
		netdat->prevSyncStat = timeSyncState;
		// client side display
		if ((networkState == NETSTATE_RCVCFG) ||
		(networkState == NETSTATE_CLIENT)) {
			// if time sync is in progress . . .
			switch (timeSyncState) {
				case TIMESYNC_INPROG:
					RenderTSync(item);
					break;
				case TIMESYNC_FINISHED:
				case TIMESYNC_POOR:
					RenderLatency(item);
					break;
				default:
					RenderLocalAddr(item);
			}
			// if the text is visible
			if (!item->hasFocus) {
				PlaceRenderItem(&(textItems[TEXTITEM_NETOPTS]));
			}
		}
		// others
		else if (localhostAddrStr[0]) {
			RenderLocalAddr(item);
			// if the text is visible
			if (!item->hasFocus) {
				PlaceRenderItem(&(textItems[TEXTITEM_NETOPTS]));
			}
		}
	}
	return 0;
}

void NetStatFocus(MenuItem *item) {
	// check for focus
	if (item->hasFocus) {
		// show the fire for options prompt
		PlaceRenderItem(&(textItems[TEXTITEM_FIREOPT]));
		if (localhostAddrStr[0] || (timeSyncState &&
		(networkState == NETSTATE_CLIENT))) {
			// remove the game options
			ClearRenderItem(&(textItems[TEXTITEM_NETOPTS]));
		}
	}
	else {
		// remove the fire for options prompt
		ClearRenderItem(&(textItems[TEXTITEM_FIREOPT]));
		if ((networkState != NETSTATE_DISCONN) && (localhostAddrStr[0] ||
		(timeSyncState && (networkState == NETSTATE_CLIENT)))) {
			// show the game options
			PlaceRenderItem(&(textItems[TEXTITEM_NETOPTS]));
		}
	}
}

int NetStatInput(MenuItem *item, Uint8 *ec, rtsa_char keyval) {
	// nothing yet
	return 0;
}

