/**
 * @file notice.h
 * Interface to the in-game notice display.
 * @ingroup NoticeSys
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "render.h"
#include "text.h"

/**
 * @defgroup NoticeSys In-game Notice Display
 *
 * Displays notices to update the player on key events happening in the game.
 *
 * @{
 */

/**
 * The different categories of notices.
 * @todo Decide if this many types are needed; maybe consolidate a few.
 */
enum NOTICE_TYPE {
	/**
	 * A notice about a player destroying a tank of the opposing team.
	 */
	NOTICE_KILL,
	/**
	 * Notice that the local player has destroyed a tank of the opposing team.
	 */
	NOTICE_LOCALKILL,
	/**
	 * A notice about a player destroying a tank on the same team.
	 */
	NOTICE_TEAMKILL,
	/**
	 * Notice that the local player has destroyed a tank on his/her team.
	 */
	NOTICE_LOCALTEAMKILL,
	/**
	 * Indication that the red team's flag is held by a blue team player.
	 */
	NOTICE_REDFLAGHELD,
	/**
	 * Indication that the blue team's flag is held by a red team player.
	 */
	NOTICE_BLUEFLAGHELD,
	/**
	 * A notice that the red team's flag has been captured by the blue team.
	 */
	NOTICE_REDFLAGCAP,
	/**
	 * A notice that the blue team's flag has been captured by the red team.
	 */
	NOTICE_BLUEFLAGCAP,
	NOTICE_MAXTYPE
};

/**
 * An expiration time used to indicate that a message does not expire and will
 * be displayed until explicitly removed.
 */
#define NOTICE_NOEXPIRATION  0xFFFFFFFF

/**
 * Adds a new notice to the display. New notices are added to the bottom of
 * the visible notices, with non-expiring notices added ahead of any expiring
 * notices.
 * @param expiryTime  The common game time when the message should be removed
 *                    from the display. If the message should persist until
 *                    explicitly removed, the value ::NOTICE_NOEXPIRATION
 *                    should be used.
 * @param type        The category for the message. It should be a value
 *                    specified from ::NOTICE_TYPE.
 * @param format      A printf() style formatting string used to generate the
 *                    text that is renderd for the notice.
 * @param ...         Any variable inputs, printf() style.
 * @author  Jeff Jackowski
 */
void AddNotice(Uint32 expiryTime, Uint8 type, rtsa_char *format, ...);

/**
 * Adds a notice about a player shooting another. Internally determines if
 * the local player did the shooting and if the player shot is on the same
 * team to select the correct message.
 * @param killerId  The ID of the player that shot another player.
 * @param killedId  The ID of the player that was shot.
 * @param score     The new score of the player that shot another.
 * @author  Jeff Jackowski
 */
void AddScoreNotice(int killerId, int killedId, int score);

/**
 * Removes a notice of the specified type. Safe to use if the notice type is
 * not currently present.
 * @param type  The category of the message to remove. The value should be
 *              specified from ::NOTICE_TYPE.
 * @author  Jeff Jackowski
 */
void RemoveNotice(Uint8 type);

/**
 * Remove all notices from the system and the display.
 * @author  Jeff Jackowski
 */
void RemoveAllNotices();

/**
 * Update all visible notices by removing any expired notices and placing
 * notices at their proper location when notices are added or removed.
 * @param time  The current common game time to compare against the expiration
 *              time of the notices.
 * @author  Jeff Jackowski
 */
void UpdateNotices(Uint32 time);

/** @} */

