/**
 * @file obstacle.h
 * Interface to code that handles impassible obstacles.
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "player.h"

/**
 * The representation of impassible objects in the game world. Screen and game
 * coordinates are kept separate in order to allow a scaled version of the
 * game to be displayed.
 * @author  Jeff Jackowski
 */
struct Obstacle_t {
	/**
	 * The data used by the rendering system to draw the obstacle. The
	 * coordinates used in ri are screen coordinates and not game coordinates;
	 * they should not be used in game play code and need to be kept in sync
	 * with the game coorinates. The only times the coordinates should need to
	 * be synchronized are when the obstacles are first created, and if the
	 * obstacles move on the battle field (but they shouldn't).
	 */
	RenderItem ri;
	/**
	 * The location of the obstacle in game coordinates. These coordinates do
	 * not include a subpixel position.
	 */
	SDL_Rect loc;
};

typedef struct Obstacle_t  Obstacle;

/**
 * A pointer to an array of obstacles that are in the game enviornment.
 */
extern Obstacle *obstacles;

/**
 * Makes all obstacles visible by setting up their internal RenderItems.
 * Collision code will work without calling this function.
 * @pre
 *    - ::obstacles is allocated.
 *    - gameConfig::obstacles is set to the number of obstacles.
 *    - The Obstacle::loc field is filled for each obstacle.
 *
 * @post  Each obstacle's RenderItem is initialized and added to the render
 *        system.
 *
 * @author  Jeff Jackowski
 */
void ObstacleInit();

/**
 * Uninitializes the obstacles. This function can be safely called when there
 * are no obstacles.
 * @post 
 *    - The RenderItem for each obstacle is removed.
 *    - The ::obstacles array is deallocated.
 *    - gameConfig::obstacles is set to zero.
 *
 * @author    Jeff Jackowski
 */
void ObstacleUninit();

/**
 * Checks for a collision between a player's tank and an obstacle. The
 * algorithm first simultaneously finds the rectangular overlapping region
 * while checking for any overlap at all. If there is some overlap, pixel
 * collision is performed on the player's sprite. The obstacle is assumed to
 * be a solid rectangular region.
 * @param  p  The player that might collide.
 * @param  o  The obstacle that might run into the player.
 * @return    True if a collision has occured.
 * @author    Jeff Jackowski
 */
Bool IsCollidingObstacle(Player *p, Obstacle *o);

