/**
 * @file player.h
 * Player data.
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#ifndef __player_h__
#define __player_h__

#include "render.h"
#include "genconmac.h"
#include "text.h"

/**
 * Stores the time of the last acknowledged update for each of the different
 * parts of the player data.
 * @deprecated  This was used with an earlier revision of the message system.
 *              A new revision may use it, so the struct will remain for now.
 * @ingroup Networking
 */
struct UpdateTimes_t {
	/**
	 * The time of the last known tank position update.
	 */
	Uint32 tankPosTime;
	/**
	 * The time of the last known shell position update.
	 */
	Uint32 shellPosTime;
	/**
	 * The time of the last known update to dead time, flags, and score.
	 */
	Uint32 dfsTime;
	/**
	 * The time of the last known identity update.
	 */
	Uint32 identTime;
};

typedef struct UpdateTimes_t  UpdateTimes;

/**
 * @defgroup GamePlayers Players
 * Maintains data on the players and their tanks in the game.
 *
 * @{
 */

/**
 * Player status flags for use with Player::flags. The lower 8 flags are sent
 * by both the server and client. The upper 8 flags are only sent by the
 * server.
 */
enum PLAYER_FLAGS {
	/**
	 * Forward motion flag.
	 */
	PLAYER_FORWARD        = 1,
	/**
	 * Backward motion flag.
	 */
	PLAYER_BACKWARD       = 2,
	/**
	 * Right rotation flag.
	 */
	PLAYER_ROTRIGHT       = 4,
	/**
	 * Just started right rotation flag.
	 */
	PLAYER_STARTROTRIGHT  = 8,
	/**
	 * Left rotation flag.
	 */
	PLAYER_ROTLEFT        = 0x10,
	/**
	 * Just started left rotation flag.
	 */
	PLAYER_STARTROTLEFT   = 0x20,
	/**
	 * True when this player's shell is in flight.
	 */
	PLAYER_SHOT           = 0x40,
	/**
	 * True if the tank uses pong shells.
	 * @warning  The pong shell feature is unimplemented.
	 */
	PLAYER_PONG           = 0x80,
	/**
	 * Indicates that the local player's identity has been updated and needs
	 * to be sent to the server.
	 */
	PLAYER_UPDATE_IDENT   = 0x100,
	/**
	 * Indicates that the tank position and/or orientation data has been
	 * changed and needs to be sent to the server. This flag should be set
	 * whenever the player's tank movement input changes and when tanks
	 * collide with tanks.
	 */
	PLAYER_UPDATE_TANK    = 0x200,
	/**
	 * Indicates the local player's shell has been fired and the sever has not
	 * yet been told of the event. The server will update the shell's position,
	 * direction, and collisions without further updates from the client.
	 */
	PLAYER_UPDATE_SHELL   = 0x400,
	/**
	 * Indicates a change in dead time, flags, and/or score.
	 */
	PLAYER_UPDATE_DFS     = 0x800,
	/**
	 * Dead flag.
	 */
	PLAYER_DEAD           = 0x1000,
	/**
	 * True if this player is in use; false if this data does not yet
	 * represent a player. This flag has no meaning in messages sent over
	 * the network.
	 */
	PLAYER_ACTIVE         = 0x2000,
	/**
	 * True when a network update must be sent.
	 * @deprecated  This information will now be stored in client and
	 *              server specific data structures.
	 */
	PLAYER_UPDATEREQ      = 0x4000,
	/**
	 * All tank & movement related flags.
	 */
	PLAYER_MOVEMENT_MASK  = 0xFF,
	/**
	 * A mask for the flags that trigger a movement sound.
	 */
	PLAYER_MOVEMENT_SOUND = 3,
	/**
	 * The mask of all the flags that the client can update, save for movement
	 * related flags.
	 */
	PLAYER_CLIENT_MASK    = 0x700,
	/**
	 * A mask of most of the flags that the server can update for the local
	 * client. The server can update all flags for any other client. The
	 * following flags should not be updated for the local client, but are in
	 * the mask so that the read function can move past the data:
	 * - PLAYER_UPDATE_TANK
	 * - PLAYER_UPDATE_SHELL
	 */
	PLAYER_LOCAL_MASK    = ~0x207F,
	/**
	 * A mask for the flags used to indicate the items being updated.
	 */
	PLAYER_UPDATE_ITEMS  = 0xF00
};

/**
 * The data describing a player in the game and player options. For options,
 * the first element in players, an array of Player, will contain the
 * settings that must be sent to the server.
 * @author Jeff Jackowski
 */
struct Player_t {
	/**
	 * The graphical represntation of the player's tank on the field.
	 * @note  This field should not be sent over the network.
	 */
	RenderItem tank;
	/**
	 * The graphical represntation of the shell shot from the tank.
	 * @note  This field should not be sent over the network.
	 */
	RenderItem shell;
	/**
	 * The small label used to identify the player operating the tank.
	 * @note  This field should not be sent over the network.
	 */
	RenderItem labelName;
	/*
	 * A larger label used on the score listing.
	 * @note  This field should not be sent over the network.
	 */
	//RenderItem largeName;
	/*
	 * The graphic the number for the player's score.
	 * @note  This field should not be sent over the network.
	 */
	//RenderItem scoreText;
	/**
	 * The times, in common game time, of when this player was last updated
	 * with information from the network.
	 * @todo  Change this to use message IDs so that local data will not be
	 *        overwritten with newer server data from a resent message.
	 */
	UpdateTimes lastUpdate;
	/**
	 * The time, in common game time, of when this player was last locally
	 * updated.
	 */
	Uint32 localUpdate;
	/**
	 * The player's current location. This value includes an 8-bit subpixel
	 * location as the least significant byte.
	 */
	Sint32 x, y;
	/**
	 * The current location of the shell. This value includes an 8-bit subpixel
	 * location as the least significant byte.
	 */
	Sint32 shellX, shellY;
	/**
	 * The distance the shell has traveled. Actually, the number of frames it
	 * has been in flight.
	 */
	Uint16 shellDist;
	/**
	 * The direction the shell is traveling.
	 */
	Uint16 shellRot;
	/**
	 * The direction of the tank. The index of the tank sprite is 
	 * (rot >> 8) & 0xF.
	 */
	Uint16 rot;
	/**
	 * The number of kills the player has scored.
	 */
	Sint16 kills;
	/**
	 * The number of flag captures the player has scored.
	 */
	Sint16 captures;
	/**
	 * Bit flags from ::PLAYER_FLAGS.
	 */
	Uint16 flags;
	/**
	 * The player's name. This may be stored using the platform's wide character.
	 * It is transmitted over the network in UTF-16.
	 */
	rtsa_char name[MAX_NAME_LEN];
	/**
	 * The ID number of the player. This should correspond to its array position
	 * and is useful when the player is passed as a pointer.
	 */
	Uint8  pid;
	/**
	 * The team that the player joined.
	 */
	Uint8  team;
	/**
	 * The number of frames a player has been dead. When the value exceeds
	 * ::DEAD_TIME, the player can respawn and the spinning tank will no longer
	 * be drawn.
	 */
	Uint8  deadTime;
};

typedef struct Player_t  Player;

/**
 * The player ID for the player that is also the user of this process. For
 * servers, the value should always be zero.
 */
extern Uint32 localPlayer;

/**
 * The set of update flags used to track locally made changes to the local
 * player so that the updated parts can be transmitted to the server. Local
 * changes to other player data is not tracked by the client because the
 * server is authoritative source for thier information.
 */
extern Uint8 localUpdateFlags;

/**
 * The global array of player data. The data is initialized to zeros at program
 * start.
 */
extern Player players[MAX_PLAYERS];

/**
 * Initializes the graphics items required to render a player.
 * @param p  The player who's graphics items need to be initialized.
 * @pre  The Player's RenderItems must not yet be initialized, and the
 *       following fields of the player must already be filled:
 *       - Player::name
 *       - Player::team
 *       - Player::pid
 *       - PLAYER_FLAGS::PLAYER_DEAD in Player::flags
 *
 *       Optionally, the x and y position can also be filled.
 *
 * @post  The Player's RenderItems are initilized and added to the appropriate
 *        layers. They are placed only if the player is not flagged as dead.
 *        The tank RenderItem is positioned to match the x and y coordinates.
 * @author  Jeff Jackowski
 */
void AddPlayer(Player *p);

/**
 * Removes a player from the game by marking the player as inactive and
 * destroying related player data.
 * @todo  Implement.
 */
void RemovePlayer(Player *p);

/** @} */

#endif

