/**
 * @file playernet.h
 * Network communication interface for player update data.
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "player.h"
#include "message.h"

/**
 * Adds update information for another player to a player update message.
 *
 * @param msg   A pointer to an existing player update message to which the
 *              player data will be appended, or NULL if a new message should be
 *              created. If the given message is too small to fit the indicated
 *              player data, a new message will be created.
 * @param p     A pointer to the player data to add to the message.
 * @param time  The common game time of the update.
 * @return      A pointer to the message used to store the player. This may be
 *              different from the message passed in as @a msg. The value is
 *              NULL on error.
 * @ingroup GamePlayers
 * @author  Jeff Jackowski
 */
MsgDescr *AppendPlayerData(MsgDescr *msg, Player *p, Uint32 time);

/**
 * Processes incomming player data. For servers, a single player update is read
 * from a single client. For clients, one or more player updates are read.
 *
 * @param origin    The originating address of the message.
 * @param buffer    The buffer that contains the message to process.
 * @param clientId  A pointer to the client ID that needs an acknowledgment.
 * @param noid      A required but unused part of the function prototype.
 * @return          True on success, false on error.
 * @ingroup GamePlayers
 * @author  Jeff Jackowski
 */
Bool HandlePlayerUpdate(SocketAddr *origin, NetBuffer *buffer, Uint8 *clientId,
Uint16 noid);

/**
 * Sends a request to the server to spawn the local player.
 * @pre  The process is a client, and the player's tank has been destroyed
 *       (stopped spinning).
 * @ingroup GamePlayers
 * @author  Jeff Jackowski
 */
Bool SendSpawnRequest();

/**
 * Processes an incomming client request to spawn their tank.
 *
 * @param origin    The originating address of the message.
 * @param buffer    The buffer that contains the message to process.
 * @param clientId  A pointer to the client ID that needs an acknowledgment.
 * @param noid      A required but unused part of the function prototype.
 * @return          True on success, false on error.
 * @ingroup GamePlayers
 * @author  Jeff Jackowski
 */
Bool HandlePlayerSpawn(SocketAddr *origin, NetBuffer *buffer, Uint8 *clientId,
Uint16 noid);
