/**
 * @file render.h
 * The header file for a fairly optimized 2D rendering system built on SDL.
 * @ingroup RenderSys
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#ifndef _render_h_
#define _render_h_

#include <SDL.h>
#include "bool.h"

/**
 * @defgroup RenderSys  Render System
 * A system to perform fairly optimized 2D rendering by only redrawing what has
 * changed every frame rather than the whole frame.
 *
 * This is the generalized flow of execution of a program using this rendering
 * system:
 * - Initalize SDL
 * - Initalize the rendering system by calling RenderInit().
 * - Perform the following loop for each frame:
 *   - Input processing and other application specific work.
 *   - Change display information. Required after a window resize or change in
 *     resolution. Use RenderClearStates() only when needed.
 *   - If it is known the entire frame must be redrawn and the previous step
 *     was skipped, call ForceRenderAll().
 *   - Remove render items. This can be done earlier, but not after placing
 *     items. Use RemoveRenderItem().
 *   - Clear render items. This can be done earlier, but not later.
 *     Use ClearRenderItem().
 *   - Add new render items by calling AddRenderItem().
 *   - Place all moved render items by changing RenderItem::loc and then
 *     calling PlaceRenderItem().
 *   - Change layer offsets by calling SetLayerOffset(). For best performance,
 *     change the location of the items in the affected layer first by changing
 *     RenderItem::loc and do not call PlaceRenderItem() on those items. After
 *     that, call SetLayerOffset().
 *   - Render a frame by calling Render().
 * - Uninitalize the rendering system by calling RenderUninit().
 * - Uninitalize SDL
 *
 * Macros that affect the renderers operation:
 * - RENDER_INDIVIDUAL_UPDATES -- When defined, the renderer will not perform a
 *   batch update on a single-buffered display. Instead, it will call
 *   SDL_UpdateRect() immediately after an item's render function is called.
 *   This can help debugging by making visible changes to the display as they
 *   are drawn.
 *
 * The system allows for single and double buffered operation. When single
 * buffered operation is performed on a windowed GUI (X11, Windows), the
 * operation is very close to double buffered and lacks flicker. It is still a
 * single buffered setup because the program will only have access to one
 * buffer. Everything is drawn to this buffer, and from there is copied to
 * the frame buffer that only the windowed GUI can access. The buffer the
 * program draws onto is not in video memory, is not visible, and therefore
 * no flicker will occur for the same reasons as using double buffering.
 *
 * Single buffered operation usually is required when hardware graphics
 * acceleraion is not available. Hardware acceleration may or may not be
 * availble with double buffering. Care must be taken with placing SDL surfaces
 * in graphics or system memory to achive the best performance. Always favoring
 * graphics memory will ensure that some operations will be slow and some
 * systems will run the program slow. Search online for more information. When
 * running without hardware acceleration, this is usually a non-issue becuase
 * graphics memory is not availble. In that case, SDL's optimized software
 * blits (espically between surfaces with the same format) and the rendering
 * system's ability to minimize updates to the display can together yield
 * decent to excellent performance.
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/ 
 * @{
 */

#ifndef BACK_STATES
/**
 * The number of previous rendered states stored. The value must be at least
 * 1 for single buffered output and 2 for double buffered output.
 * @warning Presently
 * values greater than 2 are not fully supported; error directives will point
 * out the spots if BACK_STATES is set to 3 or more. Should none of the code
 * marked by error directives
 * cause trouble, consider the result untested.
 */
#define BACK_STATES  2
#endif

#ifndef MAX_LAYERS
/**
 * The total number of layers that will be available for rendering.
 */
#define MAX_LAYERS  6
#endif

/**
 * Contains globally accessible information about the display. main() must
 * change this when the screen settings or window size change. The drawing
 * functions need it for the colors and the resolution setting.
 * @author  Jeff Jackowski
 */
struct DisplayInfo_t {
	/**
	 * The surface that represents the viewable area presented to the user.
	 */
	SDL_Surface *display;
	/**
	 * The width of the display window in pixels.
	 * @warning  Any change of this value after initalizing the render system
	 *           should be accompanied by a call to RenderClearStates().
	 */
	Uint16 pw;
	/**
	 * The height of the display window in pixels.
	 * @warning  Any change of this value after initalizing the render system
	 *           should be accompanied by a call to RenderClearStates().
	 */
	Uint16 ph;
	/**
	 * The width of the display window in logical positions.
	 * @note  This field is specific to this game and is not used by the
	 *        rendering library.
	 */
	Uint16 lw;
	/**
	 * The height of the display window in logical positions.
	 * @note  This field is specific to this game and is not used by the
	 *        rendering library.
	 */
	Uint16 lh;
	/**
	 * Width in physical positions of one logical unit.
	 * @note  This field is specific to this game and is not used by the
	 *        rendering library.
	 */
	Uint16 luw, luh;
	union {
		/**
		 * A group of global application specific flags that relate to what and
		 * how things are rendered.
		 */
		Uint8 flags;
		struct {
			/**
			 * Render the game instead of the menu. Needed because the game
			 * simulation loop in UpdatePlayers() is called whenever the process
			 * is a server or is connected to a server.
			 */
			Bool showGame     : 1;
			/**
			 * If true, the names of the players will be drawn above their
			 * tank.
			 * @note  This field is specific to this game and is not used by the
			 *        rendering library.
			 */
			Bool showLabels   : 1;
			/**
			 * True when the battle field is too large to fit on screen so it
			 * must scroll with the player's movement.
			 * @note  This field is specific to this game and is not used by the
			 *        rendering library.
			 */
			Bool scrollField   : 1;
		};
	};
};

/**
 * Global information about the display.
 */
extern struct DisplayInfo_t  dInfo;

typedef struct RenderItem_t  RenderItem;

typedef struct Layer_t  Layer;

/**
 * Evaluates to the fastest (hopefully) type of surface, hardware or software,
 * at runtime.
 * @author  Jeff Jackowski
 */
#define FASTEST_SURFACE  ((dInfo.display->flags & SDL_HWACCEL) ? \
                       SDL_HWSURFACE : SDL_SWSURFACE)

/**
 * A function pointer type that references a function called upon to render an
 * item. Allows great flexibility with what can be rendered. For items flagged
 * as transparent, the region of the target surface where the item should be
 * drawn already has current image data for layers beneath the one holding the
 * item to be rendered.
 * @param dest  The surface that is the target of the rendering operation. 
 * @param item  The RenderItem that the function is expected to render.
 *
 * relatesalso RenderItem_t
 * @author  Jeff Jackowski
 */
typedef void (*ItemRenderer)(SDL_Surface *dest, RenderItem *item);

/**
 * Renders a surface pointed to by item->data. To avoid clipping the source
 * and destination rectangles again (already clipped by the render system),
 * SDL_LowerBlit() is called.
 * @param dest  The surface that is the target of the rendering operation. 
 * @param item  The RenderItem that the function is expected to render. Its
 *              data field (RenderItem::data) must point to a valid
 *              SDL_Surface.
 *
 * relatesalso RenderItem_t
 * @author  Jeff Jackowski
 */
void GenericSurfaceRenderer(SDL_Surface *dest, RenderItem *item);

/**
 * Renders a solid rectangle who's color is stored at item->data.
 * @param dest  The surface that is the target of the rendering operation. 
 * @param item  The RenderItem that the function is expected to render. Its
 *              data field (RenderItem::data) must point to an Uin32 with a
 *              valid device specific color code.
 *
 * relatesalso RenderItem_t
 * @author  Jeff Jackowski
 */
void SolidFillRenderer(SDL_Surface *dest, RenderItem *item);

/**
 * The fundamental unit for a visible rectangular region. Each RenderItem can
 * only be used with one Layer.
 * @author  Jeff Jackowski
 */
struct RenderItem_t {
	/**
	 * The function which renders the item.
	 */
	ItemRenderer renderer;
	/**
	 * The owning layer of this item.
	 */
	Layer *parent;
	/**
	 * A pointer to the next item in the list of all visible items.
	 */
	RenderItem *next;
	/**
	 * A pointer to the previous item in the list of all visible items.
	 */
	RenderItem *prev;
	/**
	 * A pointer to application defined data.
	 */
	void *data;
	/**
	 * The logical (world) location of the item relative to the layer. The
	 * rendering system will not set or change the value of loc.
	 */
	SDL_Rect loc;
	/**
	 * The region of the item that must be rendered in the next frame. The
	 * rendering system will change the value of src to clip it to the viewable
	 * area and to extend it if needed to draw over an area made dirty by
	 * another update.
	 */
	SDL_Rect src;
	/**
	 * The destination coordinates on the physical display. This value is
	 * clipped to the display and to the union of region that has changed and
	 * regions that require an update because of other items changing. At the
	 * time renderer is called, src will contain the item relative coordinates
	 * of what must be drawn onto the rectangle in physDest.
	 */
	SDL_Rect physDest;
	/**
	 * Stores physical coordinates of previously visible locations. Index zero
	 * stores the location for the last frame, index one stores the frame
	 * before the last, ect . . . The width field must be zero if the item
	 * was not visible in the corresponding frame.
	 */
	SDL_Rect prevStates[BACK_STATES];
	union {
		/**
		 * A set of bit flags used to hold render state information.
		 */
		Uint32 flags;
		struct {
			/**
			 * Flag that indicates the item has changed since last drawn and
			 * the change is visible.
			 * @internal
			 */
			Bool dirty       : 1;
			/**
			 * Flag that indicates the item has transparency.
			 * @warning  Not yet implemented; all items are assumed to have
			 *           transparency. Implementing this could achive a performance
			 *           bonus for opaque items.
			 */
			Bool transparent : 1;
			/**
			 * Flag that indicates the item is within the viewable area.
			 */
			Bool visible     : 1;
			/**
			 * Flag that indicates the item was placed with PlaceRenderItem()
			 * since the last frame was rendered. This helps avoid duplicating
			 * calculations during rendering.
			 * @internal
			 */
			Bool placed      : 1;
			/**
			 * Flag that indicates the item should be drawn; this flag is
			 * not cleared in Render().
			 */
			Bool enableDraw   : 1;
		};
	};
	/**
	 * The physical display location where the item is placed. The width and
	 * height are no different than the size of the item as stored in loc.
	 * @warning  This location (upper left) is not clipped to the display area
	 *           by PlaceRenderItem(). Doing so would complicate figuring out
	 *           the item relative coordinates for src in the rendering process.
	 */
	Sint16 physLocX, physLocY;
	/**
	 * The index of the rect in the update list which contains this item. This
	 * is used to prevent confusion and multiple renderings of the item in the
	 * case that there are overlapping rects in the update list.
	 */
	Sint16 rectIndex[BACK_STATES];
};

/**
 * Describes a layer for rendering. Layers provide a Z-ordering for 2D graphics.
 * The ordering of overlapping items in the same layer is not guaranteed. Each
 * layer can be independently offset. Layer coordinates are translated into
 * physical coordinates by adding the offset values.
 *
 * @todo    Add a clipping rectangle that affects only owning layer.
 *
 * @author  Jeff Jackowski
 */
struct Layer_t {
	/**
	 * The first item of the list of all items present on the layer.
	 */
	RenderItem *worldListFirst;
	/**
	 * The last item of the list of all items present on the layer.
	 */
	RenderItem *worldListLast;
	/**
	 * The horizontal offset value.
	 * @warning  This value must not change from the time PlaceRenderItem() is
	 *           called until after the frame has been rendered. Any change
	 *           must also include setting the dirty flag.
	 */
	Sint16 offX;
	/**
	 * The vertical offset value.
	 * @warning  This value must not change from the time PlaceRenderItem() is
	 *           called until after the frame has been rendered. Any change
	 *           must also include setting the dirty flag.
	 */
	Sint16 offY;
	union {
		/**
		 * Bit flags used with layers.
		 */
		Uint16 flags;
		struct {
			/**
			 * Marks a layer as dirty when something on the layer changes or the
			 * layer's offset changes and an item on the layer is visible.
			 * @internal
			 */
			Bool dirty         : 1;
		};
	};
};

/**
 * The global set of Layers. Lower values are further back in the Z-order. See
 * ::LAYERS for a list of valid layers and their intended use.
 */
extern Layer layers[MAX_LAYERS];

/**
 * Initalizes the rendering system.
 * @warning  This function should only be called once before a call to
 *           RenderUninit().
 * @param numRects  The maximum number of rectangular regions that will ever
 *                  need to be updated between frames. This number must be
 *                  greater than zero, and should be a good bit higher to
 *                  render most things.
 * @param numStates The number of visual states the renderer should use. To
 *                  assure correct operation, use one state for single-buffered
 *                  output, two for double-buffered output, ect... The value
 *                  must be greater than 0 (if you've got no states then you
 *                  don't need to render) and not greater than ::BACK_STATES.
 * @return  True if the operation succeeded. Inadequate memory and invalid
 *          parameters can cause a failure.
 * @author  Jeff Jackowski
 */
Bool RenderInit(Uint16 numRects, Uint8 numStates);

/**
 * Clears all display state information and optionally changes the number of
 * states to use. This is required after changing the size of the window or
 * the display resolution, and if there is a change between single and double
 * buffered output.
 *
 * @todo  Implement this function.
 *
 * @param numStates  The number of visual states the renderer should use, or
 *                   zero if the value should not change. See RenderInit() for
 *                   more details.
 * @author  Jeff Jackowski
 */
void RenderClearStates(Uint8 numStates);

/**
 * Uninitializes the rendering system. Deallocates all memory allocated by the
 * rendering system.
 * @author  Jeff Jackowski
 */
void RenderUninit();

/**
 * Force everything to be redrwan for the next frame. For best performance,
 * call this prior to clearing, removing, or placing rener items. The extra
 * logic to compute the changed regions on screen will be skipped. Only affects
 * the next frame.
 * @author  Jeff Jackowski
 */
void ForceRenderAll();

/**
 * Tells if the next call to Render() will render an entire frame rather than
 * just a changed portion.
 * @return  True if a whole frame will be rendered, false if only a partial
 *          frame will be rendered.
 * @author  Jeff Jackowski
 */
Bool RenderWillRenderAll();

/**
 * Set the dirty flag to force the renderer to perform a partial redraw for the
 * next frame. This is not needed after calls to PlaceRenderItem() and
 * ClearRenderItem(), but is needed after changing layer offsets. The macros
 * LayerOffset(), LayerHorizOffset(), and LayerVertOffset() all call this
 * function.
 * @author  Jeff Jackowski
 */
void ForceDirtyFrame();

/**
 * Adds a new RenderItem to the world list of a Layer. It will initally be set
 * to not draw. Call PlaceRenderItem() next to position the item on screen
 * and draw it on the next call to Render().
 * @pre   @a item is not a member of another layer, and its parent pointer is
 *        NULL.
 * @param item  The item to add to the list of items a layer contains.
 *              item's src field (RenderItem::src) will be initalized such that
 *              it will indicate the entire item will need to be drawn. For
 *              this to be valid when PlaceRenderItem() is called, the item's
 *              width and height in loc (RenderItem::loc) must be set prior to
 *              calling this function.
 * @param layer The index of the Layer. For RTSA, they are enumerated in
 *              ::LAYERS.
 *
 * relatesalso Layer_t
 * @author  Jeff Jackowski
 */
void AddRenderItem(RenderItem *item, int layer);

/**
 * Removes a RenderItem from a Layer.
 * @pre   The item is not visible, either by being positioned off-screen, or
 *        through a call to ClearRenderItem(). Or where the item was visible
 *        will be redrawn through other means.
 *
 * @param item  The item to remove.
 *
 * relatesalso Layer_t
 * @author  Jeff Jackowski
 */
void RemoveRenderItem(RenderItem *item);

/**
 * Places a render item on the physical display based on information set within
 * the item. Avoid calling this function more than once per frame per item.
 * @warning  Changing the position on an item without calling PlaceRenderItem()
 *           can cause incorrect rendering of multiple items.
 *
 * @post The fields used to place the item are:
 * - RenderItem::loc - the new Layer relative location
 * - RenderItem::src - the changed region relative to the item
 *
 * @pre The fields updated by this call are:
 * - RenderItem::physLocX
 * - RenderItem::physLocY
 * - RenderItem::physDest
 * - RenderItem::src
 * - RenderItem::flags
 * - Layer::flags
 * - UpdateList_t::rects
 * - UpdateList_t::numRects
 * - UpdateList_t::flags
 *
 * relatesalso RenderItem_t
 * @author  Jeff Jackowski
 */
void PlaceRenderItem(RenderItem *item);

/**
 * Makes an item no longer visible until the next PlaceRenderItem() call on
 * the item. More efficent than calling
 * PlaceRenderItem() after putting the item out of view. Does not change the
 * location of the item. The best performance with the current implementation
 * is achived if all ClearRenderItem() calls are made before any
 * PlaceRenderItem() calls prior to the frame render.
 * @param   item  The render item which will not be visible in the next frame.
 *
 * relatesalso RenderItem_t
 * @author  Jeff Jackowski
 */
void ClearRenderItem(RenderItem *item);

/**
 * Changes the layer's offset from the physical display. This offset is added
 * to the coordinates of the layer's items. For best performance, change the
 * location of the items in the affected layer first by changing
 * RenderItem::loc and do not call PlaceRenderItem() on those items. After
 * that, call SetLayerOffset().
 * @param  layer  A pointer to the layer to modify.
 * @param  x      The new horizontal offset.
 * @param  y      The new vertical offset.
 *
 * relatesalso  Layer_t
 * @author   Jeff Jackowski
 */
void SetLayerOffset(Layer *layer, Sint16 x, Sint16 y);

/**
 * Renders a frame.
 *
 * Issues:
 * - Recomputes a number of values computed in PlaceRenderItem() in case other
 *   items placed since then require more of the item to be rendered.
 * - If values in PlaceRenderItem() were only computed here, it might limit the
 *   possibility of rendering bugs (haven't seen them yet, though) and could
 *   reduce recomputation, but will require collision checks between all items
 *   and all update rects. In the present code, those collision checks are not
 *   performed for items that have been placed.
 *
 * @author  Jeff Jackowski
 */ 
void Render();

/**
 * A utility function for resizing the program's window. It will use flags
 * stored in videoFlags, a Uint32 defined non-static in render.c. Set the
 * value to the flags needed.
 * @param   w  The new width of the window.
 * @param   h  The new height of the window.
 * @return  True on sucess, false if the window could not be resized.
 * @author  Jeff Jackowski
 */
Bool ResizeWindow(Uint16 w, Uint16 h);

/**
 * Clips a rectangle to fit within the given bounding rectangle.
 * @param clip   A pointer to the SDL_Rect to be clipped. It will be modified.
 *               If the width or height equals zero, then the input clip did
 *               overlap the bounding rectangle.
 * @param limit  A pointer to the bounding rectangle.
 * @author Jeff Jackowski
 */
#define ClipRectToRect(clip, limit)  { \
	int t; \
	if ((t = (limit)->x - (clip)->x) > 0) { \
		(clip)->w -= t; \
		(clip)->x = (limit)->x; \
	} \
	if ((t = ((clip)->x + (Sint16)(clip)->w) - ((limit)->x + (limit)->w)) > 0) { \
		(clip)->w -= t; \
	} \
	if ((Sint16)((clip)->w) < 0) { \
		(clip)->w = 0; \
	} \
	if ((t = (limit)->y - (clip)->y) > 0) { \
		(clip)->h -= t; \
		(clip)->y = (limit)->y; \
	} \
	if ((t = ((clip)->y + (Sint16)(clip)->h) - ((limit)->y + (limit)->h)) > 0) { \
		(clip)->h -= t; \
	} \
	if ((Sint16)((clip)->h) < 0) { \
		(clip)->h = 0; \
	} \
}

/**
 * Clips a rectangle to fit within a physical region defined by a width and a
 * height. Works like ClipRectToRect() but assumes limit's x & y coordinates are
 * zero. Should be more efficent than using ClipRectToRect().
 * @param clip   A pointer to the SDL_Rect to be clipped. It will be modified.
 *               If the width or height equals zero, then the input clip did
 *               overlap the bounding rectangle.
 * @param width    The maximum width.
 * @param height   The maximum height.
 * @author Jeff Jackowski
 */
#define ClipRectToWH(clip, width, height) { \
	int t; \
	if ((clip)->x < 0) { \
		(clip)->w += (clip)->x; \
		(clip)->x = 0; \
	} \
	if ((t = ((clip)->x + (Sint16)(clip)->w) - (width)) > 0) { \
		(clip)->w -= t; \
	} \
	if ((Sint16)((clip)->w) < 0) { \
		(clip)->w = 0; \
	} \
	if ((clip)->y < 0) { \
		(clip)->h += (clip)->y; \
		(clip)->y = 0; \
	} \
	if ((t = ((clip)->y + (Sint16)(clip)->h) - (height)) > 0) { \
		(clip)->h -= t; \
	} \
	if ((Sint16)((clip)->h) < 0) { \
		(clip)->h = 0; \
	} \
}

/** @} */

#endif

