/**
 * @file states.c
 * Implements a system to use and switch between the different major game
 * operating states.
 * @ingroup OperStates
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include <assert.h>
#include <stddef.h>
#include "bflderrstate.h"
#include "states.h"
#include "splashstate.h"
#include "gamestate.h"
#include "menustate.h"
#include "menu.h"

/**
 * Stores the current operating state.
 * @internal
 * @ingroup OperStates
 */
static int currState = STATE_SPLASH;

/**
 * Tracks if the current state has been initalized.
 * @internal
 * @ingroup OperStates
 */
static Bool initalized = FALSE;

/**
 * A stub implementation for unimplemented states. Moves the system to the next
 * state. If there is no next state, quit.
 * @internal
 * @ingroup OperStates
 * @author Jeff Jackowski
 */
static int Stub() {
	int state = currState + 1;
	if (state == STATE_MAX)
		return STATE_QUIT;
	return state;
}

/**
 * The list of functions used for the operation state implementations.
 * @internal
 * @ingroup OperStates
 */
static const struct State_t stateImpls[STATE_MAX] = {
	{ SplashInit, SplashRun, SplashUninit },        // STATE_SPLASH
	{ MenuRootInit, MenuRun, MenuRootUninit },      // STATE_MENUROOT
	{ NULL, Stub, NULL },                           // STATE_MENUNAME
	{ NULL, Stub, NULL },                           // STATE_MENUNET
	{ NULL, Stub, NULL },                           // STATE_MENUCONN
	{ NULL, Stub, NULL },                           // STATE_SHOWOPTS
	{ GameInit, GameRun, GameUninit },              // STATE_PLAY
	{ BfldErrInit, BfldErrRun, BfldErrUninit },     // STATE_BFLDERR
	{ NULL, Stub, NULL }                            // STATE_DISCONN
};

int GetCurrentState() {
	return currState;
}

Bool RunState() {
	int result;
	// check for an uninitalized state
	if (!initalized) {
		// initialize the state
		if (stateImpls[currState].init) {   // only run non-NULL functions
			result = (*stateImpls[currState].init)();
			// check for error
			if (result <= STATE_QUIT) {
				// non-recoverable error -- send back termination code
				return FALSE;
			}
			// failed init; recoverable error
			if (result < STATE_MAX) {
				// set the new state
				currState = result;
				// continue running states later
				return TRUE;
			}
			// otherwise the init succeeded
			assert(result == STATE_OK);
		}
		// the state is initalized
		initalized = TRUE;
	}
	// run the state
	result = (*stateImpls[currState].run)();
	// check for state change
	if (result != currState) {
		assert(result < STATE_MAX);
		// call the uninitalize function and check for error
		if (stateImpls[currState].uninit &&   // only run non-NULL functions
		((*stateImpls[currState].uninit)() != STATE_OK)) {
			// non-recoverable error -- send back termination code
			return FALSE;
		}
		// the state is uninitalized
		initalized = FALSE;
		// set the new state
		currState = result;
	}
	// check for error or quit
	if (result < 0) {
		// quit now
		return FALSE;
	}
	// no errors and no request to quit, so keep going
	return TRUE;
}

