/**
 * @file states.h
 * Defines the different major game operating states.
 * @ingroup OperStates
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "bool.h"

/**
 * @defgroup OperStates Operating States
 *
 * Implements a system to switch between the
 * @ref OperatingStates "major operating states" of the program. Each state
 * includes an initialization, uninitialization, and run function. This
 * provides for an intuitive operation, however an implementation with only
 * one function per state would allow a slimmer state machine. In that case,
 * the init and uninit functions would be whole states that normally finish
 * in a single call. I think I'll do it that way next time.
 *
 * @{
 */

/**
 * Enumerates the different operating states of the game. Each has its own
 * input-output loop.
 */
enum STATE_CODES {
	/**
	 * A state signifying a non-recoverable error has occured.
	 */
	STATE_ERROR = -2,
	/**
	 * The state that signifies normal program termination.
	 */
	STATE_QUIT,
	/**
	 * The inital state of the game; a splash screen identifying the game while
	 * media loads and a connection to a server is attempted. The next state
	 * is STATE_MENUROOT.
	 */
	STATE_SPLASH,
	/**
	 * A menu allowing the user to chose a team or quit, and informing the user
	 * of the network connection status.
	 */
	STATE_MENUROOT,
	/**
	 * An item of the root menu for player name entry. Uses a custom event loop.
	 */
	STATE_MENUNAME,
	/**
	 * A menu for configuring networking options and, when running as the
	 * server, game options.
	 */
	STATE_MENUNET,
	/**
	 * An item of the networking menu for entering the hostname of a server.
	 * Uses a custom event loop.
	 */
	STATE_MENUCONN,
	/**
	 * Displays game options prior to the start of the game, and loads media
	 * needed only for game play.
	 */
	STATE_SHOWOPTS,
	/**
	 * Game play. The user may chose to quit the program, or the game in
	 * progress. If the user quits the current game, the next state will be
	 * STATE_MENU. If the connection to the server is lost, the next state
	 * will be STATE_DISCONN.
	 */
	STATE_PLAY,
	/**
	 * An error occured while loading the battlefield file; this state shows
	 * the error to the user.
	 */
	STATE_BFLDERR,
	/**
	 * A notice that the server is no longer accessible. The next state is
	 * STATE_MENUROOT.
	 */
	STATE_DISCONN,
	/**
	 * The number of defined states.
	 */
	STATE_MAX,
	/**
	 * A code indicating that initialization was successful.
	 */
	STATE_OK
};

/**
 * The function prototype for all implementations of the different operating
 * states.
 * @return  Varies. See State_t.
 */
typedef int (*StateFunc)();

/**
 * A set of function pointers to functions that implement an operating state.
 * @author  Jeff Jackowski
 */
struct State_t {
	/**
	 * The function called to initalize the state. It will be called before
	 * the run function. It can be set to NULL if no initalization is required.
	 * @return STATE_CODES::STATE_OK if the initialization succeeded, otherwise
	 *         another state code for the next state or an error condition.
	 */
	StateFunc init;
	/**
	 * The function called repeatedly to operate the state. It is expected to
	 * capture and respond to events, render graphics, and deal with frame
	 * rate issues.
	 * @return A value defined in ::STATE_CODES for the next state, a
	 *         non-recoverable error, or program termination. If the next state
	 *         is the current state, the state's init and uninit functions will
	 *         not be called.
	 */
	StateFunc run;
	/**
	 * The function that uninitalizes a state in preperation for changing
	 * states or program termination. It can be set to NULL if no
	 * uninitalization is required. If the state's run function ended with a
	 * non-recoverable error, its uninitalize function will still be called.
	 * @return STATE_CODES::STATE_OK on success, anything else on failure.
	 */
	StateFunc uninit;
};

/**
 * Returns the program's current operating state as defined in the
 * ::STATE_CODES enum.
 * @return  The state.
 * @author  Jeff Jackowski
 */
int GetCurrentState();

/**
 * Runs the state system. The order of events are:
 * -# Check for an uninitalized state (program start condition)
 *    -# Initialize the state (error possibility)
 * -# Run the state
 * -# Check for a state change, request to terminate, or error
 *    -# Uninitalize the state (error possibility)
 *    -# Set the next state
 * -# Check for a request to terminate, or an error, from running the state
 *    -# Return FALSE
 * -# Return TRUE
 *
 * @return  True if the program should continue calling RunSate(). False if
 *          the program should terminate.
 * @author  Jeff Jackowski
 */
Bool RunState();

/** @} */

