/**
 * @file tanksprite.h
 * Defintions of functions and data that deal directly with the tank sprites.
 * @ingroup TankSprite
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "player.h"

/**
 * @defgroup TankSprite Tank Sprite
 * Defines the sprites used for the tank, the function for drawing tanks, and
 * the function for tank to tank pixel collision.
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/ 
 * @{
 */

/**
 * Sprite data for all sixteen sprites. The data is generated by
 * GenTankSprites() at runtime. The first dimention indicates the rotation of
 * the tank sprite. The second dimention indicates the horizontal line of the
 * sprite. A set bit indictaes the pixel is part of the tank.
 */
extern Uint8 sprites[16][8];

/**
 * An array of deltas for the shells inital location that is organized by the
 * tank's rotation.
 * @author  Jeff Jackowski
 */
const Uint8 shellInitLoc[16][2];

/**
 * Generates all 16 tank sprites from 3 prototype sprites. The algorithm used
 * closely resembles the original tank rendering routine. Generating the
 * sprites helps my lazyness because I don't have to come up with the sprites
 * myself or make a tool to generate the sprites. Having each sprite simplifies
 * and speeds up the rendering and collision detection algorithms.
 * @note   Only needs to be called once.
 * @author Jeff Jackowski
 */
void GenTankSprites();

/**
 * Draws a player's tank. Does not check for visibility.
 * @bug    Does not clip its drawing to the changed region.
 * @param dest  The surface that is the target of the rendering operation. 
 * @param item  The RenderItem, and ::Player, that the function will render.
 *              @warning @a item must also be a Player.
 * @author   Jeff Jackowski
 */
void DrawTank(SDL_Surface *dest, RenderItem *item);

/**
 * Tank to tank collision detection. The algorithm first finds an overlap
 * region for the players and at the same time checks for an overlap. If there
 * is an overlap, collision detection is done on the pixels of the sprites
 * used for the players.
 * @param p1  A player that might be colliding.
 * @param p2  The player that the first might be colliding with.
 * @return  True if the players are colliding with each other, false otherwise
 * @author  Jeff Jackowski
 */
Bool IsCollidingTank(Player *p1, Player *p2);

/**
 * Tank to shell collision detection. The algorithm first computes where on
 * the tank's sprite the shell is located while also checking a bounding box
 * to see if the shell is anywhere on tank's sprite. If the shell is on the
 * tank's sprite, pixel collision will be computed.
 * @param tank   The player who's tank that might be hit by the shell.
 * @param shell  The player who's shell might hit the tank.
 * @return   True if the shell hit the tank, false otherwise.
 * @author   Jeff Jackowski
 */
Bool IsCollidingShell(Player *tank, Player *shell);

/** @} */
