/**
 * @file teamsel.c
 * Implementation of the team selection menu item.
 * @ingroup MenuTeamSel
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include "teamsel.h"
#include "color.h"
#include "states.h"
#include "tanksprite.h"
#include "menustate.h"
#include "net.h"
#include <string.h>


/**
 * The space, in pixels, between the two team columns.
 * @ingroup MenuTeamSel
 */
#define MIDDLE_MARGIN   64

void TeamSelInit(MenuItem *item) {
	// setup the menu item fields
	item->minWidth = 320;
	item->minHeight = 240;
	//item->focus =
	item->input = TeamSelInput;
	item->update = TeamSelUpdate;
	item->active = TRUE;
	item->expandHeight = TRUE;
	// setup position of screen parts
	dInfo.luw = dInfo.luh = 6;
	dInfo.lw = dInfo.pw / dInfo.luw;
	dInfo.lh = dInfo.ph / dInfo.luh;
	// initalize two tanks
	menuSettings->teams[0].tank.renderer =
	menuSettings->teams[1].tank.renderer = DrawTank;
	menuSettings->teams[0].tank.loc.w = menuSettings->teams[0].tank.loc.h =
	menuSettings->teams[1].tank.loc.w = menuSettings->teams[1].tank.loc.h =
		dInfo.luw << 3;
	menuSettings->teams[0].rot = 4 << 8;
	menuSettings->teams[1].rot = 12 << 8;
	menuSettings->teams[1].team++;
	AddRenderItem(&(menuSettings->teams[0].tank), LAYER_TANKS);
	AddRenderItem(&(menuSettings->teams[1].tank), LAYER_TANKS);
	// set the colors
	menuSettings->selColor[TEAM_RED] = colorVals[COLOR_RED];
	menuSettings->selColor[TEAM_BLUE] = colorVals[COLOR_BLUE];
	menuSettings->deselColor[TEAM_RED] = ColorBlend(colorFmtInd[COLOR_RED],
	colorFmtInd[COLOR_REGBACK], 2, 3);	
	menuSettings->deselColor[TEAM_BLUE] = ColorBlend(colorFmtInd[COLOR_BLUE],
	colorFmtInd[COLOR_REGBACK], 2, 4);
	// set a condition that will cause tank zero to be selected
	menuSettings->team = 0;
	// set tank 1's deselected color
	colorVals[COLOR_BLUE] = menuSettings->deselColor[TEAM_BLUE];
}

void TeamSelUninit(MenuItem *item) {
	int loop;
	Player *p = players;
	// remove the tank render items used to denote the different teams
	RemoveRenderItem(&(menuSettings->teams[0].tank));
	RemoveRenderItem(&(menuSettings->teams[1].tank));
	// restore colors
	colorVals[COLOR_RED] = menuSettings->selColor[TEAM_RED];
	colorVals[COLOR_BLUE] = menuSettings->selColor[TEAM_BLUE];
	// clear all rendered names
	for (loop = MAX_PLAYERS; loop > 0; loop--, p++) {
		if ((p->flags & PLAYER_ACTIVE) && p->labelName.enableDraw) {
			ClearRenderItem(&(p->labelName));
		}
	}
}

int TeamSelUpdate(MenuItem *item) {
	Bool placeRemaining[TEAM_MAX] = { 0 };
	int loop, max;
	int top[TEAM_MAX];  // the top position for the next name of a teammember
	Player *p = players;
	// update the data with the latest information from the server
	// see if the visible components need to be moved
	if (!item->isPlaced) {
		// set the placed flag
		item->isPlaced = TRUE;
		// Move the components to the proper place.
		// Place the chose team prompt
		textItems[TEXTITEM_CHOSETEAM].loc.x = item->region.x +
		((item->region.w - textItems[TEXTITEM_CHOSETEAM].loc.w) >> 1);
		textItems[TEXTITEM_CHOSETEAM].loc.y = item->region.y;
		// place two tanks
		menuSettings->teams[0].tank.loc.y = menuSettings->teams[1].tank.loc.y =
		item->region.y + textItems[TEXTITEM_CHOSETEAM].loc.h + 4;
		menuSettings->teams[1].tank.loc.x = (menuSettings->teams[0].tank.loc.x =
		item->region.x + (item->region.w >> 1) - (MIDDLE_MARGIN >> 1) -
		menuSettings->teams[0].tank.loc.w) + MIDDLE_MARGIN +
		menuSettings->teams[0].tank.loc.w;
		// make both tanks visible
		PlaceRenderItem(&(menuSettings->teams[0].tank));
		PlaceRenderItem(&(menuSettings->teams[1].tank));
		// make the chose team title visible
		PlaceRenderItem(&(textItems[TEXTITEM_CHOSETEAM]));
	}
	// get the top position
	top[0] = top[1] = menuSettings->teams[0].tank.loc.y +
	menuSettings->teams[0].tank.loc.h;
	// set the maximum height value for player names
	max = item->region.y + item->region.h + 6;
	// loop through the players
	for (loop = MAX_PLAYERS; loop > 0; loop--, p++) {
		// only work on active players
		if (!(p->flags & PLAYER_ACTIVE)) continue;
		// see if the player doesn't fit but was visible
		if (placeRemaining[p->team] && p->labelName.enableDraw &&
		(top[p->team] >= (max - p->labelName.loc.h))) {
			// clear the player's name from view
			ClearRenderItem(&(p->labelName));
		}
		// see if the player will fit and is not visible or needs to be moved
		else if ((!p->labelName.enableDraw || placeRemaining[p->team]) &&
		(top[p->team] < (max - p->labelName.loc.h))) {
			if (p->team == TEAM_RED) {
				// set the vertical position
				p->labelName.loc.y = top[TEAM_RED];
				// advance the next position
				top[TEAM_RED] += p->labelName.loc.h + 2;
				// set the horizonal position -- right aligned
				p->labelName.loc.x = menuSettings->teams[TEAM_RED].tank.loc.x +
				menuSettings->teams[TEAM_RED].tank.loc.h - p->labelName.loc.h;
			}
			else {
				// set the vertical position
				p->labelName.loc.y = top[TEAM_BLUE];
				// advance the next position
				top[TEAM_BLUE] += p->labelName.loc.h + 2;
				// set the horizonal position -- left aligned
				p->labelName.loc.x = menuSettings->teams[TEAM_BLUE].tank.loc.x;
			}
			// render all others for this team
			placeRemaining[p->team] = TRUE;
			// place it
			PlaceRenderItem(&(p->labelName));
		}
		// some or all will still be in the same spot -- no need to render them
	}
	return 0;
}

int TeamSelInput(MenuItem *item, Uint8 *ec, rtsa_char keyval) {
	// check for left input when right team selected
	if ((ec[EC_ROTLEFT] == 1) && menuSettings->team) {
		// clear the input
		ec[EC_ROTLEFT]++;
		// chose the new team
		menuSettings->team = 0;
		// set tank 0's selected color
		colorVals[COLOR_RED] = menuSettings->selColor[TEAM_RED];
		// set tank 1's deselected color
		colorVals[COLOR_BLUE] = menuSettings->deselColor[TEAM_BLUE];
		// redraw both tanks
		goto TeamSelInput_redraw;
	}
	// check for right input when left team selected
	else if ((ec[EC_ROTRIGHT] == 1) && !menuSettings->team) {
		// clear the input
		ec[EC_ROTRIGHT]++;
		// chose the new team
		menuSettings->team++;
		// set tank 0's deselected color
		colorVals[COLOR_RED] = menuSettings->deselColor[TEAM_RED];
		// set tank 1's selected color
		colorVals[COLOR_BLUE] = menuSettings->selColor[TEAM_BLUE];
		// redraw both tanks
		TeamSelInput_redraw:
		PlaceRenderItem(&(menuSettings->teams[0].tank));
		PlaceRenderItem(&(menuSettings->teams[1].tank));
	}
	// check for selection input
	else if ((ec[EC_FIRE] == 1) && ((networkState == NETSTATE_SERVER)
	|| (networkState == NETSTATE_CLIENT))) {
		// play!
		return STATE_PLAY;
	}
	return 0;
}

