/**
 * @file text.h
 * Interface to code handling on-screen text.
 * @warning  Parts of the interface are dependent upon render.h and will be
 *           omitted unless render.h is included prior to this header file.
 * @ingroup TextSys
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#ifndef __text_h__
#define __text_h__

#if defined(_MSC_VER) && (_MSC_VER <= 1310)
// For some reason it seems Microsoft's wchar.h in Visual Studio 2003 lacks
// prototypes for several functions that are defined in the header for use
// with C++, nevermind they are part of the C standard.
#ifndef fwide
#define fwide(file, num)  (num)
#endif
// When used with C, Microsoft's wchar.h defines an old, non-standard
// swprintf() function. Work out getting the standard prototype.
#define swprintf  _nonstd_swprintf
#endif

/**
 * @defgroup TextSys Text System
 *
 * Renders text with the help of SDL_ttf and maintains the resulting surfaces
 * in RenderItems. Includes text constants that can, hopefully, be easily
 * changed to support other languages.
 *
 * To use the text system, first call TextInit(). Then, call one of the
 * TextSetupState* functions to prepare a set of constant strings for the
 * state. Before calling another TextSetupState* function, call
 * TextUnsetState() to remove all text items. Use TextRenderString() to create
 * new strings with variable content. Once finished with the system, call
 * TextUninit() to free memory used by the fonts. 
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/ 
 * @{
 */

#ifndef _render_h_
#include "bool.h"
#else
#include <stdarg.h>
#endif
#include <SDL_types.h>
#if defined(USE_GETTEXT) || defined(DOXYGEN)
#include <wchar.h>
#include <libintl.h>
/**
 * The character type used to store and manipulate strings for later output.
 * The type is the platform's wide character (wchar_t) when built with I18N
 * support, or a char.
 */
typedef wchar_t  rtsa_char;
/**
 * The character type used in the strings given to the SDL_ttf functions. It is
 * a Uint16 when build with I18N support, or a char.
 * @warning  This type may not be the same as rtsa_char.
 */
typedef Uint16   rtsa_outchar;
#ifdef WIN32
/**
 * On Windows, the gettext library supplies an fprintf() that is garenteed to
 * crash; use Microsoft's fprintf() instead.
 */
#undef fprintf
/**
 * On Windows, the gettext library supplies an fwprintf() that is garenteed to
 * crash; use Microsoft's fwprintf() instead.
 */
#undef fwprintf
#endif
/**
 * A string copy function, either wcscpy() for I18N support, or strcpy().
 */
#define rtsa_strcpy  wcscpy
/**
 * A string copy function, either wcsncpy() for I18N support, or strncpy().
 */
#define rtsa_strncpy wcsncpy
/**
 * A string compare function, either wcscmp() for I18N support, or strcmp().
 */
#define rtsa_strcmp  wcscmp
/**
 * A string length function, either wcslen() for I18N support, or strlen().
 */
#define rtsa_strlen  wcslen
/**
 * Finds the next instance of a specific character in a string using either
 * wstrchr() for I18N support, or strchr.
 */
#define rtsa_strchr  wcschr
/**
 * Writes a formatted string to another string using snwprintf(), snprintf(),
 * or a similar function.
 */
#define prnstr    swprintf
/**
 * Writes a formatted string to another string using vsnwprintf(), vsnprintf(),
 * or a similar function.
 */
#define vprnstr   vswprintf
/**
 * Writes a formatted string to stdout using wprintf(), printf(), or a
 * similar function.
 */
#define prnout    wprintf
/**
 * Writes a formatted string to a file using fwprintf(), fprintf(), or a
 * similar function.
 */
#define prnfile   fwprintf
/**
 * Writes a single character to a file using putwc(), or putc().
 */
#define witechar  putwc
/**
 * A macro to make using gettext() shorter by giving it a shorter name and
 * performing the needed type casts to avoid compiler warnings.
 */
#define _(text)  ((wchar_t*)gettext((char*)(text)))
/**
 * The function to render text. Uses TTF_RenderUNICODE_Blended() when built
 * with I18N support, or TTF_RenderText_Blended().
 */
#define RenderTTF  TTF_RenderUNICODE_Blended
/**
 * The function to report the height and width of text. Uses TTF_SizeUNICODE()
 * when built with I18N support, or TTF_SizeText().
 */
#define SizeTTF    TTF_SizeUNICODE
#else
typedef char   rtsa_char;
typedef char   rtsa_outchar;
#define rtsa_strcpy    strcpy
#define rtsa_strncpy   strncpy
#define rtsa_strcmp    strcmp
#define rtsa_strlen    strlen
#define rtsa_strchr    strchr
#undef  snwprintf
#define snwprintf      snprintf
#define prnstr         snprintf
#define vprnstr        vsnprintf
#define prnout         printf
#define prnfile        fprintf
#define wprintf        printf
#define witechar       putc
#define gettext(text)  (text)
#define _(text)        (text)
#define RenderTTF      TTF_RenderText_Blended
#define SizeTTF        TTF_SizeText
#endif

// Several ISO C functions are not available on older MSVCs, but the gettext
// library will supply them
#ifndef USE_GETTEXT
#if defined(_MSC_VER) && (_MSC_VER <= 1310)
#undef snwprintf
// Make all calls of swprintf() go to a function that uses the standard
// prototype. This function does not guarentee a null-terminated result,
// unlike the standard function, so problems could result. Use a newer version
// of Visual Studio (like 2005) to avoid the trouble altogether.
#define snwprintf  _snwprintf
#define vsnprintf  _vsnprintf
#define vswprintf  _vsnwprintf
#endif
#ifdef _MSC_VER
#define snprintf   _snprintf
#endif
#endif

/**
 * Indices into an array of RenderItems used to render text. Serveral of the
 * items have the same value because they are used during different
 * @ref OperatingStates "operating states" and will never be used together. 
 */
enum {
	// ---------- STATE_SPLASH ----------
	/**
	 * The title of the game in a large font.
	 */
	TEXTITEM_TITLE,
	/**
	 * The author of the game.
	 */
	TEXTITEM_TRANS,
	/**
	 * The translator for the language the game is currently using.
	 */
	TEXTITEM_AUTH,
	/**
	 * Copyright year.
	 */
	TEXTITEM_COPYRIGHT,
	/**
	 * The multi-line GPL notice.
	 */
	TEXTITEM_GPL,
	// ---------- STATE_BFLDERR ----------
	// Also uses TEXTITEM_TITLE
	/**
	 * The error message.
	 */
	TEXTITEM_ERROR = 1,
	/**
	 * Prompt to press fire to continue.
	 */
	TEXTITEM_ERRCONT,
	// ---------- STATE_MENU ----------
	// Also uses TEXTITEM_TITLE
	/**
	 * Prompt to chose a team to join.
	 */
	TEXTITEM_CHOSETEAM = 1,
	/**
	 * Prompt to see more options by pressing fire.
	 */
	TEXTITEM_FIREOPT,
	/**
	 * The network status message, one of:
	 * - ::STRING_NETLOOKING
	 * - ::STRING_NETCONTACT
	 * - ::STRING_NETCONNECT
	 * - ::STRING_NETSERVER
	 */
	TEXTITEM_NETSTAT,
	/**
	 * The networked game options.
	 */
	TEXTITEM_NETOPTS,
	/**
	 * The number of red team members.
	 */
	TEXTITEM_REDMEM,
	/**
	 * The number of blue team members.
	 */
	TEXTITEM_BLUEMEM,
	// ---------- STATE_PLAY ----------
	/**
	 * Red team's label for the score display.
	 */
	TEXTITEM_REDTEAM = 0,
	/**
	 * Blue team's label for the score display.
	 */
	TEXTITEM_BLUETEAM,
	/**
	 * Red team's score. This value will vary during the game.
	 */
	TEXTITEM_REDSCORE,
	/**
	 * Blue team's score. This value will vary during the game.
	 */
	TEXTITEM_BLUESCORE,
	/**
	 * A prompt for the user to press fire to make their tank respawn.
	 */
	TEXTITEM_RESPAWN,
	TEXTITEM_MISC,
	TEXTITEM_MAX
};

/**
 * The available fonts. The freely distributable Bitstream Vera
 * font availble from http://www.gnome.org/fonts/ is used for now. To support
 * more languages, especially non-european ones, a different font will be
 * required.
 */
enum FONTS {
	/**
	 * A large font used to display these messages:
	 * - ::STRING_REDFLAGHELD
	 * - ::STRING_REDFLAGCAP
	 * - ::STRING_REDWIN
	 * - ::STRING_BLUEFLAGHELD
	 * - ::STRING_BLUEFLAGCAP
	 * - ::STRING_BLUEWIN
	 * - ::STRING_CHOSETEAM
	 *
	 * @note  The more appropriate fonts availble from the AtariAge web site
	 *        (http://www.atariage.com/2600/archives/AtariFonts/?SystemID=2600)
	 *        appear to be illegal to distribute, with the exception of the
	 *        Eras fonts.
	 */
	FONT_MESSAGE,
	/**
	 * A medium sized font used to display score lists.
	 */
	FONT_SCORE,
	/**
	 * A small font used for player name labels.
	 */
	FONT_PLAYER,
	FONT_MAX
};

#if defined(_render_h_) || defined(DOXYGEN)
/**
 * The array of text items that can be displayed. If their data pointer
 * (RenderItem::data) is NULL, then the item is not presently in use, is not
 * a memeber of a Layer, and a surface containing the text does not exist.
 */
extern RenderItem textItems[TEXTITEM_MAX];
#endif

/**
 * Indices into an array of constant strings used for messages
 * presented to the user.
 */
enum {
	/**
	 * Label for the red team.
	 */
	STRING_REDTEAM,
	/**
	 * Label for the blue team.
	 */
	STRING_BLUETEAM,
	/**
	 * Notice that the blue team holds the red flag.
	 */
	STRING_REDFLAGHELD,
	/**
	 * Notice that the red team's flag was successfully captured.
	 */
	STRING_REDFLAGCAP,
	/**
	 * Notice that the red team holds the blue flag.
	 */
	STRING_BLUEFLAGHELD,
	/**
	 * Notice that the blue team's flag was successfully captured.
	 */
	STRING_BLUEFLAGCAP,
	/**
	 * Notice that someone killed one of their teammates.
	 */
	STRING_TEAMKILL,
	/**
	 * Notice that the local player killed one of his/her teammates.
	 */
	STRING_LOCALTEAMKILLER,
	/**
	 * Notice that the local player was killed by one of his/her teammates.
	 */
	STRING_LOCALTEAMKILLED,
	/**
	 * Notice that someone killed one of their opponents.
	 */
	STRING_KILL,
	/**
	 * Notice that the local player killed one of his/her opponents.
	 */
	STRING_LOCALKILL,
	/**
	 * Notice that the red team has won the game.
	 */
	STRING_REDWIN,
	/**
	 * Notice that the blue team has won the game.
	 */
	STRING_BLUEWIN,
	/**
	 * Message informing the player that pressing fire will respawn their tank.
	 */
	STRING_RESPAWN,
	/**
	 * The program's name.
	 */
	STRING_TITLE,
	/**
	 * The author's name and email address.
	 */
	STRING_AUTH,
	/**
	 * The translator's credit. Set to "translator" for English (no translator).
	 */
	STRING_TRANS,
	/**
	 * Copyright notice.
	 */
	STRING_COPYRIGHT,
	/**
	 * GPL notice.
	 */
	STRING_GPL,
	/**
	 * Request that the user select a team to join.
	 */
	STRING_CHOSETEAM,
	/**
	 * Notice to the player that pressing fire will show an options menu.
	 */
	STRING_FIREOPT,
	/**
	 * A prompt for the user to press the fire button to continue on to the
	 * next screen.
	 */
	STRING_FIRECONT,
	/**
	 * Notice that the program is attempting to find a server on the LAN.
	 */
	STRING_NETLOOKING,
	/**
	 * Notice that the program is attempting to contact a specific server, such
	 * as one specified by the player. The string should include a "%s" that
	 * will be expanded to the hostname being contacted.
	 */
	STRING_NETCONTACT,
	/**
	 * A message telling the user that the game has contacted a server
	 * (hostname is %s) and is waiting on game option/config data before the
	 * user can join the game in progress.
	 */
	STRING_NETRCVCFG,
	/**
	 * Notice that a connection has been made to a server and this process is
	 * functioning as a server. The string should
	 * include a "%s" that will be expanded to the hostname of the server.
	 */
	STRING_NETCLIENT,
	/**
	 * A message informing the user that the game client is disconnecting from
	 * the server.
	 */
	STRING_NETCLIDISCONN,
	/**
	 * Notice that this process is functioning as a game server. The string
	 * should include a "%s" that will be expanded to the hostname of the local
	 * system.
	 */
	STRING_NETSERVER,
	/**
	 * A message informing the user that the server is disconnecting all game
	 * clients.
	 */
	STRING_NETSRVDISCONN,
	/**
	 * A message telling the user that the server (hostname %s) will not allow
	 * the game client to connect.
	 */
	STRING_NETDENY,
	/**
	 * A status message indicating the game has disconnected from the server
	 * or all clients.
	 */
	STRING_NETDISCONN,
	/**
	 * An error message informing the user that the process of finding a server
	 * on the LAN could not be performed. This does not mean that a server
	 * could not be found, but that the finding could not be done. The error
	 * string is %ls and its translations are in net.c.
	 */
	STRING_NETSEARCHERR,
	/**
	 * An error message indicating that the game could not run as a server.
	 * The error string is %ls and its translations are in net.c.
	 */
	STRING_NETSRVERR,
	/**
	 * An error message telling the user that the game acting as a client
	 * could not connect to a server. The error is %ls.
	 */
	STRING_NETCONERR,
	/**
	 * A status message showing this computer's network address (%s).
	 */
	STRING_NETADDR,
	/**
	 * A message informing the user that time synchronization with the server
	 * is in progress.
	 */
	STRING_NETSYNC,
	/**
	 * A message with the latency result of time synchronization. The latency
	 * value is given as a double (I recommed using %0.1lf for one digit after
	 * the decimal point). The units are milliseconds.
	 */
	STRING_NETLAT,
	/**
	 * The player name used when a name cannot be determined from the
	 * enviornment variables.
	 */
	STRING_DEFAULTNAME,
	/**
	 * The time format. Only minutes and seconds are provided.
	 */
	STRING_TIMEFORM,
	/**
	 * The time that remains for the round of combat.
	 */
	STRING_TIMELEFT,
	/**
	 * A format for status output showing a summary of information read from
	 * the battlefield file(s). This output is only shown on the console
	 * to assist people in creating their own battlefields.
	 */
	STRING_BF_STAT,
	/**
	 * A success/debugging message sent to the console that indicates the
	 * battlefield data is ready for network transmission.
	 */
	STRING_CMPLOPT,
	/**
	 * The format to present any errors encountered while reading the
	 * battlefield file(s).
	 */
	STRING_ERR_READ,
	/**
	 * The format to present any errors encountered while processing the
	 * battlefield data that isn't attributed to any particular file.
	 */
	STRING_ERR_PROC,
	/**
	 * Too little memory to handle the battlefield file.
	 */
	STRING_ERR_MEM,
	/**
	 * The battlefield file could not be opened.
	 */
	STRING_ERR_FOPEN,
	/**
	 * The battlefield file ends inside an include statement.
	 */
	STRING_ERR_EOFINC,
	/**
	 * The name of a battlefield file to include continues to the end of the
	 * file. It should be terminated by either a semicolon (';') or the end
	 * of a line.
	 */
	STRING_ERR_EOFNAME,
	/**
	 * The name of a battlefield file to include is too long.
	 */
	STRING_ERR_NAMELONG,
	/**
	 * The size of the battlefield is too small.
	 */
	STRING_ERR_FLDSMALL,
	/**
	 * The size of the battlefield is too large.
	 */
	STRING_ERR_FLDLARGE,
	/**
	 * The battlefield size is not specified.
	 */
	STRING_ERR_NOSIZE,
	/**
	 * An invalid keyword is in the battlefield file.
	 */
	STRING_ERR_INVLDKEY,
	/**
	 * No spawn area for the red team.
	 */
	STRING_ERR_NOREDSPN,
	/**
	 * No spawn area for the blue team.
	 */
	STRING_ERR_NOBLSPN,
	/**
	 * No goal area specified for the red team.
	 */
	STRING_ERR_NOREDGOAL,
	/**
	 * No goal area specified for the blue team.
	 */
	STRING_ERR_NOBLGOAL,
	/**
	 * Illegal use of a backslash in the path to a file.
	 */
	STRING_ERR_BACKSLASH,
	/**
	 * I/O error while reading from a battlefield file.
	 */
	STRING_ERR_IO,
	/**
	 * Line break before the end of a token.
	 */
	STRING_ERR_EOL,
	/**
	 * Too few parameters were given to a statement in a battlefield file.
	 */
	STRING_ERR_PARAM,
	/**
	 * A specified coordinate is out of the allowable range.
	 */
	STRING_ERR_XYRANGE,
	/**
	 * A specified height or width is out of the allowable range.
	 */
	STRING_ERR_WHRANGE,
	/**
	 * An invalid team was specified as the owning team of either a goal or
	 * spawn area.
	 */
	STRING_ERR_BADOWN,
	/**
	 * An error string for no error just in case.
	 */
	STRING_ERR_NONE,
	/**
	 * Network non-error for an operation that would have blocked if the socket
	 * wasn't non-blocking.
	 */
	STRING_ERR_WOULDBLOCK,
	/**
	 * Bad file descriptor.
	 */
	STRING_ERR_BADF,
	/**
	 * Bad pointer given; dereferencing it will case a segmentation fault.
	 */
	STRING_ERR_FAULT,
	/**
	 * The operation was interrupted.
	 */
	STRING_ERR_INTR,
	/**
	 * A general error indicating that something is invalid.
	 */
	STRING_ERR_INVAL,
	/**
	 * The given file descriptor is not a socket but was used as a socket.
	 */
	STRING_ERR_NOTSOCK,
	/**
	 * The requested socket address is already in use.
	 */
	STRING_ERR_ADDRINUSE,
	/**
	 * The network is down.
	 */
	STRING_ERR_NETDOWN,
	/**
	 * The network is unreachable.
	 */
	STRING_ERR_NETUNREACH,
	/**
	 * The connection was reset.
	 */
	STRING_ERR_CONNRESET,
	/**
	 * The connection was refused.
	 */
	STRING_ERR_CONNREFUSED,
	/**
	 * The host is unreachable.
	 */
	STRING_ERR_HOSTUNREACH,
	/**
	 * A header for a specific error on a host lookup.
	 */
	STRING_ERR_HOSTLOOK,
	/**
	 * SDL was not initialized.
	 */
	STRING_ERR_SDL,
	/**
	 * A query for video output information (SDL_GetVideoInfo()) failed.
	 */
	STRING_ERR_VIDQ,
	/**
	 * An attempt to set the video output mode (SDL_SetVideoMode()) failed.
	 */
	STRING_ERR_VIDMODE,
	/**
	 * The SDL_ttf library failed to initialize.
	 */
	STRING_ERR_TTFINIT,
	/**
	 * The rendering system failed to initialize.
	 */
	STRING_ERR_RENDER,
	/**
	 * A font file could not be opened.
	 */
	STRING_ERR_FNTOPEN,
	/**
	 * The font(s) failed to load.
	 */
	STRING_ERR_FNTINIT,
	/**
	 * Failed to open an audio output device.
	 */
	STRING_ERR_OPENAUDIO,
	/**
	 * Failed to load sound from a specific file.
	 */
	STRING_ERR_LOADSND,
	/**
	 * No sounds were loaded.
	 */
	STRING_ERR_NOSNDS,	
	/**
	 * The most generic error -- something is wrong and there is no other error
	 * string for it.
	 */
	STRING_ERR_DEFAULT,
	/**
	 * Using single buffered output.
	 */
	STRING_STAT_SINGLEBUFF,
	/**
	 * Using double buffered output.
	 */
	STRING_STAT_DOUBLEBUFF,
	/**
	 * No hardware acceleration on video output.
	 */
	STRING_STAT_NOHWACCEL,
	/**
	 * Using hardware acceleration on video output.
	 */
	STRING_STAT_HWACCEL,
	/**
	 * The video output buffer is stored in system/main memory.
	 */
	STRING_STAT_OUTSYSMEM,
	/**
	 * The video output buffer is stored in the graphics hardware's memory.
	 */
	STRING_STAT_OUTGRPMEM,
	/**
	 * Formatting string to show the video driver in use.
	 */
	STRING_STAT_VIDDRNAME,
	/**
	 * Formatting string for some joystick or other controller information.
	 */
	STRING_STAT_JOYSTICK,
	/**
	 * Notification that audio output could not be initialized.
	 */
	STRING_STAT_NOAUDIO,
	/**
	 * Notification that the program has entered the main state loop.
	 */
	STRING_STAT_ENTERSTATE,
	/**
	 * Notification that the program has left the main state loop.
	 */
	STRING_STAT_EXITSTATE,
	STRING_MAX
};

/**
 * An array of strings with message text. Limits the code changes to support
 * additional languages.
 */
extern rtsa_char * strings[STRING_MAX];

/**
 * Initalizes the text system's string table and support for translation
 * through GNU gettext (if compiled to support gettext). Does not initialize
 * fonts or any graphics releated data -- FontInit() handles that.
 * @return  True on success, false if the system could not be initalized.
 * @author  Jeff Jackowski
 */
Bool TextInit();

/**
 * Initalizes fonts and other graphical parts of the text system.
 * @return  True on success, false if the system could not be initalized.
 * @author  Jeff Jackowski
 */
Bool FontInit();

/**
 * Configures the text system for use with STATE_CODES::STATE_SPLASH.
 * @return  True if the operation succeeded, false if it failed.
 * @author  Jeff Jackowski
 */
Bool TextSetupStateSplash();

/**
 * Configures the text system for use with STATE_CODES::STATE_MENUROOT and
 * other menu states.
 * @return  True if the operation succeeded, false if it failed.
 * @author  Jeff Jackowski
 */
Bool TextSetupStateMenu();

/**
 * Configures the text system for use with STATE_CODES::STATE_PLAY.
 * @return  True if the operation succeeded, false if it failed.
 * @author  Jeff Jackowski
 */
Bool TextSetupStatePlay();

/**
 * Removes rendered text items in preperation for a new set to be rendered.
 * @author  Jeff Jackowski
 */
void TextUnsetState();

/**
 * Uninitalizes the text system and loaded fonts. Does the opposite of both
 * TextInit() and FontInit().
 * @author  Jeff Jackowski
 */
void TextUninit();

/**
 * Converts a string from the internal format to big endian UTF-16.
 * @param dest    The destination string buffer.
 * @param destlen The length of the destination buffer in characters
 *                (2 bytes each).
 * @param src     The source string in the native format.
 * @param srclen  The length of the source string in characters.
 * @return  The number of characters written to the destination buffer, or a
 *          negative value on an error.
 * @author  Jeff Jackowski
 */
int TextNativeToUTF16(Uint16 *dest, size_t destlen, rtsa_char *src,
size_t srclen);

/**
 * Sets the title used on the program's window. The title used is taken from
 * ::strings[::STRING_TITLE] if UTF-8 support is available in SDL, otherwise
 * an English title is used.
 * @author  Jeff Jackowski
 */
void TextSetWindowTitle();

/**
 * Converts a string from big endian UTF-16 to the internal format.
 * @param dest    The destination string buffer.
 * @param destlen The length of the destination buffer in characters.
 * @param src     The source string in the native format.
 * @param srclen  The length of the source string in characters (2 bytes each).
 * @return  The number of characters written to the destination buffer, or a
 *          negative value on an error.
 * @author  Jeff Jackowski
 */
int TextUTF16toNative(rtsa_char *dest, size_t destlen, Uint16 *src,
size_t srclen);

#if defined(_render_h_) || defined(DOXYGEN)
/**
 * Renders a string to an SDL_Surface stored at text->data, sets the size to
 * match, and sets the render function to GenericSurfaceRenderer(). If a
 * surface already exists in the render item, it will be destroyed before being
 * replaced with the newly rendered text.
 * @param text    The item that will represent the rendered text. The surface
 *                pointed to by RenderItem::data should either be a valid
 *                surface, which will be freed, or NULL.
 * @param font    The index of the font to use. See ::FONTS.
 * @param maxwidth  The maximum width of the text in pixels. If zero or less,
 *                  no maximum width will be enforced.
 * @param color   A display format independent color to use for the text.
 * @param format  The formatting string, printf() style, used to control the
 *                output.
 * @param  ...    Additional variables used to fill in fields of the formatting
 *                string.
 * @return  True on success, false if the string could not be rendered.
 * @author  Jeff Jackowski
 */
Bool TextRenderString(RenderItem *text, int font, int maxwidth, SDL_Color color,
const rtsa_char *format, ...);

/**
 * Renders a string to an SDL_Surface stored at text->data, sets the size to
 * match, and sets the render function to GenericSurfaceRenderer(). If a
 * surface already exists in the render item, it will be destroyed before being
 * replaced with the newly rendered text.
 * @param text    The item that will represent the rendered text. The surface
 *                pointed to by RenderItem::data should either be a valid
 *                surface, which will be freed, or NULL.
 * @param font    The index of the font to use. See ::FONTS.
 * @param maxwidth  The maximum width of the text in pixels. If zero or less,
 *                  no maximum width will be enforced.
 * @param color   A display format independent color to use for the text.
 * @param format  The formatting string, printf() style, used to control the
 *                output.
 * @param  args   A variable argument list used to fill in fields of the
 *                formatting string.
 * @return  True on success, false if the string could not be rendered.
 * @author  Jeff Jackowski
 */
Bool TextRenderStringV(RenderItem *text, int font, int maxwidth, SDL_Color color,
const rtsa_char *format, va_list args);

#endif

#endif

/** @} */

