/**
 * @file timesync.h
 * Interface to the @ref NetTimeSync module.
 * @ingroup NetTimeSync
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

#include <SDL_types.h>

/**
 * @defgroup NetTimeSync Time Synchronization
 *
 * A system of approximating the time the server is using to assist in having
 * all clients show a game state that is as current as possible.
 *
 * The system consists of two parts: client and server. The client first
 * connects to a server, then calls StartClientSync(). The time sync process
 * occurs on the @ref IOThread. ::timeSyncState will
 * change to reflect the status of the time sync operation (see TSYNC_STATE
 * for the valid codes).
 *
 * The server is started by a call to StartServerTimeSync() which will start
 * another thread to handle incomming client time sync requests. To stop the
 * time sync server, call StopServerTimeSync().
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/ 
 *
 * @author Jeff Jackowski
 * @ingroup Networking
 * @{
 */

enum TSYNC_STATE {
	/**
	 * Indicates that the time synchronization process is not running and there
	 * are no valid results.
	 */
	TIMESYNC_IDLE,
	/**
	 * Indicates that time synchronization is in progress and results are not
	 * yet available.
	 */
	TIMESYNC_INPROG,
	/**
	 * Indicates that the time synchronization process has completed
	 * successfully and valid results are available.
	 */
	TIMESYNC_FINISHED,
	/**
	 * Indicates that the time synchronization process could not determine a
	 * good time delta and any attempt to play may not work well.
	 */
	TIMESYNC_POOR,
	/**
	 * Indicates that the time synchronization process terminated on an error
	 * and no results are available.
	 */
	TIMESYNC_ERROR
};

/**
 * Data from a single time sync sample. Several are collected to produce the
 * final results. The units are milliseconds.
 */
struct TimeSample_t {
	/**
	 * The approximate time required for data to reach the server. The
	 * computation of this value assumes that the round-trip time is double
	 * the time from client to server.
	 */
	double latency;
	/**
	 * The computed time delta based on the returned server time and the
	 * latency of this sample.
	 */
	double delta;
};

typedef struct TimeSample_t  TimeSample;

/**
 * The final results of the last time sync operation that ended with a state
 * of either TSYNC_STATE::TIMESYNC_FINISHED or TSYNC_STATE::TIMESYNC_POOR.
 */
extern TimeSample timeSyncResults;
/**
 * The time delta used to produce common game time. This is the integer version
 * of ::timeSyncResults.delta.
 */
extern Sint32 timeDelta;
/**
 * The state of the time synchronization system; a value from TSYNC_STATE.
 */
extern int timeSyncState;

/**
 * Starts the time synchronization process on the @ref IOThread.
 * @pre    The process is a client that has successfully connected to a server.
 * @author Jeff Jackowski
 */
void StartClientTimeSync();

/**
 * Starts the time synchronization server on a separate thread.
 *
 * @todo   Fix this to better support IPv6.
 *
 * @return True on success, false if the server thread could not be started.
 * @author Jeff Jackowski
 */
Bool StartServerTimeSync();

/**
 * Stops the time synchronization server and returns the time sync state to
 * TSYNC_STATE::TIMESYNC_IDLE.
 * @author Jeff Jackowski
 */
void StopServerTimeSync();

/** @} */

