/**
 * @file vector.h
 * The interface to a generic vector with binary heap functionality.
 * @ingroup VectorG
 * @author Jeff Jackowski  (jeffj@ro.com)
 *
 * Copyright (C) 2007  Jeff Jackowski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA. Or visit their website at: http://www.gnu.org/
 */

/**
 * @defgroup VectorG Vector
 * A resizable array that can optionally act as a binary heap.
 * @ingroup DataStruct
 *
 * @{
 */

#ifndef _vector_h_
#define _vector_h_

#include "bool.h"

/**
 * A function that determines if an item is less than another.
 *
 * @param item1  The address within the vector of the first operand.
 * @param item2  The address within the vector of the second operand.
 * @return  True if item1 < item2.
 */
typedef Bool (*LessThanOp)(void *item1, void *item2);

/**
 * A function that handles deinitialization for an item.
 *
 * @param item  The item to deinitialize. It must not be deallocated in this
 *              funtion.
 */
typedef void (*DestructorOp)(void *item);

/**
 * A generic destructor for use with vectors of pointers. It calls free() on
 * the pointer in the Vector if the pointer is non-NULL.
 * @param  item  The item to deinitialize by deallocating it. The item is
 *               really a double pointer.
 */
void GenericPointerFree(void *item);

/**
 * Computes the index of the parent item in a binary heap given the child's
 * index.
 *
 * @param i  The child's index.
 * @return   The parent's index.
 * @relates  Vector_t
 */
#define HEAP_PARENT_INDEX(i)  (((i) - 1) >> 1)

/**
 * Computes the index of the first child item in a binary heap given the
 * parent's index.
 *
 * @param i  The parent's index.
 * @return   The first child's index.
 * @relates  Vector_t
 */
#define HEAP_CHILD_INDEX(i)   (((i) << 1) + 1)

/**
 * A generic vector that can hold a varying number of items that are all the
 * same size. This structure holds information describing a vector and the data
 * contained within.
 * @author  Jeff Jackowski
 */
struct Vector_t {
	/**
	 * The function used to compare items in the vector for sorting operations.
	 * It may be set to NULL if the sorting functions are not used.
	 */
	LessThanOp lessOp;
	/**
	 * The function used to deinitialize items in the vector prior to
	 * deallocation. If set to NULL, no such function will be called.
	 */
	DestructorOp destructOp;
	/**
	 * The contiguous array of items. With the exception of VectorInit(), all
	 * vector functions assume array points to valid allocated space.
	 * @warning  Do not change this value directly.
	 */
	void  *array;
	/**
	 * The size of the item type stored in the vector in bytes. This should be
	 * the same value returned by the sizeof operator.
	 * @warning  Do not change this value after initializing the vector.
	 */
	unsigned int itemSize;
	/**
	 * The number of items present in the vector.
	 * @warning  Do not change this value directly.
	 */
	unsigned int items;
	/**
	 * The allocated size of the vector in terms of the number of elements that
	 * can be stored.
	 * @warning  Do not change this value directly.
	 */
	unsigned int size;
	/**
	 * The minimum number of items that the vector will hold.
	 * @warning  Do not change this value after initializing the vector.
	 */
	unsigned int minSize;
};

typedef struct Vector_t Vector;

/**
 * Initializes a Vector by setting some initial values and allocating space for
 * the array. This function assumes the array pointer in the given vector is
 * uninitialized. When compiled for debugging, each byte of the array data will
 * be initialized to 0xCD to help find cases where data is incorrectly used
 * uninitialized.
 * @param vec       The vector to initialize.
 *        @pre      The following fields must be
 *                  set prior to calling this function:
 *                  - Vector::itemSize
 *                  - Vector::minSize
 *
 *        @post     The following fields are not initalized by this function and
 *                  can be set before or after the call:
 *                  - Vector::lessOp -- only needed if the heap sorting
 *                    functions are used
 *                  - Vector::destructOp -- must be set either to a destructor
 *                    function, or to NULL
 *
 * @param initSize  The initial size of the vector given in the number of items.
 *        @warning  The value of initSize must not be less than Vector::minSize.
 * @return  True on success, false if the array could not be allocated
 *          (inadequate memory).
 * @relates  Vector_t
 * @author  Jeff Jackowski
 */
Bool VectorInit(Vector *vec, unsigned int initSize);

/**
 * Destroys the contents of a Vector and deallocates its array. If
 * Vector::destructOp is not NULL, it will be called for each item still
 * inside the vector.
 * @param vec  The vector to destroy. The array will be deallocated, but vec
 *             will not so that it can be allocated somewhere other than the
 *             heap. The fields of vec will not be changed unless compiled for
 *             debugging. Debug builds will fill every byte of vec with 0xAB
 *             to assist in finding inappropriate use of destroyed vectors.
 * @relates  Vector_t
 * @author  Jeff Jackowski
 */
void VectorDestroy(Vector *vec);

/**
 * Copies the contents of a Vector into a new vector.
 * @param target  An uninitialzied vector that will receive the copy.
 *                Unlike with VectorInit(), there is no need to initialize any
 *                fields because they will be copied from the source vector.
 * @param source  The vector with the contents to copy.
 * @return   True on success. Failure will occur if there is inadequate memory.
 * @relates  Vector_t
 * @author  Jeff Jackowski
 */
Bool VectorCopy(Vector *target, Vector *source);

/**
 * Destroys the contents of a Vector, but not the Vector. The Vector's array
 * is not deallocated, but it may shrink. If Vector::destructOp is not NULL,
 * it will be called for each item still inside the vector.
 *
 * @param vec  The vector to operate on.
 * @relates  Vector_t
 * @author  Jeff Jackowski
 */
void VectorClear(Vector *vec);

/**
 * Adds an item to the end of the Vector and returns a pointer to the new,
 * uninitialized item.
 * 
 * @param vec  The Vector to add upon.
 * @return     The address of the new item, or NULL if resizing the vector
 *             failed.
 * @relates    Vector_t
 * @author     Jeff Jackowski
 */
void *VectorAdd(Vector *vec);

/**
 * Copies the given item to the end of the Vector and returns a pointer to the
 * item within the Vector.
 * 
 * @param vec  The Vector to operate upon.
 * @param item The item to copy into the Vector.
 * @return     The address of the item in the Vector, or NULL if resizing the
 *             vector failed.
 * @relates    Vector_t
 * @author     Jeff Jackowski
 */
void *VectorPush(Vector *vec, void *item);

/**
 * Copies the given item into the Vector and places it in heap order.
 * 
 * @pre  All items already in the vector are in heap order, or the vector is
 *       empty.
 *
 * @pre  The given vector must have a function set for Vector::lessOp prior
 *       to calling this function.
 * 
 * @param vec  The Vector to operate on.
 * @param item The item to copy into the Vector.
 * @return     The address of the item in the Vector, or NULL if resizing the
 *             vector failed.
 * @relates  Vector_t
 * @author  Jeff Jackowski
 */
void *VectorHeapPush(Vector *vec, void *item);

/**
 * Removes the item at the end of a Vector. If Vector::destructOp is not NULL,
 * it will be called on the item.
 * 
 * @param vec  The Vector to operate on.
 * @relates  Vector_t
 * @author  Jeff Jackowski
 */
void VectorPop(Vector *vec);

/**
 * Removes the first item in the Vector and maintains heap order. If
 * Vector::destructOp is not NULL, it will be called on the removed item.
 * 
 * @pre  All items already in the Vector are in heap order, or the vector is
 *       empty.
 *
 * @pre  The given vector must have a function set for Vector::lessOp prior
 *       to calling this function.
 * 
 * @param vec  The Vector to operate on.
 * @relates  Vector_t
 * @author   Jeff Jackowski
 */
void VectorHeapPop(Vector *vec);

/**
 * Uses the heap sort algorithm to sort the items in the Vector.
 * 
 * @pre  The given vector must have a function set for Vector::lessOp prior
 *       to calling this function.
 *
 * @post  The vector's items are in accending order, not heap order.
 * @param vec  The Vector to operate on.
 * @relates  Vector_t
 * @author   Jeff Jackowski
 */
void VectorHeapSort(Vector *vec);

/**
 * Inserts an uninitalized item within the given vector. When compiled for
 * debugging, each byte of the new item data will
 * be initialized to 0xCD to help find cases where data is incorrectly used
 * uninitialized.
 * 
 * @param vec    A pointer to the Vector to operate on.
 * @param index  The index to place the new item. All items presently at or
 *               past this index will be moved to a higher index. The index may
 *               be any value from zero to one past the end of the vector.
 * @return   The address of the inserted item, or NULL if resizing the
 *           vector failed.
 * @relates  Vector_t
 * @author   Jeff Jackowski
 */
#define VectorInsert(vec, index)  VectorInsertMany(vec, index, 1)

/**
 * Makes room for multiple items in the Vector. The new items will be
 * uninitialzied. When compiled for debugging, each byte of the new item data
 * will be initialized to 0xCD to help find cases where data is incorrectly
 * used uninitialized.
 * 
 * @param vec    The Vector to operate on.
 * @param index  The first index of the range to insert. All items presently at
 *               or past this index will be moved to a higher index. The index
 *               may be any value from zero to one past the end of the vector.
 * @param count  The number of items to insert.
 * @return   The address of the first inserted item, or NULL if resizing the
 *           vector failed.
 * @relates  Vector_t
 * @author   Jeff Jackowski
 */
void *VectorInsertMany(Vector *vec, unsigned int index, unsigned int count);

/**
 * Removes an item from a Vector.
 * 
 * @param vec    A pointer to the Vector to operate on.
 * @param index  The index to remove. All items past this index will be moved
 *               to a lower index.
 * @relates  Vector_t
 * @author   Jeff Jackowski
 */
#define VectorRemove(vec, index)  VectorRemoveMany(vec, index, 1)

/**
 * Removes multiple items from a Vector.
 * 
 * @param vec    The Vector to operate on.
 * @param index  The first index to remove. All items past this index will be
 *               moved to a lower index.
 * @param count  The number of items to remove. The items to remove must not
 *               go past the end of the items in the Vector's array.
 * @relates  Vector_t
 * @author   Jeff Jackowski
 */
void VectorRemoveMany(Vector *vec, unsigned int index, unsigned int count);

/**
 * Reorders the items in a Vector into heap order.
 * 
 * @pre  The given vector must have a function set for Vector::lessOp prior
 *       to calling this function.
 *
 * @post  The vector's items are in heap order.
 * 
 * @param vec  The Vector to operate on.
 * @relates  Vector_t
 * @author   Jeff Jackowski
 */
void VectorMakeHeap(Vector *vec);

/**
 * Reserves space for a given number of items. If the Vector can already store
 * the specified number of items or more, no action is taken.
 * 
 * @param vec    The Vector to operate on.
 * @param size   The minimum number of items the Vector should store after this
 *               function is finished.
 * @return  True on success.
 * @relates Vector_t
 * @author  Jeff Jackowski
 */
Bool VectorReserve(Vector *vec, unsigned int size);

/**
 * A convienience macro for getting the first item in a Vector.
 * 
 * @param  vec   A pointer to the Vector with the item.
 * @return  A pointer to the item or NULL if it does not exist.
 * @relates Vector_t
 * @author  Jeff Jackowski
 */
#define VectorFirst(vec)  (((vec)->items) ? (vec)->array : NULL)

/**
 * A convienience macro for getting the last item in a Vector.
 * 
 * @param  vec  A pointer to the Vector with the item.
 * @return  A pointer to the item or NULL if it does not exist.
 * @relates Vector_t
 * @author  Jeff Jackowski
 */
#define VectorLast(vec)   (((vec)->items) ? \
             &(((char*)(vec)->array)[((vec)->items - 1) * (vec)->itemSize]) : \
			 NULL)

/**
 * A convienience macro for getting some item in a Vector.
 * 
 * @param  vec   A pointer to the Vector with the item.
 * @param  index The index of the item to retrieve.
 * @warning      Do not write code that modifies the value of index -- it will
 *               be run twice.
 * @return  A pointer to the item or NULL if it does not exist.
 * @relates Vector_t
 * @author  Jeff Jackowski
 */
#define VectorItem(vec, index)  (((unsigned)(index) < (vec)->items) ? \
                   &(((char*)(vec)->array)[index * (vec)->itemSize]) : NULL)

#endif

/** @} */

